// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Mnemonic message object templates.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef mn_messagebase_hh__
#define mn_messagebase_hh__

#include "mnemonic/ptr.hh"
#include "mnemonic/priority_list.hh"
#include "mnemonic/midentifier.hh"

#include <stdexcept>
#include <typeinfo>
#include <set.h>
#include <algo.h>

class untyped_interest;
class untyped_message;

class messagebase_error : public logic_error {
	public:
		enum status_type { no_recipients, no_handling_recipients }; 
		messagebase_error(status_type, const char *str);
		messagebase_error(status_type, const char *messagetype, const char *str);
		status_type status(void) const;
		string      status_as_text(void) const;
		string      message_type(void) const;
		virtual const char* what () const;
	private:
		string      messagetype_;
		status_type stat_;
		string		local_what_;
};

ostream& operator<<(ostream&, const messagebase_error&);

class group_identifiers {
	public:
		bool add_group_id(const midentifier&);
		bool remove_group_id(const midentifier&);
		bool add_group_ids(const group_identifiers&);
		bool remove_group_ids(const group_identifiers&);
		void set_group_ids(const group_identifiers&);

		typedef set<midentifier, less<midentifier> >::iterator       iterator;
		typedef set<midentifier, less<midentifier> >::const_iterator const_iterator;
		set<midentifier, less<midentifier> >  ids;
		mutable thread_mutex                  ids_mutex;

		bool operator<(const group_identifiers& other) const;

		friend ostream& operator<<(ostream&, const group_identifiers&);
};

class untyped_interest : public group_identifiers {
	public:
		enum interest_priority { highest=0, normal=127, lowest=255 };
		untyped_interest(void);
		virtual void unregister(void);
		virtual ~untyped_interest();
		enum callresult { call_next, pass_on, exit_loop, refused };

		// call_next: message accepted, call next interest
		// pass_on:   message accepted but refused on second thought FIXME: this doubles refused now.
		// exit_loop: don't call any other interests
		// refused:   unhandled, call next interest

		enum acceptance_type { ignore_ids,
									  mid_contains_iid,
									  iid_contains_mid,
									  mid_equals_iid   } acceptance;
};

class untyped_message { // one static instance of this object for each message type
	public:
		untyped_message();
		virtual ~untyped_message();
		void register_interest(untyped_interest::interest_priority priority, ptr<untyped_interest>);
		void unregister_interest(untyped_interest *u);
		bool handle_docall_result(untyped_interest::callresult result, int& reccount) const;
		class send_result {
			public:
				send_result(void);
				send_result(int);
				bool is_unset(void) const;
				int number_of_receivers(void) const;
			private:
				int numrec_;
		};
	
		typedef priority_list<ptr<untyped_interest>,untyped_interest::interest_priority> recipients_list;
		recipients_list   recipients;
		thread_mutex      recipients_lock;
};


template<class M>
class message : virtual public group_identifiers {
	public:
		virtual ~message() {};
		virtual untyped_message::send_result send(void);
		static  untyped_message static_part;
	protected:
		bool    send_to_interests(bool id_registered, int& handled_by);

};

template<class M>
class interest_base : virtual public untyped_interest {
	public:
		virtual void unregister(void);
		virtual untyped_interest::callresult docall(message<M>& m)=0;
		
};

template<class T, class M>
class interest : virtual public interest_base<M> {
	public:
		static ptr<interest<T,M> > create(T* t, untyped_interest::callresult (T::*f)(M&), 
													 untyped_interest::interest_priority);
		virtual untyped_interest::callresult docall(message<M>& m);
	protected:
		interest(T* t, untyped_interest::callresult (T::*f)(M&));
		T* oil_;
		untyped_interest::callresult (T::*oil_function_)(M&);
};



/* Implementations follow */
/* message */

template<class M>
untyped_message::send_result message<M>::send(void)
	{
	ids_mutex.lock();
	if(ids.size()==0)
		ids.insert(midentifier());  // if we have no group number yet, we take one.
	ids_mutex.unlock();

	if(static_part.recipients.size()==0) {
		throw messagebase_error(messagebase_error::no_recipients, typeid(this).name(),"");
		}

	int handled_by=0;
	
	if(send_to_interests(true,handled_by))   // send to all interests which have registered the group id
		send_to_interests(false,handled_by);  // ditto, to those which have not, unless exit_loop prevents
	                                         //    us from doing that.
	if(handled_by==0) {
		throw messagebase_error(messagebase_error::no_handling_recipients, typeid(this).name(),"");
		}
	// FIXME: interests that only get notified when a given message has been delivered
	// successfully are probably useful too. 
	return untyped_message::send_result(handled_by);
	}
	
// returns false if sending resulted in exit_loop return status.
// Only call this when the ids_mutex is locked!
template<class M>
bool message<M>::send_to_interests(bool id_registered, int& handled_by)
	{
	untyped_message::recipients_list::iterator i;
	ptr<untyped_interest > p;

	static_part.recipients_lock.lock();
	bool iid_contains_mid_true, mid_contains_iid_true, ids_allow_accept;
	bool retstat=true;

	for(i=static_part.recipients.begin(); i!=static_part.recipients.end(); ++i) {
		(*i)->ids_mutex.lock();
		ids_allow_accept=false;
		if((*i)->ids.size()==0) {
			if(id_registered) {
				(*i)->ids_mutex.unlock();
				continue;
				}
			}
		else {
			iid_contains_mid_true=includes((*i)->ids.begin(), (*i)->ids.end(),
													 ids.begin(), ids.end());
			mid_contains_iid_true=includes(ids.begin(), ids.end(),
													 (*i)->ids.begin(), (*i)->ids.end());
			if((*i)->acceptance==untyped_interest::iid_contains_mid 
				&& iid_contains_mid_true) ids_allow_accept=true;
			if((*i)->acceptance==untyped_interest::mid_contains_iid 
				&& mid_contains_iid_true) ids_allow_accept=true;
			if((*i)->acceptance==untyped_interest::mid_equals_iid
				&& iid_contains_mid_true && mid_contains_iid_true) ids_allow_accept=true;

			if((ids_allow_accept && !id_registered) || (!ids_allow_accept && id_registered)) {
				(*i)->ids_mutex.unlock();
				continue;
				}
			}
		if((*i)->acceptance!=untyped_interest::ignore_ids && !ids_allow_accept) {
			(*i)->ids_mutex.unlock();
			continue;
			}
		p=*i;
		assert((&(*p))!=0);
		(*i)->ids_mutex.unlock();
		static_part.recipients_lock.unlock();
		ptr<interest_base<M> > pp;
		pp.dynamiccast(p);
		assert(pp!=0);
		untyped_interest::callresult result=pp->docall(*((M*)this));
		static_part.recipients_lock.lock();
		if(static_part.handle_docall_result(result,handled_by)==false) {
			retstat=false;
			break;
			}
		}
	static_part.recipients_lock.unlock();
	return retstat;
	}


/* interest */

template<class M>
void interest_base<M>::unregister(void)
	{
	message<M>::static_part.unregister_interest((untyped_interest *)(this));
	}


template<class T,class M>
 ptr<interest<T,M> > interest<T,M>::create(T* t, untyped_interest::callresult (T::*f)(M&), untyped_interest::interest_priority priority){
	ptr<interest<T,M> > p=new interest<T,M>(t,f);
	message<M>::static_part.register_interest(priority,p);
	return p;
	}

template<class T, class M>
interest<T,M>::interest(T* t, untyped_interest::callresult (T::*f)(M&)) 
	{
	oil_=t;
	oil_function_=f;
	}

template<class T, class M>
untyped_interest::callresult interest<T, M>::docall(message<M>& m)
	{
	return (oil_->*oil_function_)((M&)m);
	}


#endif
