// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Mnemonic message object base code.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <mnemonic/messagebase.hh>
#include <typeinfo>
#include <algorithm>

#define DEBUG_CLASS "core"
#include <mnemonic/debug.hh>


messagebase_error::messagebase_error(messagebase_error::status_type st, const char *str) :
	logic_error(str)
	{
	stat_=st;
	local_what_ = string("[messagebase_error]: ") + message_type() + ": " +
		status_as_text() + " (" + str + ")";
	}

messagebase_error::messagebase_error(messagebase_error::status_type st, const char *mt, const char *str) :
	logic_error(str)
	{
	messagetype_=mt;
	stat_=st;
	local_what_ = string("[messagebase_error]: ") + message_type() + ": " +
		status_as_text() + " (" + str + ")";
	}

messagebase_error::status_type messagebase_error::status(void) const
	{
	return stat_;
	}

string messagebase_error::status_as_text(void) const
	{
	switch(stat_) {
		case no_recipients:
			return "no recipients for message";
		case no_handling_recipients:
			return "none of the recipients handled the message";
		default:
			return "ERROR: unknown exception, please report to bugs@mnemonic.browser.org";
		}
	}

string messagebase_error::message_type(void) const
	{
	return messagetype_;
	}

const char* messagebase_error::what() const
	{
	return local_what_.c_str();
	}

ostream& operator<<(ostream& str, const messagebase_error& ex)
	{
	str << "[messagebase_error]:" << endl
		  << "  status:  " << ex.status_as_text() << endl
		  << "  message: " << ex.message_type() << endl
		  << "  info:    " << ex.what();
	return str;
	}

untyped_message::untyped_message()
	{
	}

untyped_message::~untyped_message()
	{
//	debug(this<<" untyped message destructed"); 
/*	debug(this<<"message unregistered");
	untyped_message::recipients_list::iterator i;
	recipients_lock.lock();
	for(i=recipients.begin(); i!=recipients.end(); i++) {
		debug((*i).count()<<"[");
		recipients.erase(i);
		debug("]");
		}
	recipients_lock.unlock(); */
	}

bool untyped_message::send_result::is_unset(void) const
	{
	if(numrec_==-1) return true;
	else return false;
	}

untyped_message::send_result::send_result(void)
	{
	numrec_=-1;
	}

untyped_message::send_result::send_result(int num)
	{
	numrec_=num;
	}

int untyped_message::send_result::number_of_receivers(void) const
	{
	if(numrec_==-1) return 0;
	else return numrec_;
	}

untyped_interest::untyped_interest(void)
	{
	acceptance=ignore_ids;
	}

void untyped_interest::unregister(void)
	{
	}

untyped_interest::~untyped_interest()
	{
	}

bool group_identifiers::add_group_id(const midentifier& theid)
	{
	ids_mutex.lock();
	if(ids.count(theid)==1) {
		ids_mutex.unlock();
		return false;
		}
	ids.insert(theid);
	ids_mutex.unlock();
	return true;
	}

bool group_identifiers::add_group_ids(const group_identifiers& theid)
	{
	ids_mutex.lock();
	iterator it=theid.ids.begin();
	while(it!=theid.ids.end()) {
		ids.insert(*it);
		++it;
		}
	ids_mutex.unlock();
	return true;
	}

bool group_identifiers::remove_group_id(const midentifier& theid)
	{
	ids_mutex.lock();
	if(ids.count(theid)==0) {
		ids_mutex.unlock();
		return false;
	}		
	ids.erase(theid);
	ids_mutex.unlock();
	return true;
	}

bool group_identifiers::remove_group_ids(const group_identifiers& theid)
	{
	ids_mutex.lock();
	iterator it=theid.ids.begin();
	while(it!=theid.ids.end()) {
		if(ids.count((*it))!=0)
			ids.erase(*it);
		++it;
		}
	ids_mutex.unlock();
	return true;
	}

void group_identifiers::set_group_ids(const group_identifiers& other_ids)
	{
	// FIXME: do we have to lock the ids of other_ids too?
	ids_mutex.lock();
	ids=other_ids.ids;
	ids_mutex.unlock();
	}

bool group_identifiers::operator<(const group_identifiers& other) const
	{
	// FIXME: Hmm, this one is locked too long, maybe.
//	debug("operator<");
	ids_mutex.lock();
	set<midentifier, less<midentifier> >::const_iterator it1=ids.begin();
	set<midentifier, less<midentifier> >::const_iterator it2=other.ids.begin();
	while(it1!=ids.end() && it2!=other.ids.end()) {
		if((*it1)<(*it2)) {
			ids_mutex.unlock();
			return true;
			}
		++it1;
		++it2;
		}
	ids_mutex.unlock();
//	debug("done");
	return false;
	}

ostream& operator<<(ostream& str, const group_identifiers& gid)
	{
	// FIXME: no idea, but you cannot lock this here.
//	gid.ids_mutex.lock();
	set<midentifier, less<midentifier> >::const_iterator it=gid.ids.begin();
	while(it!=gid.ids.end()) { 
		str << (*it) << " ";
		++it;
		}
//	gid.ids_mutex.unlock();
	return str;
	}

void untyped_message::register_interest(untyped_interest::interest_priority priority, 
													 ptr<untyped_interest> intr)
	{
	recipients_lock.lock();
// unfortunately, this does not work:
//	debug("interest in " << typeid((*intr)).name() << " registered");
//	debug(this<<"message registered");
	recipients.insert(priority,intr);
	recipients_lock.unlock();
	}

void untyped_message::unregister_interest(untyped_interest *rm)
	{
//	debug(this<<"message unregistered");
	untyped_message::recipients_list::iterator i;
	recipients_lock.lock();
	for(i=recipients.begin(); i!=recipients.end(); ++i) {
		if(&(*(*i)) == rm) {
			recipients.erase(i);
			}
		}
	recipients_lock.unlock();
	}

// return false when the message should not be sent to any further recipients.
bool untyped_message::handle_docall_result(untyped_interest::callresult result, int& reccount) const
	{
	switch(result) {
		case untyped_interest::call_next: 
			++reccount;
//			cerr << "message should go to next recipient" << endl;
			return true;
		case untyped_interest::pass_on:
//			cerr << "message should be passed on to next recipient" << endl;
			return true;
		case untyped_interest::refused:
//			cerr << "message was refused, send to next recipient" << endl;
			return true;
		case untyped_interest::exit_loop:
//			cerr << "do not send message to any further interested recipients" << endl;
			++reccount;
			return false;
		default:
			debug("throw");
			throw logic_error("illegal return value for docall");
		}
	return false;
	}


counter::counter(void)
	{
	count_=1;
	}

counter::~counter()
	{
	}



template untyped_message::recipients_list;
template list<pair<untyped_interest::interest_priority, ptr<untyped_interest> > >;
template list<untyped_message::recipients_list::iterator *>;
template set<midentifier, less<midentifier> >;
typedef set<midentifier, less<midentifier> >::const_iterator idset_it;
template bool includes(idset_it, idset_it, idset_it, idset_it);
