// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  C++ classes to represent a W3C DOM CSS.

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef css_hh__
#define css_hh__

#include <string>
#include <dom/dom.hh>
#include <url/url_string.hh>
#include <vector.h>
#include <multimap.h>
#include <map.h>
#include <stdint.h>

#include <mnemonic/ptr.hh>

namespace CSS {

using namespace DOM;

class stylesheet;
class numberproperty;
class colorproperty;
class urlproperty;
class stringproperty; 
class rule;
class selector;
class declaration;

class declaration {
	public:
		virtual ~declaration(); 

		enum number_properties { margin_top=1,
										 margin_bottom=2,
										 margin_left=3,
										 margin_right=4,
										 max_height=5, 
										 max_width=6,
										 min_height=7,
										 min_width=8,
										 border_spacing=9,
										 font_size=10,
										 border_width=11
		};
		enum color_properties  { background_color=101,
										 border_color=102,
										 border_top_color=103,
										 border_bottom_color=104,
										 border_left_color=105,
										 border_right_color=106,
										 color=107,
										 outline_color=108,
										 text_shadow_color=109 
		};
		enum string_properties { font_family=150 
		};
		enum url_properties    { background_image=201,
										 content_uri=201,
										 cursor_uri=202,
										 list_style_image=203 
		};
		enum other_properties  { background_attachment=204,
		                         display=205,
										 font_style=206,
										 font_weight=207,
										 text_align=208,
										 border_style=209,
										 border_top_style=210,
										 border_bottom_style=211,
										 border_left_style=212,
										 border_right_style=213,
		};

		class numberproperty {
			public:
				enum css_unit { EM, EX, PX, IN, CM, MM, PT, PC, PERC, 
									 DEG, GRAD, RAD, HZ, KHZ, SEC, MILLISEC };

				numberproperty(void);
				numberproperty(const string&);
				numberproperty(double, css_unit);
				
				double   value(void) const;
				css_unit unit(void) const;
				string   str(void) const;
			private:
				double   value_;
				css_unit unit_;
		};
		
		class urlproperty {
			public:
				urlproperty(const url_string& );
				url_string url(void) const;
			protected:
				url_string theurl_;
		};
		
		class colorproperty {
			public:
				enum specialcolor { transparent, inherit };
				
				colorproperty(uint8_t, uint8_t, uint8_t);
				colorproperty(float, float, float);
				colorproperty(specialcolor);
				colorproperty(const DOMString&);
				
				DOMString    colorName(void) const;
				float        redPercentage(void) const;
				float        greenPercentage(void) const;
				float        bluePercentage(void) const;
				int16_t      redValue(void) const;
				int16_t      greenValue(void) const;
				int16_t      blueValue(void) const;
			protected:
				DOMString    name_;
				int16_t      red_, green_, blue_;
		};
		
		class stringproperty {
			public:
				stringproperty(const DOMString&);
				
				DOMString   value(void) const;
			protected:
				DOMString   text_;
		};
		
		enum background_attachment_value { scroll=1, fixed };
		enum background_repeat_value     { repeat=1, repeat_x, repeat_y, no_repeat };
		enum border_collapse_value       { collapse=1, separate };
		enum border_style_value          { border_style_none=1, hidden, dotted, dashed, solid,
													  border_double, groove, ridge, inset, outset };
		enum caption_side_value          { caption_top=1, caption_bottom, caption_left, caption_right };
		enum clear_value                 { clear_none=1, clear_left, clear_right, clear_both };
		enum display_value               { display_inline=1, block, list_item,
													  run_in, compact, marker,
													  table, inline_table, table_row_group,
													  table_column, table_column_group,
													  table_header_group, table_footer_group,
													  table_row, table_cell, table_caption, display_none,
													  display_empty };
		// display_empty means `we don't know this element, so just insert the content' (eg. <foo>).
		// display_none  means `do not display this subtree'
		enum font_style_value            { font_normal, font_italic, font_oblique };
		enum font_weight_value           { weight_normal, weight_bold };
		enum text_align_value            { align_left, align_right, align_center, align_justify };

		virtual numberproperty             *get_number_property(number_properties) const;
		virtual colorproperty              *get_color_property(color_properties) const;

		virtual numberproperty             *get_border_spacing(void) const;
		virtual display_value               get_display(void) const;
		virtual colorproperty              *get_color(void) const;
		virtual stringproperty             *get_font_family(void) const;
		virtual font_style_value            get_font_style(void) const;
		virtual font_weight_value           get_font_weight(void) const;
		virtual text_align_value            get_text_align(void) const;
		virtual numberproperty             *get_font_size(void) const;
		virtual border_style_value          get_border_top_style(void) const;
		virtual border_style_value          get_border_bottom_style(void) const;
		virtual border_style_value          get_border_left_style(void) const;
		virtual border_style_value          get_border_right_style(void) const;

		
		bool inherit(url_properties);
		bool inherit(color_properties);
		bool inherit(number_properties);
		bool inherit(string_properties);
		bool inherit(other_properties);

		void set_number_property(number_properties, numberproperty *);
		void set_color_property(color_properties, colorproperty *);

		void set_background_attachment(background_attachment_value);
		void set_background_position(numberproperty, numberproperty);
		enum background_position_shorthand { top, 
														 top_left,
														 top_center,
														 top_right,
														 left_center,
														 center,
														 right_center,
														 bottom_left,
														 bottom_center,
														 bottom_right };
		void set_background_position(numberproperty);
		void set_background_position(background_position_shorthand);
		void set_background_repeat(background_repeat_value);

		void set_border_collapse(border_collapse_value);
		void set_border_color(colorproperty);
		void set_border_spacing(numberproperty *);
		void set_border_top_color(colorproperty);
		void set_border_bottom_color(colorproperty);
		void set_border_left_color(colorproperty);
		void set_border_right_color(colorproperty);
		void set_border_style(border_style_value);
		void set_border_top_style(border_style_value);
		void set_border_bottom_style(border_style_value);
		void set_border_left_style(border_style_value);
		void set_border_right_style(border_style_value);
		void set_border_top_width(numberproperty);
		void set_border_bottom_width(numberproperty);
		void set_border_left_width(numberproperty);
		void set_border_right_width(numberproperty);

		void set_top(numberproperty);
		void set_bottom(numberproperty);
		void set_left(numberproperty);
		void set_right(numberproperty);
		void set_caption_side(caption_side_value);
		void set_clear(clear_value);
		void set_display(display_value);
		void set_color(colorproperty *);
		void set_font_family(stringproperty *prop);
		void set_font_style(font_style_value);
		void set_font_weight(font_weight_value);
		void set_text_align(text_align_value);
		void set_font_size(numberproperty *);

		bool set_inherit(url_properties);
		bool set_inherit(color_properties);
		bool set_inherit(number_properties);
		bool set_inherit(other_properties);

//	protected:
		union propertyunion {
				void   *pointer;
				uint8_t  number;
		}; 
		map<uint8_t, propertyunion>   properties_; // 0 means value `inherit' in declaration
};

class selector {
	public:
		class selector_element {
			public:
				selector_element(const DOMString&);

				enum pseudotypes  { empty_pseudotype,
                                first_child,
										  link, visited, 
										  hover, active, focus,
										  lang,
										  first_line, first_letter,
										  before, after,
										  page,
										  font_face };

				class attribute_element {
					public:
						DOMString attribute;
						enum select_on { is_set, equals, contains_only_one,
											  contains_as_first };
						
						select_on value_interpret;
						DOMString value;
				};

				DOMString                 name;     // "*" means arbitrary
				vector<DOMString>         classes;
				vector<pseudotypes>       pseudoclasses;
				vector<DOMString>         ids;
				vector<attribute_element> attributes;
		};
		bool applies(const Node *, selector_element::pseudotypes pt) const;

		enum selector_relation { descendant, child, preceded };
		
		vector<selector_element>  elements_;
		vector<selector_relation> relations_;
};

class stylesheet {
	public:
		// FIXME: parent default =0 
		stylesheet(ptr<DOM::Document> doc, stylesheet *parent);
		~stylesheet();

		bool               disabled(void) const;
		void               set_disabled(bool);
		Node              *owning_node(void) const;
		stylesheet        *parent_stylesheet(void) const;
		DOMString          href(void) const;
		DOMString          title(void) const;
		DOMString          media(void) const;
		ptr<DOM::Document> document(void) const;
							
		const rule        *add_rule(const rule *);

		enum whichstage { specified, computed };

		class node_style : public declaration {
			public:
				void copy_from(const declaration *);
		};

		node_style *style(const DOM::Node *);
		node_style *style(const DOM::Node *, selector::selector_element::pseudotypes);

		typedef multimap<DOMString, const rule *>::iterator       iterator;
		typedef multimap<DOMString, const rule *>::const_iterator const_iterator;
		const_iterator begin(void)  const;
		const_iterator end(void)    const;
	private:
		DOMString    medium_;   // Ehm, mediA?
		DOMString    origin_;   // to store eg the url? (to avoid CSSImportRule)

		ptr<DOM::Document> doc_;
		stylesheet  *parent_;
		stylesheet  *first_child_, *last_child_;
		stylesheet  *prev_sib_,    *next_sib_;

		map<const DOM::Node *, node_style *> styleinfo_;
		multimap<DOMString, const rule *>    rules_;    // indexed on name of element to which rule applies

		void determine_properties_(const Node *nod, node_style *st, selector::selector_element::pseudotypes);
};


class rule {
	public:
		rule(selector *, declaration *);
		rule(selector *, declaration *, bool owndec);
		~rule();  // we own the two pointers

		stylesheet        *parent_stylesheet(void) const;
		rule              *parent_rule(void) const;
		const selector    *get_selector(void) const;
		const declaration *get_declaration(void) const;
	protected:
		rule           *parent_;
		stylesheet     *parentsheet_;
		selector       *selector_;
		declaration    *declaration_;
		bool            own_declaration_;  
      // Declarations may be used by multiple rules, but only
		// one should actually delete it after use.
};




}; //namespace


#endif
