// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  C++ classes to represent a W3C Document Object Model document.

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef dom_hh_
#define dom_hh_

#include <string>
#include <vector.h>
#include <map.h>
#include <stdexcept>

// TODO:
//       implement Range (that's fairly easy, the complicated bit is how to relate it to display)

namespace DOM {

class DocumentFragment;
class NamespaceDecl;
class Document;
class Node;
class NamedNodeMap;
class CharacterData;
class Attr;
class Element;
class Text;
class Comment;
class CDATASection;
class DocumentType;
class Notation;
class Entity;
class EntityReference;
class ProcessingInstruction;


typedef string DOMString;
typedef bool   boolean;


// DOM 1
class DOMException { 
	public:
		enum ExceptionCode { INDEX_SIZE_ERR=1, DOMSTRING_SIZE_ERR, 
									HIERARCHY_REQUEST_ERR, WRONG_DOCUMENT_ERR, 
									INVALID_CHARACTER_ERR, NO_DATA_ALLOWED_ERR, 
									NO_MODIFICATION_ALLOWED_ERR, NOT_FOUND_ERR, 
									NOT_SUPPORTED_ERR, INUSE_ATTRIBUTE_ERR };

		DOMException(ExceptionCode);

		ExceptionCode code;
};


// DOM 1
class DOMImplementation {
	public:
		boolean   hasFeature(const DOMString& feature, const DOMString& version);
};


// DOM 2 
class NodeFilter {
	public:
		NodeFilter(int whatToShow);
		virtual boolean acceptNode(Node *) const;

		// DOM 2 is unclear about where this should go and prefers it outside
		// of NodeFilter. Idiots.
		static const int  TW_DEFAULT       = 0x0022;
		static const int  TW_ELEMENT       = 0x0002;
		static const int  TW_PI            = 0x0008;
		static const int  TW_COMMENT       = 0x0010;
		static const int  TW_TEXT          = 0x0020;
		static const int  TW_ENTITYREF     = 0x0040;
		static const int  TW_ALL           = 0xFFFF;
	private:
		int  whattoshow_;
};


// nonstandard additional class
class ElementNodeFilter : public NodeFilter {
	public:
		ElementNodeFilter(int whatToShow, const DOMString& tagname);
		virtual ~ElementNodeFilter();

		virtual boolean acceptNode(Node *) const;
	private:
		DOMString tagname_;
};


// DOM 1
class Node {
	public:
		Node(void);
		virtual ~Node();

		enum NodeType { ELEMENT_NODE, ATTRIBUTE_NODE, TEXT_NODE, 
							 CDATA_SECTION_NODE, ENTITY_REFERENCE_NODE, 
							 ENTITY_NODE, PROCESSING_INSTRUCTION_NODE,
							 COMMENT_NODE, DOCUMENT_NODE, DOCUMENT_TYPE_NODE,
							 DOCUMENT_FRAGMENT_NODE, NOTATION_NODE };

      // DOM 2
		class NodeIterator {
			public:
				NodeIterator(Node* startnode, bool deep);
				NodeIterator(Node* startnode, bool deep, bool siblings_at_toplevel);
				
				Node *thisnode(void);
				Node *nextnode(void);
				Node *prevnode(void);

				// non-standard additions
				int           levels_changed(void) const;
				void          skip_current_children(void);
				NodeIterator& operator++(void);
				NodeIterator& operator--(void);
				Node         *operator*(void);
			protected:
				virtual bool increment_(void);
				virtual bool decrement_(void);
				bool                         deep_;
				bool                         siblings_at_toplevel_;
				Node                        *currentnode_;
				Node                        *startnode_;
				int                          levels_changed_;
				bool                         skip_current_children_;
		};

		virtual NodeType           nodeType(void) const=0;
		virtual const DOMString&   nodeName(void) const=0;
		virtual const DOMString&   nodeValue(void) const;
		const NamespaceDecl       *nodeNamespace(void) const; // the NamespaceDecl is owned by the Document,
		                                                // not the element in which the declaration is made.

		virtual NamedNodeMap *attributes(void) const;
		virtual void          set_nodeValue(const DOMString& ) throw(DOMException);

		Node                 *parentNode(void) const;
		Node                 *firstChild(void) const;
		Node                 *lastChild(void) const;
		Node                 *previousSibling(void) const;
		Node                 *nextSibling(void) const;
		Document             *ownerDocument(void) const;

		Node                 *insertBefore(Node *newChild, NodeIterator *refChild);
		Node                 *replaceChild(Node *newChild, Node oldChild);
		Node                 *removeChild(Node *oldChild);
		virtual Node         *appendChild(Node *newChild);
		boolean               hasChildNodes(void) const;
		Node                  cloneNode(boolean deep);
	protected:
//		Document       *owner_;  // waste of space
		Node           *parent_;
		Node           *first_child_, *last_child_;
		Node           *prev_sib_,    *next_sib_;
};


// DOM 1
/* 
   Gratuitous grumbling: NodeList's only advantage over an iterator is that
	even users who can't understand "next" ought to be able to deal with "+1".
	Whether that's a community worth catering to, rather than educating or
	addressing via a higher-level tool, is something we'll never agree on.
	______________________________________
	Joe Kesselman  / IBM Research

	Ergo: we do not support NodeList.
*/


// non-standard (and I don't like DOM2)
class NamespaceDecl {
	public:
		NamespaceDecl(const DOMString& prefix, const DOMString& uri);
		DOMString uri(void) const;
		DOMString prefix(void) const;
	private:
		DOMString prefix_;
		DOMString uri_;
};

// DOM 1
class Document : public Node {
	public:
		Document(const string& url_spec="", const string& content_type="", const DocumentType * dt=0);
		virtual ~Document();
		
		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;

		// non-standard. Document owns doctype and default namespace declaration.
		void                   set_doctype(const DocumentType *doctype);
		void                   set_default_namespace(const NamespaceDecl *ns);

		const DocumentType    *doctype(void) const;
		string                 url(void) const;
		string                 content_type(void) const;
		DOMImplementation     *implementation(void) const;
		Element               *documentElement(void) const;

		Element               *createElement(const DOMString& tagName) throw(DOMException);
		DocumentFragment      *createDocumentFragment(void);
		Text                  *createTextNode(const DOMString& data);
		Comment               *createComment(const DOMString& data);
		CDATASection          *createCDATASection(const DOMString& data) throw(DOMException);
		ProcessingInstruction *createProcessingInstruction(const DOMString& target, const DOMString& data);
		Attr                  *createAttribute(const DOMString& name);
		EntityReference       *createEntityReference(const DOMString& name);

		// Document owns the NamespaceDecls.
		void                          setNamespace(Node *, const NamespaceDecl *);
		pair<Node *, NamespaceDecl *> getNamespace(const DOMString& prefix);
		const NamespaceDecl          *nodeNamespace(const Node *) const;

	private:
		const DocumentType      *doctype_;
		string                   url_, content_type_;
		DOMImplementation       *implementation_;

		// We keep track of were a particular namespace declaration was made
		// We could keep a list of pointers to namespace declarations in each node, but
		// that would use much more memory as most nodes will _not_ have such a declaration.
		mutable map<const Node *, const NamespaceDecl *, less<const Node *> >  namespacedecls_;
};


// DOM 1
class DocumentFragment : public Node {
	public:
		DocumentFragment(void);
		virtual ~DocumentFragment();

		virtual Node::NodeType    nodeType(void) const;
		virtual const DOMString&  nodeName(void) const;

		friend Document;
};


// DOM 1
class NamedNodeMap {
	public:
		Node           *getNamedItem(const DOMString& name);
		Node           *setNamedItem(Node *);
		Node           *removeNamedItem(const DOMString& name);
		Node           *item(unsigned long index);
		
		unsigned long   length(void) const;
};


// DOM 1
class CharacterData : public Node {
	public:
		CharacterData(const DOMString& data);
		virtual ~CharacterData();

		unsigned long length(void) const;
		DOMString     data(void) const;
		void          set_data(const DOMString& );

		DOMString     substringData(unsigned long offset, unsigned long count);
		void          appendData(const DOMString& arg);
		void          insertData(const unsigned long offset, const DOMString& arg);
		void          deleteData(unsigned long offset, unsigned long count);
		void          replaceData(unsigned long offset, unsigned long count, const DOMString& arg);

		friend Document;
	protected:
		string  data_; 
};


// DOM 1
class Attr : public Node {
	public:
		Attr(const DOMString& name);
		Attr(const DOMString& name, const DOMString& value);
		virtual ~Attr();

		virtual Node::NodeType    nodeType(void) const;
		virtual const DOMString&  nodeName(void) const;
		virtual const DOMString&  nodeValue(void) const;

		const DOMString&   name(void) const;
		boolean            specified(void) const;
		const DOMString&   value(void) const;
		void               set_value(const DOMString& );

		friend Document;
	private:
		DOMString attrname_;
		DOMString attrvalue_;
		boolean   specified_;
};


// DOM 1
class Element : public Node {
	public: 
		Element(const DOMString& tagName);
		virtual ~Element();

		virtual Node::NodeType    nodeType(void) const;
		virtual const DOMString&  nodeName(void) const;
		NamedNodeMap          *attributes(void) const;
		// non-standard addition
		Attr                  *firstAttribute(void) const;

		const DOMString&  tagName(void) const;
		const DOMString&  getAttribute(const DOMString& name) const;
		void              setAttribute(const DOMString& name, const DOMString& value);
		void              removeAttribute(const DOMString& name);
		Attr             *getAttributeNode(const DOMString& name) const;
		Attr             *setAttributeNode(const Attr& newAttr);
		Attr             *removeAttributeNode(const Attr& oldAttr);
		void              normalize(void);

		friend Document;
	private:
		DOMString      tagname_;
		Attr          *attributes_;  // first attribute, all other attributes are children.
};


// DOM 1
class Text : public CharacterData {
	public:
		Text(const DOMString& data);
		virtual ~Text();

		virtual Node::NodeType nodeType(void) const;
		virtual const DOMString& nodeName(void) const;
		virtual const DOMString& nodeValue(void) const;

		Text              splitText(unsigned long offset);

		friend Document;
};


// DOM 1
class Comment : public CharacterData {
	public:
		Comment(const DOMString& data);
		virtual ~Comment();

		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;
		virtual const DOMString& nodeValue(void) const;

		friend Document;
};


// DOM 1
class CDATASection : public Text {
	public:
		CDATASection(const DOMString& data);
		virtual ~CDATASection();

		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;
		virtual const DOMString& nodeValue(void) const;

		friend Document;
};


// DOM 1
class DocumentType : public Node {
	public:
		DocumentType(const DOMString& doctype);
		virtual ~DocumentType();

		virtual Node::NodeType    nodeType(void) const;
		virtual const DOMString&  nodeName(void) const;

		const DOMString& name(void) const;
		NamedNodeMap    *entities(void) const;
		NamedNodeMap    *notations(void) const;
	private:
		DOMString doctypename_;
};


// DOM 1
class Notation : public Node {
	public:
		Notation(const DOMString& notation);
		virtual ~Notation();

		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;

		DOMString      publicID(void) const;
		DOMString      systemID(void) const;

		friend Document;
	private:
		DOMString notationname_;
};


// DOM 1
class Entity : public Node {
	public:
		Entity(const DOMString &entityname);
		virtual ~Entity();

		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;

		DOMString      publicID(void) const;
		DOMString      systemID(void) const;
		DOMString      notationName(void) const;

		friend Document;
	private:
		DOMString entityname_;
};


// DOM 1
class EntityReference : public Node {
	public:
		EntityReference(const DOMString& name);
		virtual ~EntityReference();
		
		virtual Node::NodeType   nodeType(void) const;
		virtual const DOMString& nodeName(void) const;

		friend Document;
	private:
		DOMString entityref_;
};


// DOM 1
class ProcessingInstruction : public Node {
	public:
		ProcessingInstruction(const DOMString& target, const DOMString& data);
		virtual ~ProcessingInstruction();

		virtual Node::NodeType nodeType(void) const;
		virtual const DOMString& nodeName(void) const;
		virtual const DOMString& nodeValue(void) const;

		DOMString     target(void) const;
		DOMString     data(void) const;
		void          set_data(const DOMString& );

		friend Document;
	private:
		string target_, data_;
};

}; // namespace

#endif
