// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  C++ classes to represent the stylesheet part of DOM level 2.

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <dom/css.hh>
#include <algo.h>
#include <strstream.h>

#define DEBUG_CLASS "CSS"
#include <mnemonic/debug.hh>

using namespace CSS;

stylesheet::stylesheet(ptr<DOM::Document> doc, stylesheet *theparent)
	: doc_(doc), parent_(theparent)
	{
	}

stylesheet::~stylesheet()
	{
	iterator it=rules_.begin();
	while(it!=rules_.end()) {
		delete (*it).second;
		++it;
		}
	debug("CSS::stylesheet deleted");
	}

const rule *stylesheet::add_rule(const rule *rul)
	{
	assert(rul->get_selector()->elements_.size()>0);
	DOMString nam=rul->get_selector()->elements_[0].name;
	rules_.insert(pair<DOMString, const rule *>(nam,rul));
	return rul;
	}

ptr<DOM::Document> stylesheet::document(void) const
	{
	return doc_;
	}

stylesheet::node_style *stylesheet::style(const DOM::Node *nod)
	{
	map<const DOM::Node *, node_style *>::iterator it=styleinfo_.find(nod);
	if(it!=styleinfo_.end()) return (*it).second;
	else {
		node_style *st=new node_style;
		determine_properties_(nod,st,selector::selector_element::empty_pseudotype);
		styleinfo_[nod]=st;
		return st;
		}
	}

stylesheet::node_style *stylesheet::style(const DOM::Node *nod, selector::selector_element::pseudotypes pt)
	{
	map<const DOM::Node *, node_style *>::iterator it=styleinfo_.find(nod);
	if(it!=styleinfo_.end()) return (*it).second;
	else {
		node_style *st=new node_style;
		determine_properties_(nod,st,pt);
		styleinfo_[nod]=st;
		return st;
		}
	}

stylesheet::const_iterator stylesheet::begin(void) const
	{
	return rules_.begin();
	}

stylesheet::const_iterator stylesheet::end(void) const
	{
	return rules_.end();
	}

void stylesheet::determine_properties_(const Node *nod, node_style *st, 
													selector::selector_element::pseudotypes pt)
	{
	pair<iterator, iterator> itp=rules_.equal_range(nod->nodeName());

	while(itp.first!=itp.second) {
		assert((*(itp.first)).second!=0);
		const selector *sel=(*(itp.first)).second->get_selector();
		assert(sel!=0);

		if(sel->applies(nod, pt)) {
			st->copy_from((*(itp.first)).second->get_declaration());
			}
		++(itp.first);
		}
	}

void stylesheet::node_style::copy_from(const declaration *prop)
	{
	map<uint8_t, propertyunion>::const_iterator it=prop->properties_.begin();
	while(it!=prop->properties_.end()) {
		properties_[(*it).first]=(*it).second;
//		cout << (int)((*it).first) << ", "  << (int)((*it).second.number) << endl;
		++it;
		}
	}

declaration::~declaration()
	{
	map<uint8_t, propertyunion>::iterator it=properties_.begin();
	while(it!=properties_.end()) {
		if((*it).first<static_cast<uint8_t>(background_attachment))
			delete (*it).second.pointer;
		++it;
		}
	}

declaration::numberproperty *declaration::get_number_property(number_properties np) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(np);
	if(it==properties_.end()) return 0;
	else return (declaration::numberproperty *)((*it).second.pointer);
	}

declaration::colorproperty *declaration::get_color_property(color_properties np) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(np);
	if(it==properties_.end()) return 0;
	else return (declaration::colorproperty *)((*it).second.pointer);
	}

declaration::numberproperty *declaration::get_border_spacing(void) const
	{
	// FIXME
	return 0; 
	}

declaration::colorproperty *declaration::get_color(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(color);
	if(it==properties_.end()) return 0;
	else return (colorproperty *)((*it).second.pointer);
	}

declaration::display_value declaration::get_display(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(display);
	if(it==properties_.end()) return declaration::display_empty;
	else return (declaration::display_value)((*it).second.number);
	}

declaration::stringproperty *declaration::get_font_family(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(font_family);
	if(it==properties_.end()) return 0;
	else return (declaration::stringproperty *)((*it).second.pointer);
	}

declaration::font_style_value declaration::get_font_style(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(font_style);
	if(it==properties_.end()) return declaration::font_normal;
	else return (declaration::font_style_value)((*it).second.number);
	}

declaration::font_weight_value declaration::get_font_weight(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(font_weight);
	if(it==properties_.end()) return declaration::weight_normal;
	else return (declaration::font_weight_value)((*it).second.number);
	}

declaration::text_align_value declaration::get_text_align(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(text_align);
	if(it==properties_.end()) return declaration::align_left;
	else return (declaration::text_align_value)((*it).second.number);
	}

declaration::numberproperty *declaration::get_font_size(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(font_size);
	if(it==properties_.end()) return 0;
	else return (declaration::numberproperty *)((*it).second.pointer);
	}

declaration::border_style_value declaration::get_border_top_style(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(border_top_style);
	if(it==properties_.end()) return declaration::border_style_none;
	else return (declaration::border_style_value)((*it).second.number);
	}

declaration::border_style_value declaration::get_border_bottom_style(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(border_bottom_style);
	if(it==properties_.end()) return declaration::border_style_none;
	else return (declaration::border_style_value)((*it).second.number);
	}

declaration::border_style_value declaration::get_border_left_style(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(border_left_style);
	if(it==properties_.end()) return declaration::border_style_none;
	else return (declaration::border_style_value)((*it).second.number);
	}

declaration::border_style_value declaration::get_border_right_style(void) const
	{
	map<uint8_t, propertyunion>::const_iterator it=properties_.find(border_right_style);
	if(it==properties_.end()) return declaration::border_style_none;
	else return (declaration::border_style_value)((*it).second.number);
	}

void declaration::set_number_property(number_properties wh, declaration::numberproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[wh]=tmp;
	}

void declaration::set_color_property(color_properties wh, colorproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[wh]=tmp;
	}

void declaration::set_border_spacing(declaration::numberproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[border_spacing]=tmp;
	}

void declaration::set_border_style(border_style_value prop)
	{
	propertyunion tmp;
	tmp.number=prop;
	properties_[border_top_style]=tmp;
	properties_[border_bottom_style]=tmp;
	properties_[border_left_style]=tmp;
	properties_[border_right_style]=tmp;
	}

void declaration::set_color(colorproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[color]=tmp;
	}

void declaration::set_display(declaration::display_value prop)
	{
	propertyunion tmp;
	tmp.number=prop;
	properties_[display]=tmp;
	}

void declaration::set_font_family(declaration::stringproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[font_family]=tmp;
	}

void declaration::set_font_style(declaration::font_style_value prop)
	{
	propertyunion tmp;
	tmp.number=prop;
	properties_[font_style]=tmp;
	}

void declaration::set_font_weight(declaration::font_weight_value prop)
	{
	propertyunion tmp;
	tmp.number=prop;
	properties_[font_weight]=tmp;
	}

void declaration::set_text_align(declaration::text_align_value prop)
	{
	propertyunion tmp;
	tmp.number=prop;
	properties_[text_align]=tmp;
	}

void declaration::set_font_size(declaration::numberproperty *prop)
	{
	propertyunion tmp;
	tmp.pointer=prop;
	properties_[font_size]=tmp;
	}

rule::rule(selector *sel, declaration *dec)
	: selector_(sel), declaration_(dec), own_declaration_(true)
	{
	}

rule::rule(selector *sel, declaration *dec, bool od)
	: selector_(sel), declaration_(dec), own_declaration_(od)
	{
	}

rule::~rule()
	{
	if(selector_)                        delete selector_;
	if(declaration_ && own_declaration_) delete declaration_;
	}

const selector *rule::get_selector(void) const
	{
	return selector_;
	}

const declaration *rule::get_declaration(void) const
	{
	return declaration_;
	}

selector::selector_element::selector_element(const DOMString& nam)
	: name(nam)
	{
	}

bool selector::applies(const Node *nd, selector_element::pseudotypes pt) const
	{
	if(!nd) return false;

	// FIXME: this does not do nearly enough, but suffices for now.
	int i=elements_.size()-1;
	while(i>=0) {
		if(nd->nodeName()==elements_[i].name) {
			debug("checking " << elements_[i].name);
			if(pt!=selector_element::empty_pseudotype) {
				debug("pseudoclass");
				if(find(elements_[i].pseudoclasses.begin(),
						  elements_[i].pseudoclasses.end(),pt)!=elements_[i].pseudoclasses.end()) {
					debug("pseudoclass ok");
					return true;
					}
				}
			else {
				if(elements_[i].pseudoclasses.size()==0)
					return true;
				}
			}
		--i;
		}

	return false;
	}

declaration::numberproperty::numberproperty(void)
	: value_(0), unit_(IN)
	{
	}

declaration::numberproperty::numberproperty(const string& s)
	{
	value_=atof(s.c_str());
	string::size_type pos=s.find_first_not_of("-+0123456789. ");
	string un=s.substr(pos);
	transform(un.begin(),un.end(),un.begin(),tolower);
	if(un=="em")        unit_=EM;
	else if(un=="ex")   unit_=EX;
	else if(un=="px")   unit_=PX;
	else if(un=="in")   unit_=IN;
	else if(un=="cm")   unit_=CM;
	else if(un=="mm")   unit_=MM;
	else if(un=="pt")   unit_=PT;
	else if(un=="pc")   unit_=PC;
	else if(un=="%")    unit_=PERC;
	else if(un=="deg")  unit_=DEG;
	else if(un=="grad") unit_=GRAD;
	else if(un=="rad")  unit_=RAD;
	else if(un=="hz")   unit_=HZ;
	else if(un=="khz")  unit_=KHZ;
	else if(un=="sec")  unit_=SEC;
	else if(un=="msec") unit_=MILLISEC;
	else throw logic_error(string("unit ")+un+string(" is not a CSS unit"));
	}

declaration::numberproperty::numberproperty(double val, declaration::numberproperty::css_unit un)
	: value_(val), unit_(un)
	{
	}

double declaration::numberproperty::value(void) const
	{
	return value_;
	}

declaration::numberproperty::css_unit declaration::numberproperty::unit(void) const
	{
	return unit_;
	}

string declaration::numberproperty::str(void) const
	{
	strstream st;
	st << value_;
	switch(unit_) {
		case EM:        st << "em"; break;
		case EX:        st << "ex"; break;
		case PX:        st << "px"; break;
		case IN:        st << "in"; break;
		case CM:        st << "cm"; break;
		case MM:        st << "mm"; break;
		case PT:        st << "pt"; break;
		case PC:        st << "pc"; break;
		case PERC:      st << "%"; break;
		case DEG:       st << "deg"; break;
		case GRAD:      st << "grad"; break;
		case RAD:       st << "rad"; break;
		case HZ:        st << "hz"; break;
		case KHZ:       st << "khz"; break;
		case SEC:       st << "sec"; break;
		case MILLISEC:  st << "msec"; break;
		}
	st << ends;
	string ret=st.str();
	delete [] st.str();
	return ret;
	}

declaration::colorproperty::colorproperty(const DOMString& nam)
	: name_(nam)
	{
	}

DOMString declaration::colorproperty::colorName(void) const
	{
	return name_;
	}

declaration::stringproperty::stringproperty(const DOMString& str)
	: text_(str)
	{
	}

DOMString declaration::stringproperty::value(void) const
	{
	return text_;
	}
