// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  C++ classes to represent a W3C Document Object Model document.

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <dom/dom.hh>

#define DEBUG_CLASS "DOM"
#include <mnemonic/debug.hh>
using namespace DOM;


DOMException::DOMException(DOMException::ExceptionCode cd)
	{
	code=cd;
	}


//--------------------------------------------------------------

boolean DOMImplementation::hasFeature(const DOMString& feature, const DOMString& version)
	{
	if(feature=="HTML") {
		if(version=="" || version=="1.0")
			return true;
		}
	if(feature=="XML") {
		if(version=="" || version=="1.0")
			return true;
		}
	return false;
	}

//--------------------------------------------------------------

NodeFilter::NodeFilter(int whatToShow)
	: whattoshow_(whatToShow)
	{
	}

boolean NodeFilter::acceptNode(Node *nod) const
	{
	return true; // FIXME
	}


//--------------------------------------------------------------

ElementNodeFilter::ElementNodeFilter(int whatToShow, const DOMString& tagname)
	: NodeFilter(whatToShow), tagname_(tagname)
	{
	}

ElementNodeFilter::~ElementNodeFilter()
	{
	}

boolean ElementNodeFilter::acceptNode(Node *nod) const
	{
	return NodeFilter::acceptNode(nod); // FIXME
	}

//--------------------------------------------------------------

Node::Node(void)
	: parent_(0), first_child_(0), last_child_(0), prev_sib_(0), next_sib_(0)
	{
	}

Node::~Node(void)
	{
	Node *currentnode=first_child_;
	Node *prevnode;
	while(currentnode!=0) {
		prevnode=currentnode;
		currentnode=currentnode->next_sib_;
		delete prevnode;
		}
	}

void Node::set_nodeValue(const DOMString& val) throw(DOMException)
	{
	throw DOMException(DOMException::NO_MODIFICATION_ALLOWED_ERR);
	}

NamedNodeMap *Node::attributes(void) const
	{
	return 0;
	}

const DOMString& Node::nodeValue(void) const
	{
	static string tmp("");
	return tmp;
	}

const NamespaceDecl *Node::nodeNamespace(void) const
	{
	Document *doc=ownerDocument();
	assert(doc!=0);
	const NamespaceDecl *dec=doc->nodeNamespace(this);
	assert(dec!=0);
	return dec;
	}

Node *Node::parentNode(void) const
	{
	return parent_;
	}

Node *Node::firstChild(void) const
	{
	return first_child_;
	}

Node *Node::lastChild(void) const
	{
	return last_child_;
	}

Node *Node::appendChild(Node *newChild)
	{
	newChild->parent_=this;
	if(last_child_!=0)
		last_child_->next_sib_=newChild;
	else
		first_child_=newChild;
	newChild->prev_sib_=last_child_;
	last_child_=newChild;
	newChild->next_sib_=0;
	return newChild;
	}

Node *Node::previousSibling(void) const
	{
	return prev_sib_;
	}

Node *Node::nextSibling(void) const
	{
	return next_sib_;
	}

Document *Node::ownerDocument(void) const
	{
	Node *doc=parent_;
	if(doc!=0) {
		while(doc->parentNode()) 
			doc=doc->parentNode();
		}
	else doc=const_cast<Node *>(this);

	Document *truedoc=dynamic_cast<Document *>(doc);
	return truedoc;
	}

Node::NodeIterator::NodeIterator(Node *startnode, bool deep)
	: deep_(deep), currentnode_(startnode), startnode_(startnode), skip_current_children_(false)
	{
	}

Node::NodeIterator::NodeIterator(Node *startnode, bool deep, bool siblings_at_toplevel)
	: deep_(deep), siblings_at_toplevel_(siblings_at_toplevel),
	  currentnode_(startnode), startnode_(startnode), skip_current_children_(false)
	{
	}

Node *Node::NodeIterator::nextnode(void)
	{
	if(!increment_()) {
		currentnode_=0;
		return 0;
		}
	return currentnode_;
	}

Node *Node::NodeIterator::prevnode(void)
	{
	if(!decrement_()) {
		currentnode_=0;
		return 0;
		}
	return currentnode_;
	}

Node *Node::NodeIterator::thisnode(void)
	{
	return currentnode_;
	}

void Node::NodeIterator::skip_current_children(void)
	{
	skip_current_children_=true;
	}

Node *Node::NodeIterator::operator*(void)
	{
	return currentnode_;
	}

Node::NodeIterator& Node::NodeIterator::operator++(void)
	{
	nextnode();
	return *this;
	}

Node::NodeIterator& Node::NodeIterator::operator--(void)
	{
	prevnode();
	return *this;
	}

int Node::NodeIterator::levels_changed(void) const
	{
	return levels_changed_;
	}

bool Node::NodeIterator::increment_(void)
	{
	if(!skip_current_children_ && currentnode_->firstChild() != 0) {
		levels_changed_=1;
		currentnode_=currentnode_->firstChild();
		return true;
		}
	else {
		skip_current_children_=false;
		levels_changed_=0;
		while(currentnode_->nextSibling()==0) {
			currentnode_=currentnode_->parentNode();
			levels_changed_--;
			if(currentnode_==0 || currentnode_==startnode_->parentNode())
				return false;
			}
		if(siblings_at_toplevel_==false)
			if(currentnode_==startnode_)
				return false;
		currentnode_=currentnode_->nextSibling();
		return true;
		}
	}

bool Node::NodeIterator::decrement_(void)
	{
	if(currentnode_->lastChild() != 0) {
		currentnode_=currentnode_->lastChild();
		return true;
		}
	else {
		while(currentnode_->previousSibling()==0) {
			currentnode_=currentnode_->parentNode();
			if(currentnode_==0 || currentnode_==startnode_->parentNode())
				return false;
			}
		currentnode_=currentnode_->previousSibling();
		return true;
		}
	}

//--------------------------------------------------------------

NamespaceDecl::NamespaceDecl(const DOMString& prefix, const DOMString& uri)
	: prefix_(prefix), uri_(uri)
	{
	}

DOMString NamespaceDecl::uri(void) const
	{
	return uri_;
	}

DOMString NamespaceDecl::prefix(void) const
	{
	return prefix_;
	}


//--------------------------------------------------------------

Document::Document(const string& url_spec="", const string& content_type="", const DocumentType* doctype=0) 
	: doctype_(doctype), url_(url_spec), content_type_(content_type), implementation_(0)
	{
	}

Document::~Document(void)
	{
	map<const Node *, const NamespaceDecl *, less<const Node *> >::iterator it=namespacedecls_.begin();
	while(it!=namespacedecls_.end()) {
		delete (*it).second;
		++it;
		}
	if(doctype_!=0)
		delete doctype_;
	debug("DOM::Document deleted");
	}

Node::NodeType Document::nodeType(void) const
	{
	return Node::DOCUMENT_NODE;
	}

const DOMString& Document::nodeName(void) const
	{
	static string tmp("#document");
	return tmp;
	}

void Document::set_doctype(const DocumentType *dt) 
	{
	doctype_=dt;
	}

void Document::set_default_namespace(const NamespaceDecl *ns)
	{
	namespacedecls_[this]=ns;
	}

const DocumentType *Document::doctype(void) const
	{
	return doctype_;
	}

string Document::url(void) const
	{
	return url_;
	}

string Document::content_type(void) const
	{
	return content_type_;
	}

DOMImplementation *Document::implementation(void) const
	{
	return implementation_;
	}

/* Element *Document::documentElement(void) const
	{
	return this; // FIXME: that's not quite according to the specs.
	 } */

Element *Document::createElement(const DOMString& tagName) throw(DOMException)
	{
	string tmp=tagName;
/*	if(doctype_==0) {
		for(unsigned int i=0; i<tmp.size(); ++i)
			tmp[i]=toupper(tmp[i]);
		} */
	return new Element(tmp);
	}

DocumentFragment *Document::createDocumentFragment(void)
	{
	return new DocumentFragment();
	}

Text *Document::createTextNode(const DOMString& data)
	{
	return new Text(data);
	}

Comment *Document::createComment(const DOMString& data)
	{
	return new Comment(data);
	}

CDATASection *Document::createCDATASection(const DOMString& data) throw(DOMException)
	{
	return new CDATASection(data);
	}

ProcessingInstruction *Document::createProcessingInstruction(const DOMString& target, const DOMString& data)
	{
	return new ProcessingInstruction(target,data);
	}

Attr* Document::createAttribute(const DOMString& name)
	{
	return new Attr(name);
	}

EntityReference *Document::createEntityReference(const DOMString& name)
	{
	return new EntityReference(name);
	}

void Document::setNamespace(Node *nd, const NamespaceDecl *dec)
	{
	namespacedecls_[nd]=dec;
	}

pair<Node *, NamespaceDecl *> Document::getNamespace(const DOMString& prefix)
	{
	return pair<Node *, NamespaceDecl *>(0,0);
	}

const NamespaceDecl *Document::nodeNamespace(const Node *nod) const
	{
	do {
		map<const Node *, const NamespaceDecl *, less<Node *> >::const_iterator it;
		it=namespacedecls_.find(nod);
		if(it!=namespacedecls_.end()) {
			return (*it).second;
			}
		nod=nod->parentNode();
		} while(nod!=0);
	throw logic_error("[lib-dom]: reached top of DOM tree while searching for namespace declaration");
	}

//--------------------------------------------------------------

DocumentFragment::DocumentFragment(void)
	{
	}

DocumentFragment::~DocumentFragment()
	{
	}

Node::NodeType DocumentFragment::nodeType(void) const
	{
	return Node::DOCUMENT_FRAGMENT_NODE;
	}

const DOMString& DocumentFragment::nodeName(void) const
	{
	static string tmp("#document-fragment");
	return tmp;
	}


//--------------------------------------------------------------

unsigned long NamedNodeMap::length(void) const
	{
	throw logic_error("not implemented");
	return 0;
	}


//--------------------------------------------------------------

CharacterData::CharacterData(const DOMString& data)
	: data_(data)
	{
	}

CharacterData::~CharacterData()
	{
	}

void CharacterData::appendData(const DOMString& dat)
	{
	data_+=dat;
	}

//--------------------------------------------------------------

Attr::Attr(const DOMString& name)
	: attrname_(name), specified_(false)
	{
	}

Attr::Attr(const DOMString& name, const DOMString &value)
	: attrname_(name), attrvalue_(value), specified_(true)
	{
	}

Attr::~Attr()
	{
	}

Node::NodeType Attr::nodeType(void) const
	{
	return Node::ATTRIBUTE_NODE;
	}

const DOMString& Attr::nodeName(void) const
	{
	return attrname_;
	}

const DOMString& Attr::nodeValue(void) const
	{
	return attrvalue_;
	}

const DOMString& Attr::name(void) const
	{
	return attrname_;
	}

boolean Attr::specified(void) const
	{
	return specified_;
	}

const DOMString& Attr::value(void) const
	{
	return attrvalue_;
	}

void Attr::set_value(const DOMString& val)
	{
	attrvalue_=val;

	if(attrvalue_!="") // strip off quotes.
		if(attrvalue_[0]==34)
			attrvalue_=attrvalue_.substr(1,attrvalue_.size()-2);


	specified_=true;
	}

//--------------------------------------------------------------

Element::Element(const DOMString &name)
	: tagname_(name), attributes_(0)
	{
	}

Element::~Element()
	{
	if(attributes_!=0)
		delete attributes_;
	// attributes_ is a node which will delete the children automatically.
	}

Node::NodeType Element::nodeType(void) const
	{
	return Node::ELEMENT_NODE;
	}

const DOMString& Element::nodeName(void) const
	{
	return tagname_;
	}

NamedNodeMap *Element::attributes(void) const
	{
	throw logic_error("attribute unimplemented");
	return 0;
	}

Attr *Element::firstAttribute(void) const
	{
	return attributes_;
	}

const DOMString& Element::tagName(void) const
	{
	return tagname_;
	}

const DOMString& Element::getAttribute(const DOMString& name) const
	{
	static string tmps("");
	Attr *tmp=getAttributeNode(name);
	if(tmp==0 || tmp->specified()==false)
		return tmps;
	else
		return tmp->value();
	}

void Element::setAttribute(const DOMString& name, const DOMString &value)
	{
	Attr *tmp=getAttributeNode(name);
	if(tmp==0) {
		tmp=new Attr(name);
		tmp->set_value(value);
		if(attributes_==0)
			attributes_=tmp;
		else {
			Attr *here=attributes_;
			while(here->firstChild()!=0) {
				here=static_cast<Attr *>(here->firstChild());
				}
			here->appendChild(tmp);
			}
		}
	else {
		tmp->set_value(value);
		}
	}

Attr *Element::getAttributeNode(const DOMString& name) const
	{
	Attr *here=attributes_;
	while(here!=0) {
		if(here->name()==name)
			return here;
		here=static_cast<Attr *>(here->firstChild());
		}
	return 0; 
	}

//--------------------------------------------------------------

Text::Text(const DOMString& data)
	: CharacterData(data)
	{
	}

Text::~Text()
	{
	}

Node::NodeType Text::nodeType(void) const
	{
	return Node::TEXT_NODE;
	}

const DOMString& Text::nodeName(void) const
	{
	static string tmp("#text");
	return tmp;
	}

const DOMString& Text::nodeValue(void) const
	{
	return data_;
	}


//--------------------------------------------------------------

Comment::Comment(const DOMString& data)
	: CharacterData(data)
	{
	}

Comment::~Comment()
	{
	}

Node::NodeType Comment::nodeType(void) const
	{
	return Node::COMMENT_NODE;
	}

const DOMString& Comment::nodeName(void) const
	{
	static string tmp("#comment");
	return tmp;
	}

const DOMString& Comment::nodeValue(void) const
	{
	return data_;
	}




//--------------------------------------------------------------

CDATASection::CDATASection(const DOMString& data) 
	: Text(data)
	{
	}

CDATASection::~CDATASection()
	{
	}

Node::NodeType CDATASection::nodeType(void) const
	{
	return Node::CDATA_SECTION_NODE;
	}

const DOMString& CDATASection::nodeName(void) const
	{
	static string tmp("#cdata-section");
	return tmp;
	}

const DOMString& CDATASection::nodeValue(void) const
	{
	return data_;
	}


//--------------------------------------------------------------

DocumentType::DocumentType(const DOMString& doctypename)
	: doctypename_(doctypename)
	{
	}

DocumentType::~DocumentType()
	{
	}

Node::NodeType DocumentType::nodeType(void) const
	{
	return Node::DOCUMENT_TYPE_NODE;
	}

const DOMString& DocumentType::nodeName(void) const
	{
	return doctypename_;
	}

const DOMString& DocumentType::name(void) const
	{
	return doctypename_;
	}

NamedNodeMap *DocumentType::entities(void) const
	{
	return 0;
	}

NamedNodeMap *DocumentType::notations(void) const
	{
	return 0;
	}

//--------------------------------------------------------------

Notation::Notation(const DOMString& notname)
	: notationname_(notname)
	{
	}

Notation::~Notation()
	{
	}

Node::NodeType Notation::nodeType(void) const
	{
	return Node::NOTATION_NODE;
	}

const DOMString& Notation::nodeName(void) const
	{
	return notationname_;
	}


//--------------------------------------------------------------

Entity::Entity(const DOMString& entityname)
	: entityname_(entityname)
	{
	}

Entity::~Entity()
	{
	}

Node::NodeType Entity::nodeType(void) const
	{
	return Node::ENTITY_NODE;
	}

const DOMString& Entity::nodeName(void) const
	{
	return entityname_;
	}


//--------------------------------------------------------------

EntityReference::EntityReference(const DOMString &name)
	: entityref_(name)
	{
	}

EntityReference::~EntityReference(void)
	{
	}

Node::NodeType EntityReference::nodeType(void) const
	{
	return Node::ENTITY_REFERENCE_NODE;
	}

const DOMString& EntityReference::nodeName(void) const
	{
	return entityref_;
	}


//--------------------------------------------------------------

ProcessingInstruction::ProcessingInstruction(const DOMString& target, 
															const DOMString& data) 
	: target_(target), data_(data)
	{
	}

ProcessingInstruction::~ProcessingInstruction()
	{
	}

Node::NodeType ProcessingInstruction::nodeType(void) const
	{
	return Node::PROCESSING_INSTRUCTION_NODE;
	}

const DOMString& ProcessingInstruction::nodeName(void) const
	{
	return target_;
	}

const DOMString& ProcessingInstruction::nodeValue(void) const
	{
	return data_;
	}

