// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  GTK-- layout boxes, implementation.

  Copyright 1998-1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "gtklayout/gtkboxes.hh"
#include <stdexcept>
#include <algo.h>
#include <math.h>
#include <vector.h>
#include <typeinfo>

#include <gdk/gdkprivate.h>
#include <gtk--/main.h>
#include <gdk/gdkrgb.h>

#define DEBUG_CLASS "GTKLAYOUT"
#include <mnemonic/debug.hh>
#include <mnemonic/debugger.hh>
#include <mnemonic/stopwatch.hh>

#include <layout/fonts.hh>
#include <layout/colors.hh>
#include "gtklayout/gtkfonts.hh"
#include "gtklayout/gtkcolors.hh"

// gtkfactory

layout::viewer_box *gtkfactory::create_viewer_box(void)
	{
	return 0;
	}

layout::text_box *gtkfactory::create_text_box(const string& txt)
	{
	layout::text_box *tmp=new gtktext_box(txt);
	return tmp;
	}

layout::glue *gtkfactory::create_glue(void)
	{
	layout::glue *tmp=new gtkglue;
	return tmp;
	}

layout::hbox *gtkfactory::create_hbox(void)
	{
	layout::hbox *tmp=new gtkhbox;
	return tmp;
	}

layout::vbox *gtkfactory::create_vbox(void)
	{
	layout::vbox *tmp=new gtkvbox;
	return tmp;
	}

layout::border_box *gtkfactory::create_border_box(void)
	{
	layout::border_box *tmp=new gtk_border_box;
	return tmp;
	}

layout::event_box *gtkfactory::create_event_box(void)
	{
	layout::event_box *tmp=new gtkeventbox;
	return tmp;
	}

layout::button_box *gtkfactory::create_button_box(const string& str)
	{
	return 0;//new gtkbuttonbox(str); WARNING: button_box contains a GTK thing so
	// you have to gdk_threads_enter. Or make that button dynamic and handle it in 
	// the gtkbuttonbox (cleaner).
	}

// gtkviewer

gtkviewer::gtkviewer(void)
	: scanvas_(const_cast<gtkviewer *>(this)), color_mgr_(this), font_mgr_(this)
   {
   }

layout::area *gtkviewer::current_area(void) const
	{
	return &scanvas_;
	}

viewercanvas& gtkviewer::scanvas(void) const
	{
	return scanvas_;
	}

void gtkviewer::reflow(void)
	{
	scanvas_.canvas()->relayout_all();
	}

void gtkviewer::redraw(layout::rectangle rt)
	{
	debug(this);
	debug("request for redraw of " << rt.top_left.x << "," << rt.top_left.y
			<< "-" << rt.size.x << "," << rt.size.y);
	gdk_threads_enter();
	scanvas_.redraw_rectangle(rt.top_left.x, rt.top_left.y, rt.size.x, rt.size.y);
	gdk_threads_leave();
	}

int gtkviewer::draw(layout::point loc, const box::draw_info& inf, box *fe)
	{
	color_mgr().activate_foreground(properties()->current_background_color());
	debug(inf.viewport_loc.x << " " << inf.viewport_loc.y);
	debug(inf.viewport_size.x << " " << inf.viewport_size.y);
	debug(inf.bitmap_offset.x << " " << inf.bitmap_offset.y);
	scanvas_.canvas()->pixmap().draw_rectangle(scanvas_.canvas()->gc(),TRUE,
 															  inf.bitmap_offset.x,  
															  inf.bitmap_offset.y,
  															  inf.viewport_size.x, inf.viewport_size.y);
	return container_box::draw(loc, inf, fe);
	}

void gtkviewer::freeze(int32_t hsize, int32_t vsize, box *forced_end)
	{
	hbox::freeze(hsize, vsize, 0);
	if(scanvas_.width()==1 && hsize!=100) {
		scanvas_.canvas()->set_size(100,100);
		scanvas_.show();
		} 
	}

layout::font_manager& gtkviewer::font_mgr(void)
	{
	return font_mgr_;
	}

layout::color_manager& gtkviewer::color_mgr(void)
	{
	return color_mgr_;
	}



// viewercanvas

viewercanvas::viewercanvas(gtkviewer *owner)
	: owner_(owner)
	{
	}

void viewercanvas::draw_offscreen_impl(gtkcanvas::draw_info& inf)
	{
	if(trylock()) {
		layout::box::draw_info tmp;
		// This seems a bit of a waste, but I didn't want to make the canvaswidgets depend
		// on lib-layout. So even though the draw_info structures are identical in content,
		// we still have to convert to the box::draw_info one here.
		tmp.viewport_loc.x=inf.viewport_loc.x;
		tmp.viewport_loc.y=inf.viewport_loc.y;
		tmp.viewport_size.x=inf.viewport_size.x;
		tmp.viewport_size.y=inf.viewport_size.y;
		tmp.bitmap_offset.x=inf.bitmap_offset.x;
		tmp.bitmap_offset.y=inf.bitmap_offset.y;
		tmp.viewport_botright.x=inf.viewport_loc.x+inf.viewport_size.x;
		tmp.viewport_botright.y=inf.viewport_loc.y+inf.viewport_size.y;
		if(inf.move_too) {
			debug("moving first");
			assert(owner_!=0);
			owner_->position(::layout::point(0,0),tmp,0);
			}
		debug("previous redraw was:");
		debug("from " << previousinfo.viewport_loc.x << "," << previousinfo.viewport_loc.y);
		debug("to   " << previousinfo.viewport_botright.x << "," << previousinfo.viewport_botright.y);
		debug("drawing!!! viewport info:");
		debug("from " << tmp.viewport_loc.x << "," << tmp.viewport_loc.y);
		debug("to   " << tmp.viewport_botright.x << "," << tmp.viewport_botright.y);
		// The idea here will be to remember the last partially drawn box and the
		// first partially drawn box of each draw. Then, when we scroll down, go
		// directly to the last partially drawn box of the previous draw. Analogous
		// for scrolling up. We will have to keep track of all iterators up to the
		// first and last drawing point for this to work. Scrolling up is going to 
		// be tricky as the tree is only singly linked...
		if(previousinfo.viewport_botright.y==tmp.viewport_loc.y &&
			previousinfo.viewport_botright.x==tmp.viewport_botright.x &&
			previousinfo.viewport_loc.x==tmp.viewport_loc.x) {
			debug("scrolling DOWN by " << tmp.viewport_size.y << " pixels");
			}
		else if (previousinfo.viewport_loc.y==tmp.viewport_botright.y &&
					previousinfo.viewport_botright.x==tmp.viewport_botright.x &&
					previousinfo.viewport_loc.x==tmp.viewport_loc.x) {
			debug("scrolling UP by " << tmp.viewport_size.y << " pixels");
			}
		owner_->draw(::layout::point(0,0),tmp,0);
		unlock();
		previousinfo=tmp;
		}
	else {
		debug("tree locked, no redraw");
		}
	}

gint viewercanvas::motion_notify_event_impl(GdkEventMotion *event)
	{
	int x, y;
	GdkModifierType state;
	
	inf_.etype=layout::box::event_info::mouse_move;
	if (event->is_hint) {
      gdk_window_get_pointer (event->window, &x, &y, &state);
		inf_.hitpoint.x=x+(long)(canvas()->get_hadjustment()->get_value());
		inf_.hitpoint.y=y+(long)(canvas()->get_vadjustment()->get_value());
		}
	else {
		inf_.hitpoint.x=(int32_t)(event->x)+(long)(canvas()->get_hadjustment()->get_value());
		inf_.hitpoint.y=(int32_t)(event->y)+(long)(canvas()->get_vadjustment()->get_value());
		state = (GdkModifierType) event->state;
      }
	
	// FIXME: should not go in this library
	if(canvas()->window().gdkobj()==event->window) {
//		gdk_window_set_cursor(canvas()->window().gdkobj(),0);
		debug("mouse at " << inf_.hitpoint.x << "," << inf_.hitpoint.y);
		owner_->propagate_event(::layout::point(0,0), inf_, 0);
		}
	return 0;
	}

gint viewercanvas::button_press_event_impl(GdkEventButton *event)
	{
	if(canvas()->window().gdkobj()==event->window) {
		inf_.etype=layout::box::event_info::button_down;
		inf_.hitpoint.x=(int32_t)(event->x)+(long)(canvas()->get_hadjustment()->get_value());
		inf_.hitpoint.y=(int32_t)(event->y)+(long)(canvas()->get_vadjustment()->get_value());
		debug("button " << event->button << " pressed at " << inf_.hitpoint.x << "," << inf_.hitpoint.y);
		owner_->propagate_event(::layout::point(0,0), inf_, 0);
		}
	return 0;
	}

gint viewercanvas::button_release_event_impl(GdkEventButton *event)
	{
	if(canvas()->window().gdkobj()==event->window) {
		inf_.etype=layout::box::event_info::button_up;
		inf_.hitpoint.x=(int32_t)(event->x)+(long)(canvas()->get_hadjustment()->get_value());
		inf_.hitpoint.y=(int32_t)(event->y)+(long)(canvas()->get_vadjustment()->get_value());
		debug("button " << event->button << " released at " << inf_.hitpoint.x << "," << inf_.hitpoint.y);
		owner_->propagate_event(::layout::point(0,0), inf_, 0);
		}
	return 0;
	}

gtkcanvas::point viewercanvas::layout_objects_impl(gtkcanvas::point& siz)
	{
	if(getenv("PRINTTREE")) {
#ifdef DEBUG
		debugger::instance().lock();
		cerr << "WARNING: environment variable PRINTTREE is set, which slows down" << endl
			  << "         Mnemonic tremendously." << endl;
		debugger::instance().unlock();
#endif
		print_layout_tree(owner_);
		}
	debug("requested size " << siz.x << "x" << siz.y);
	gtkcanvas::point tmp;
	if(trylock()) {
		owner_->set_absolute_size(siz.x,0,0);
		
		if(owner_->check_recompute_children() || owner_->check_propagate_changes()) { 
			layout::point realsize=owner_->size(siz.x, 0, owner_->check_propagate_changes(),0);
			owner_->freeze(realsize.x, 0, 0); // remember: a zero in the y direction means that anything is fine
			if(getenv("PRINTTREE")) {
				print_layout_tree(owner_);
				}
			}
		else { 
			debug("weird, no relayout required at all?");
			}
		
		tmp.x=owner_->frozen_size().x;
		tmp.y=owner_->frozen_size().y;
		unlock();
		debug("layout_objects done " << owner_);
		}
	else { 
		debug("tree locked");
		}
	return tmp;
	}



// Text boxes draw themselves on the canvas exported by the container object.

gtktext_box::gtktext_box(const string& str)
	: text_box(str)
   {
   }

int gtktext_box::draw(layout::point loc, const box::draw_info& inf, box *fe)
   {
   // Testing for every box is a waste of time; it is faster to draw a few more words than
   // necessary and only test individual lines. The only exception to that is when we are
	// using t1 fonts.
	
#ifdef USE_T1LIB
	if(static_cast<gtk_fontspec *>(properties()->current_font())->font_inf().connected()==false) {
		int ourloc=am_outside(loc,inf);
		// FIXME: am_outside needs to return octant numbers, but we will
		// just fake a draw.
		if(ourloc!=5) return 5;
		
		// TT also requires the background to be set.
		parent()->color_mgr().activate_background(properties()->current_background_color());
		}
#endif
	parent()->color_mgr().activate_foreground(properties()->current_foreground_color());
	parent()->font_mgr().draw_string(properties()->current_font(), text_,
												frozen_location().x+loc.x-inf.viewport_loc.x
												+inf.bitmap_offset.x,
												frozen_location().y 
												+loc.y-inf.viewport_loc.y
												+inf.bitmap_offset.y
												);

	if(parent()->subtree_is_interactive()) {
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
		vc->canvas()->pixmap().draw_rectangle(
			vc->canvas()->gc(),TRUE,
			frozen_location().x+loc.x-inf.viewport_loc.x
			+inf.bitmap_offset.x,
			frozen_location().y+loc.y-inf.viewport_loc.y
			+inf.bitmap_offset.y+1,
			frozen_size().x,1);
		}
	return 5;
   }


// Button boxes attach themselves to a fixed location on the Gtk::Layout exported by
// the container object. 

gtkbuttonbox::gtkbuttonbox(const string& txt)
	: img_(txt)
   {
   }

layout::point gtkbuttonbox::frozen_size(void) const
   {
	return layout::point(img_.width(), img_.height());
   }

int gtkbuttonbox::position(layout::point loc, const box::draw_info& inf, box *fe)
	{
	debug("moving child " << typeid(*this).name());

	viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
	vc->canvas()->move(img_,
							 frozen_location().x+loc.x,
							 frozen_location().y+loc.y-frozen_size().y+shift());
	
	return 5;
	}

void gtkbuttonbox::box_added_impl(void)
	{
	viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
	vc->canvas()->put(img_,10000,10000);
	img_.show();
	}


// image boxes use Gdk_Pixmaps to draw image on the canvas

gtkimagebox::gtkimagebox(int32_t w, int32_t h)
	: pixmap_(0),clipmap_(0), modified_(true)
	{
	// FIXME: something fishy is going on here. If you do NOT set these by hand,
	// they don't get set to zero by the resizable_box constructor. Why not?
	width_=w;
	height_=h;
	}

gtkimagebox::~gtkimagebox()
	{
	nod();
	if(pixmap_) delete pixmap_;
	if(clipmap_) delete clipmap_;
	}

void gtkimagebox::modify(ptr<scaled_raster_image> rim)
	{
	rim_=rim;
	if(rim_!=0) {
		if(width_!=rim_->scaled_width() || height_!=rim_->scaled_height()) {
			width_=rim_->scaled_width();
			height_=rim_->scaled_height();
			shift_=0;
			mark_to_root();
			mark_recompute_children();
			mark_propagate_changes();
			}
		}
	modified_=true; // triggers rebuilding of pixmap
	}

int gtkimagebox::draw(layout::point loc, const box::draw_info& inf, box *fe)
	{
	debug("drawing imagebox");
   int ourloc=am_outside(loc,inf);
	if(ourloc!=5) return ourloc;
	
	int32_t xstart=loc.x+frozen_location().x-inf.viewport_loc.x;
	int32_t ystart=loc.y+frozen_location().y-frozen_height_+shift_-inf.viewport_loc.y;
	int32_t xlen=frozen_width_;
	int32_t ylen=frozen_height_;
	if(xstart<0) {
		xlen+=xstart;
		xstart=0;
		}
	if(ystart<0) {
		ylen+=ystart;
		ystart=0;
		}
	if(xlen>inf.viewport_size.x)
		xlen=inf.viewport_size.x;
	if(ylen>inf.viewport_size.y)
		ylen=inf.viewport_size.y;
	
	viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
	if(modified_) {
		if(rim_!=0) {
			debug("image modified");
			if(!pixmap_.connected() || 
				rim_->scaled_height()!=frozen_height_ || rim_->scaled_width()!=frozen_width_) {
				debug("size changed, rebuilding pixmap");
				if(pixmap_.connected())  pixmap_.release();
				if(clipmap_.connected()) clipmap_.release();
//				if (rim_->scaled_image_alpha())	
//					clipmap_=new Gdk_Bitmap;
				pixmap_.create(vc->canvas()->window(),rim_->scaled_width(), 
									rim_->scaled_height(),
									vc->canvas()->window().get_depth());
				}
			if (rim_->scaled_image_alpha())	{
				nod();
    			int h=rim_->scaled_height();
				int w=rim_->scaled_width();
				int bw=w/8;
				if (w%8) bw++;
				gchar *bits=new gchar[h*bw ];
				long i=0;
				long j=0; 
				int b=0;
				gchar t=0;
				long k=0; 
				for (i=0;i<h;i++){
					for (j=0;j<w;j++){
						if (rim_->scaled_image_alpha()[i*w+j]>0){
							t+=1<<b;
							}
						b++;
						if (b==8) {
							b=0;
							bits[k]=t;
							t=0;
							k++;
							}
						}
					if (b!=0) {
						bits[k]=t;
						k++;
						b=0;
						t=0;
						}
					}
				clipmap_.create(vc->canvas()->window(),bits,rim_->scaled_width(),rim_->scaled_height());
				delete[] bits;
				}
			nod();
			gdk_draw_rgb_image(pixmap_.gdkobj(),
									 vc->canvas()->gc(),
									 0,0,
									 frozen_width_,frozen_height_,
									 GDK_RGB_DITHER_NORMAL,
									 ((guchar*)rim_->scaled_image_data()),
									 3*rim_->scaled_width());
			}
		modified_=false;
		}
	nod();
	if(pixmap_.connected()) {
		gint bgw, bgh;
		nod();
		gdk_window_get_size(pixmap_.gdkobj(), &bgw, &bgh);
		nod();
		if(frozen_location().y+loc.y-inf.viewport_loc.y+inf.bitmap_offset.y+
			inf.viewport_size.y+min(frozen_height_,bgh)>0) {
			nod();
			Gdk_GC& tgc=vc->canvas()->gc();
			nod();
			if (rim_->scaled_image_alpha())	tgc.set_clip_mask(clipmap_);
			nod();
			tgc.set_clip_origin(
				frozen_location().x+loc.x-inf.viewport_loc.x+inf.bitmap_offset.x,
				frozen_location().y+loc.y-frozen_height_+shift_
				-inf.viewport_loc.y+inf.bitmap_offset.y
				);
			nod();	
			vc->canvas()->pixmap().draw_pixmap(vc->canvas()->gc(),pixmap_,
														  0,0,
														  frozen_location().x+loc.x-inf.viewport_loc.x+inf.bitmap_offset.x,
														  frozen_location().y+loc.y-frozen_height_+shift_
														  -inf.viewport_loc.y+inf.bitmap_offset.y,
														  min(frozen_width_,bgw),
														  min(frozen_height_,bgh)
														  ); 
			nod();											  
			if (rim_->scaled_image_alpha())	tgc.set_clip_mask();
			nod();
			}
		nod();												  
		} 
	nod();
	debug("drawing imagebox done");
	return ourloc;
	}



// Glue just acts as a placeholder, pretty trivial stuff.

int gtkglue::draw(layout::point loc, const box::draw_info& inf, box *fe)
	{
	if(parent()->subtree_is_interactive()) {
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
		vc->canvas()->pixmap().draw_rectangle(
			vc->canvas()->gc(),TRUE,
			frozen_location().x+loc.x-inf.viewport_loc.x
			+inf.bitmap_offset.x,
			frozen_location().y+loc.y-inf.viewport_loc.y
			+inf.bitmap_offset.y+1,
			frozen_width_,1);
		}
	return 5;
	}


// HVboxes are the only boxes that actually have the nontrivial gtk display
// stuff onboard, since they create Gtk::Layout container object with a canvas
// inside.

void gtkhbox::fill_background_impl(layout::point loc, const box::draw_info& inf)
	{	
	debug("ready to draw hvcolorbox background");

	int32_t xstart=loc.x+frozen_location().x-inf.viewport_loc.x;
	int32_t ystart=loc.y+frozen_location().y-frozen_height_+shift_-inf.viewport_loc.y;
	int32_t xlen=frozen_width_;
	int32_t ylen=frozen_height_;
	if(xstart<0) {
		xlen+=xstart;
		xstart=0;
		}
	if(ystart<0) {
		ylen+=ystart;
		ystart=0;
		}
	if(xlen>inf.viewport_size.x)
		xlen=inf.viewport_size.x;
	if(ylen>inf.viewport_size.y)
		ylen=inf.viewport_size.y;
	if(ylen>0) {
		debug("coloring height " << ylen);
		parent()->color_mgr().activate_foreground(properties()->current_background_color());
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
		vc->canvas()->pixmap().draw_rectangle(vc->canvas()->gc(),TRUE,
														  xstart+inf.bitmap_offset.x,ystart+inf.bitmap_offset.y,
														  xlen,ylen);
		}
	debug("all drawing done");
	}

void gtkvbox::fill_background_impl(layout::point loc, const box::draw_info& inf)
	{	
	debug("ready to draw hvcolorbox background");

	int32_t xstart=loc.x+frozen_location().x-inf.viewport_loc.x;
	int32_t ystart=loc.y+frozen_location().y-frozen_height_+shift_-inf.viewport_loc.y;
	int32_t xlen=frozen_width_;
	int32_t ylen=frozen_height_;
	if(xstart<0) {
		xlen+=xstart;
		xstart=0;
		}
	if(ystart<0) {
		ylen+=ystart;
		ystart=0;
		}
	if(xlen>inf.viewport_size.x)
		xlen=inf.viewport_size.x;
	if(ylen>inf.viewport_size.y)
		ylen=inf.viewport_size.y;
	if(ylen>0) {
		debug("coloring height " << ylen);
		parent()->color_mgr().activate_foreground(properties()->current_background_color());
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
		vc->canvas()->pixmap().draw_rectangle(vc->canvas()->gc(),TRUE,
														  xstart+inf.bitmap_offset.x,ystart+inf.bitmap_offset.y,
														  xlen,ylen);
		}
	debug("all drawing done");
	}

/*
     void draw_arc       (Gdk_GC       &gc,
                          gint          filled,
                          gint          x,
                          gint          y,
                          gint          width,
                          gint          height,
                          gint          angle1,
                          gint          angle2);
								  */

int gtk_border_box::draw(layout::point loc, const box::draw_info& inf, box *fe)
	{
	int num=layout::vbox::draw(loc,inf,fe);

	int32_t xstart=loc.x+frozen_location().x-inf.viewport_loc.x;
	int32_t ystart=loc.y+frozen_location().y-frozen_height_+shift_-inf.viewport_loc.y;
	int32_t xlen=frozen_width_;
	int32_t ylen=frozen_height_;
	if(xstart<0) {
		xlen+=xstart;
		xstart=0;
		}
	if(ystart<0) {
		ylen+=ystart;
		ystart=0;
		}
	if(xlen>inf.viewport_size.x)
		xlen=inf.viewport_size.x;
	if(ylen>inf.viewport_size.y)
		ylen=inf.viewport_size.y;
	if(ylen>0) {
		parent()->color_mgr().activate_foreground(properties()->current_foreground_color());
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());
		vc->canvas()->pixmap().draw_rectangle(vc->canvas()->gc(),TRUE,
														  xstart+inf.bitmap_offset.x, ystart+inf.bitmap_offset.y,
														  xlen,ylen);
		}
	return num;
	}

gtkeventbox::gtkeventbox(void)
	{
	gdk_threads_enter();
	hand_=gdk_cursor_new(GDK_HAND2);
	gdk_threads_leave();
	}

gtkeventbox::~gtkeventbox()
	{
	if(hand_)    delete hand_;
	}

bool gtkeventbox::accept_event(box::event_info& et)
	{
	debug("gtkeventbox accept_event");
	if(layout::event_box::accept_event(et)) {
		viewercanvas *vc=static_cast<viewercanvas *>(parent()->current_area());				
		gdk_window_set_cursor(vc->canvas()->window(),hand_);
		return true;
		}
	else return false;
	}

