// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  GTK font code.

  Copyright 1998-1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string>
#include <gtklayout/gtkboxes.hh>
#include <gtklayout/gtkfonts.hh>
#include <gtklayout/gtkcolors.hh>

#ifdef USE_T1LIB
#include <t1lib.h>
#include <t1libx.h>
#include <gdk--/color.h>
#include <gdk--/gc.h>
#include <gdk/gdkx.h>
#include <gdk/gdkrgb.h>
#endif

#include <stdexcept>

#define DEBUG_CLASS "GTKFONTS"
#include <mnemonic/debug.hh>
#include <mnemonic/stopwatch.hh>

bool gtk_font_manager::t1ok_=false;

gtk_fontspec::gtk_fontspec(const layout::xlfd& xl, bool uset1)
	: layout::font(xl)
	{
	debug("creating font " << name());
	gdk_threads_enter();
	if(gdk_font_load(name().c_str())==0) {
		debug("not available as gdk font");
#ifdef USE_T1LIB
		if(uset1) {
			debug("trying t1lib");
			string goal=xl_.family+string(".pfb");
			debug("for font " << goal);
			for(t1num_=0; t1num_<T1_Get_no_fonts(); ++t1num_) {
				if(T1_GetFontFileName(t1num_)==goal)
					break;
				}
			if(t1num_==T1_Get_no_fonts()) {
				debug("also not available as t1lib font");
				debug("fallback to default font");
				xl_=layout::xlfd();
				font_.create(name());
				}
			else {
				debug("found a matching t1lib font");
				int res=T1_LoadFont(t1num_);
				debug("result of loadfont=" << res);
				}
			}
		else {
			debug("fallback to default font");
			xl_=layout::xlfd();
			font_.create(name());
			}
#else
		debug("fallback to default font");
		xl_=layout::xlfd();
		font_.create(name());
#endif
		}
	else {
		font_.create(name());
		debug("font id " << font_.font_id());
		}
	gdk_threads_leave();
	}

gtk_fontspec::~gtk_fontspec()
	{
	}

int32_t gtk_fontspec::string_width(const string& txt) const
	{
	if(txt=="\x0a") {
		return 0;
		}
	else {
#ifdef USE_T1LIB
		if(font_.connected()==false) {
			int num=T1_GetStringWidth(t1num_, const_cast<char *>(txt.c_str()), 0,0,0);
//			debug("string " << txt << " width " << num);
			return (int)(num/1000.0*xl_.size.pixels(95.0));
			}
#endif 
		gint lbearing, rbearing, swidth, ascent, descent;
				
		assert(font_.connected());
		font_.string_extents(txt, lbearing, rbearing, swidth, ascent, descent);
		return swidth;
		}
	}

int32_t gtk_fontspec::standard_ascent(void) const
	{
	if(standard_ascent_==-1) {
#ifdef USE_T1LIB
		if(font_.connected()==false) {
			BBox tmp=T1_GetFontBBox(t1num_);
//			debug("BBOX: " << tmp.llx << "," << tmp.lly << " " << tmp.urx << "," << tmp.ury);
			standard_ascent_=(int32_t)(tmp.ury/1000.0*xl_.size.pixels(95.0));
			standard_descent_=-(int32_t)(tmp.lly/1000.0*xl_.size.pixels(95.0));
			space_width_=(int32_t)(T1_GetCharWidth(t1num_,' ')/1000.0*xl_.size.pixels(95.0));
			return standard_ascent_;
			}
#endif
		gint lbearing, rbearing, swidth, ascent, descent;
		font_.string_extents("y", lbearing, rbearing, swidth, ascent, descent);
		standard_ascent_=(int32_t)(font_.height())-descent;
		standard_descent_=descent;
		space_width_=font_.char_width(' ');
		}
	return standard_ascent_;
	}

int32_t gtk_fontspec::standard_descent(void) const
	{
	if(standard_ascent_==-1) {
		standard_ascent(); // trigger calcs
		}
	return standard_descent_;
	}

int32_t gtk_fontspec::space_width(void) const
	{
	if(standard_ascent_==-1) {
		standard_ascent(); // trigger calcs
		}
	return space_width_;
	}

Gdk_Font& gtk_fontspec::font_inf(void) const
	{
	return font_;
	}

int gtk_fontspec::t1_inf(void) const
	{
	return t1num_;
	}

gtk_font_manager::gtk_font_manager(layout::box *owner)
	: owner_(owner), initialised_(false)
	{
	debug("constructing gtk font manager");
#ifdef USE_T1LIB
	if(getenv("MN_DISABLE_T1FONTS")==0 && !t1ok_) {
		if ((T1_InitLib(NO_LOGFILE)==NULL)) {
			t1ok_=false;
			debug("failed to initialise T1lib");
			cerr << endl << "T1lib is not installed correctly (probably your t1lib.config cannot be" << endl
				  << "found, or the FontDatabase file is not setup properly). Anti-aliased" << endl
				  << "fonts will not be available. See the T1lib documentation for details." << endl << endl
				  << "To disable this message, set the environment variable MN_DISABLE_T1FONTS." << endl;
			}
		else t1ok_=true;
		}
//	T1_SetLogLevel(3);
#endif
	debug("gtk_font_manager constructed");
	}

layout::font *gtk_font_manager::create_font_impl(const layout::xlfd& xl)
	{
	layout::font *tmp=new gtk_fontspec(xl, t1ok_);
	return tmp;
	}

void gtk_font_manager::draw_string(layout::font *ft, const string& txt, int32_t x, int32_t y)
	{
	gtk_fontspec *gft=static_cast<gtk_fontspec *>(ft);
	viewercanvas *vc=static_cast<viewercanvas *>(owner_->current_area());
#ifdef USE_T1LIB
	if(gft->font_inf().connected()==false && t1ok_) {
		if(!initialised_) {
			T1_SetX11Params(GDK_COLORMAP_XDISPLAY( gdk_rgb_get_cmap() ),
								 GDK_VISUAL_XVISUAL(	gdk_rgb_get_visual() ),
								 gdk_visual_get_best_depth(),
								 GDK_COLORMAP_XCOLORMAP( gdk_rgb_get_cmap() )
								 );
			debug("visual depth " << gdk_visual_get_best_depth());
			if(gdk_visual_get_best_depth()==8) {
				cerr << "T1 fonts on an 8-bit visual do not look very good yet." << endl
					  << "You may want to try the `--install-cmap' option when " << endl
					  << "starting mnemonic." << endl;
				}
			T1_AASetBitsPerPixel(gdk_visual_get_best_depth());
			T1_AASetLevel(T1_AA_HIGH);
			initialised_=true;
			}
		T1_AASetStringX(GDK_WINDOW_XWINDOW(vc->canvas()->pixmap().gdkobj()), 
							 GDK_GC_XGC(vc->canvas()->gc().gdkobj()), 
							 0,      // mode (0==transparent, 1==opaque (faster? not really)
							 x, y,   // coordinates
							 gft->t1_inf(),      // font id
							 const_cast<char *>(txt.c_str()), 0, // string
							 0,      // spaceoff
							 0,      // kerning
							 gft->xl().size.pixels(95.0),   // size
							 NULL    // transform matrix
							 ); 
		}
	else 
#endif
		vc->canvas()->pixmap().draw_string(gft->font_inf(),
													  vc->canvas()->gc(),
													  x,y,
													  txt);
	}

