// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Output device independent layout algorithms, interface.

  Copyright 1999, 2000 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef boxes_hh_
#define boxes_hh_

#include <string>
#include <vector.h>
#include <map.h>
#include <set.h>
#include <stdint.h>

#include <layout/numbers.hh>
#include <mnemonic/thread_mutex.hh>

namespace layout {

 // classes defined here
 class box;
 class resizable_box;
 class container_box;
 class area;
 class reflow_manager;
 class margin_manager;

 class glue;
 class text_box;
 class property_box;
 class event_box;
 class table;
 class hbox;
 class vbox;
 class parbox;

 // external
 class font;
 class color;
 class font_manager;
 class color_manager;

 /* 
    The inheritance tree for geometrically nested boxes looks as follows:

                           box
                            |
               +------------+------------+
               |                         |
           text_box                 resizable_box
                                         |
                                   +-----+------------+
                                   |                  |
                                  glue            container_box
                                                      |
      													+--------+-+------+-------+
      													|          |      |       |
      											  event_box    h/vbox parbox  table


   Properties are stored separately.

 */
                  
 class property_box {
	 public:
		 property_box(property_box *inherit, font *fnt=0, color *foreground=0, color *background=0);

		 virtual font  *current_font(void) const;
		 virtual color *current_foreground_color(void) const;
		 virtual color *current_background_color(void) const;

		 property_box  *inherit_from(void) const;
		 void           set_inherit(property_box *);
	 protected:
		 font         *font_;
		 color        *fg_color_, *bg_color_;
		 property_box *inherit_;
 };

 class property_manager {
	 public:
		 ~property_manager();

		 // Registering a property box multiple times is allowed.
		 void register_property(property_box *);
	 private:
		 set<property_box *> properties_;
 };

 class box {
	 public:
		 box(void);
		 virtual ~box();

		 // Normal boxes can only change position.
		 inline void            freeze_location(const point&);

		 // Read-out of frozen numbers.
		 virtual point          frozen_size(void) const;
		 virtual int32_t        shift(void) const;
		 enum baseline_dir { horizontal_baseline, vertical_baseline };
		 virtual baseline_dir   baseline(void) const;
		 inline  point          frozen_location(void) const;      // of the alignment point

		 inline void            mark_to_root(void); // starts at parent
		 inline void            set_parent(container_box *p);
		 inline void            set_next_sibling(box* s);
		 inline void            set_properties(property_box *p);

		 // Static properties.
		 enum break_property { mandatory_break_discard, optional_break_discard, no_break }; 
		 virtual break_property break_info(void) const;   // refers to breaking before the box
		 virtual bool           subtree_is_interactive(void) const;
		 virtual bool           ignore_geometry(void) const;
		 virtual bool           flat_hierarchy(void) const;
		 
		 virtual string         info(void) const;              // for debugging purposes

		 // drawing and positioning
		 class draw_info {
			 public:
				 point viewport_loc, viewport_size; 
				 point viewport_botright;  // precomputed sum of the above, to speedup redraws
				 point bitmap_offset;               // canvas origin does not have to coincide 
				                                    //   with viewport origin
		 };

		 //   1  2  3    am_outside returns these number to denote relative position
		 //   4  5  6    of the current box wrt. the viewport. In order for anything
		 //   7  8  9    but a 5 to be returned, the box must be completely outside the
       //              viewport.
		 inline  int     am_outside(const point&, const draw_info& ) const;

		 //  point_inside returns the same numbers as am_outside to indicate relative
		 //  position of the point wrt. the present box (which is 5).
		 virtual int     point_inside(point, point) const;
		 virtual int     draw(point, const draw_info &, box *forced_end);
		 virtual void    draw_outline(point, const draw_info &);
		 virtual int     position(point, const draw_info &, box *forced_end);

		 // Event propagation.
		 class event_info {
			 public:
				 event_info(void);

				 enum event_type { button_down, button_up, mouse_move };
				 event_type  etype;
				 point       hitpoint;
				 box        *accepting_box; // last box that accepted button down/up
				 box        *hovering_box;  // last hovering box
		 };
		 // propagate returns the same codes as point_inside. In addition,
		 // 10 = abort, event was accepted
		 virtual int     propagate_event(point, event_info &, box *forced_end);
		 virtual bool    accept_event(event_info& ev); // true if accepted, put box pointer in ev.

		 // These two keep calling the parent until they hit an actual implementation of the
		 // algorithm. Call these to trigger a reflow/redraw of the tree when you do not
		 // have direct access to the top-level box.
		 virtual void    reflow(void);
		 virtual void    redraw(void);        // redraw rectangle of this box
		 virtual void    redraw(rectangle);

		 // Tree navigation
		 class child_iterator {
			 public:
				 child_iterator(box *firstchild);
				 child_iterator(box *firstchild, bool cangoup);
				 
				 inline child_iterator&  operator++(void);
				 inline box             *operator*(void);
				 inline void             operator=(const child_iterator& other);
				 inline bool             operator==(const child_iterator& other) const;
				 bool   cangoup_;
			 private:
				 box     *start_node_;
				 box     *current_node_;
		 };
		 inline container_box   *parent(void) const;
		 inline box             *next_sibling(void) const;
		 inline property_box    *properties(void) const;

		 // Access to global info somewhere up the tree.
		 virtual font_manager&     font_mgr(void);
		 virtual color_manager&    color_mgr(void);
		 virtual margin_manager   *margin_mgr(void) const;
		 virtual reflow_manager   *reflow_mgr(void) const;
		 virtual property_manager *property_mgr(void) const;
		 virtual area             *current_area(void) const;

		 virtual void box_added_impl(void); // called immediately after box is added to tree
	 private:

		 // Eventually we'll have an abstract box class which allows for boxes which
		 // implement these differently. So they are private here.
		 point           frozen_location_; 
		 box            *next_sibling_; 
		 container_box  *parent_;
		 property_box   *properties_;
		 
		 // total size: 1 pointer offset plus 3 pointers plus two int32_t, or
       //             32-bit arch: 6*4     = 24 bytes;
       //             64-bit arch: 4*8+2*4 = 40 bytes;
 };

 class area {
	 public:
		 virtual ~area();
		 virtual float dpi(void) const;

		 // If you want access to the tree in a thread, you have to first acquire the lock.
		 void          lock(void);
		 bool          trylock(void);
		 void          unlock(void);
	 private:
		 thread_mutex   mutex_;
 };

 class reflow_manager {
	 public:
		 // There are two reasons why a box needs to recompute the sizes of the children: 1) because
		 // some of the children have changed size or 2) because the box itself has made a change in
		 // its absolute or fractional size.
		 //
		 // Any modification to a box is propagated upwards by setting the `recompute_children'
		 // flag of parent boxes (by calling mark_to_root()). This ensures that case 1) is taken
		 // care of. Case 2) is handled by setting the `propagate_changes' flag as well as the
		 // recompute_children one. Once a box encounters this flag true, it sets the dont_cache
		 // flag so as to prevent any children from using cached results.
		 //
		 // Children need to be refrozen if 1) frozen_size of parent does not match the
		 // newly specified size (which is essentially `propagate_changes' but we don't have
		 // to buffer this one) or 2) refreeze_children is set (which is also done by mark_to_root()).

		 void mark_recompute_children(const box *);   // for size
		 void mark_propagate_changes(const box *);    // for size
		 void mark_refreeze_children(const box *);    // for freeze
		 void mark_needs_redraw(const box *);         // currently unused

		 void unmark_recompute_children(const box *);
		 void unmark_propagate_changes(const box *);
		 void unmark_refreeze_children(const box *);
		 void unmark_needs_redraw(const box *);

		 bool check_recompute_children(const box *) const;
		 bool check_propagate_changes(const box *) const;
		 bool check_refreeze_children(const box *) const;
		 bool check_needs_redraw(const box *) const;
	 protected:
		 set<const box *> boxes_recompute_children_;
		 set<const box *> boxes_propagate_changes_;
		 set<const box *> boxes_refreeze_children_;
		 set<const box *> boxes_needing_redraw_;
 };

 class resizable_box : public box {
	 public:
		 resizable_box(void);

		 virtual point   frozen_size(void) const;
		 virtual int32_t shift(void) const;

		 // Requested natural sizes. Boxes _must_ be able to satisfy what they
		 // request here (ie. no promises that cannot be guaranteed). The size
		 // returned may be bigger than the available size though, in which case
		 // the parent will have to adjust, but the returned result should be
		 // as close to (hsize,vsize) as possible.
		 virtual point   size(int32_t hsize,  int32_t vsize, bool dont_cache, box *forced_end);

		 // Setting sizes, either in absolute units or in fractions of the available space.
		 // Zeroes mean `wrap content'. Only call on boxes that have already been added to
		 // the tree.
		 virtual void    set_absolute_size(int32_t hsize, int32_t vsize, int32_t sh);
		 virtual void    set_size_fraction(float hsize, float vsize);
		 virtual void    set_size_mixed(float hsize, int32_t vsize);
		 virtual void    set_size_mixed(int32_t hsize, float vsize);

		 // Note: these sizes should be taken as absolute!!! Do not compute percentages of these!!!
		 // This member always sets the frozen_size equal to w,h.
		 virtual void    freeze(int32_t w, int32_t h, box *forced_end); 
	 protected:
		 // FIXME: we can use eg. negative numbers to denote percentages, and save a byte.
		 mutable struct boxflags {
				 bool width_is_percentage_        : 1;  
				 bool height_is_percentage_       : 1;
		 } hwflags_;

		 mutable int32_t width_, height_, shift_;
		 mutable int32_t frozen_width_,frozen_height_;
 };

 class container_box : public resizable_box {
	 public:
		 container_box(void);
		 virtual ~container_box(void);

		 // Tree navigation.
		 child_iterator  begin(void) const;
		 inline box     *first_child(void) const;
		 inline box     *last_child(void) const;

		 virtual box    *add_child(box *bx);

		 // Flags and variables.
		 void            set_baseline_is_top(bool);
		 void            set_center_content(bool);
		 void            set_fill_background(bool);

		 virtual int     draw(point, const draw_info &, box *);
		 virtual int     position(point, const draw_info &, box *);
		 virtual int     propagate_event(point, event_info& , box *);

		 virtual void    set_absolute_size(int32_t hsize, int32_t vsize, int32_t sh);
		 virtual void    set_size_fraction(float hsize, float vsize);


		 // Relayout and refreeze flags.
		 inline bool     check_recompute_children(void) const;
		 inline bool     check_propagate_changes(void) const;
		 inline bool     check_refreeze_children(void) const;
							  
		 inline void     mark_recompute_children(void);
		 inline void     mark_propagate_changes(void);
		 inline void     mark_refreeze_children(void);
							  
		 inline void     unmark_recompute_children(void);
		 inline void     unmark_propagate_changes(void);
		 inline void     unmark_refreeze_children(void);
	 protected:
		 box            *first_child_;
		 box            *last_child_;
		 mutable point   children_size_;    // meaning depends on box type

		 virtual void fill_background_impl(point, const draw_info&);
		 struct {
				 bool         baseline_is_top_           : 1;
				 bool         center_content_            : 1;
				 bool         fill_background_           : 1;
				 bool         ignore_initial_ignorables_ : 1;
				 bool         alternative_children_int_  : 1; // tricky space-saving feature
				 bool         recompute_children_        : 1;
				 bool         propagate_changes_         : 1;
				 bool         refreeze_children_         : 1;
		 } cbox_flags_;
 };

 class glue : public resizable_box {
	 public:
		 glue(void);

		 // We override these because we can query the font manager for the width of a space.
		 virtual point   size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
	
		 virtual break_property break_info(void) const; 		 
 };

 class text_box : public box {
	 public:
		 text_box(const string& txt);

		 virtual point   frozen_size(void) const;
		 virtual int32_t shift(void) const;

		 break_property  break_info(void) const;
		 virtual string  info(void) const;

	 protected:
		 string           text_;
		 mutable int16_t  measured_width_; // measuring width is very time-consuming, so we buffer
  };

 class event_box : public container_box {
	 public:
		 virtual bool   ignore_geometry(void) const;
		 virtual bool   flat_hierarchy(void) const;
		 virtual bool   subtree_is_interactive(void) const;
		 virtual bool   accept_event(event_info& ev);
 };

 class table : public container_box {
	 public:
		 table(void);
		 enum valign_type { align_top, align_bottom, align_center };

		 // For each box, we store the cell positions. 
		 void     		  set_current_cell(uint16_t row, uint16_t col);
		 void     		  set_current_cell(uint16_t row, uint16_t col, valign_type);
		 void     		  empty_cell(void);
		 void     		  end_row(void);
					 		  
		 uint16_t   	  current_row(void) const;
		 uint16_t   	  current_column(void) const;

		 virtual point   size(int32_t hsize,  int32_t vsize, bool dont_cache, box *forced_end);
		 virtual void    freeze(int32_t, int32_t, box *forced_end);
		 // Boxes are drawn in the order in which they were added to the table, so you can
		 // have overlapping ones.
		 virtual int     draw(point, const draw_info &, box *);
		 virtual int     propagate_event(point, event_info& , box *);

		 virtual box    *add_child(box *, uint16_t colspan, uint16_t rowspan); 
		 virtual box    *add_child(box *, uint16_t colspan, uint16_t rowspan, valign_type); 
	 private:
		 mutable vector<int32_t> mincolwidths_;
		 mutable vector<int32_t> natcolwidths_;
		 mutable vector<int32_t> maxcolwidths_;
		 mutable vector<int32_t> rowascents_;
		 mutable vector<int32_t> rowdescents_;
		 
		 class cellinfo {
			 public:
				 uint16_t      row, column;
				 uint16_t      rowspan, colspan;
				 valign_type valign;
		 };
		 mutable map<box *, cellinfo> cells_;  // FIXME: should not be mutable, but [] is not const.
	 protected:
		 uint16_t cur_row_, cur_col_;
		 void compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
		 void add_cell_info_(box *, uint16_t, uint16_t);
		 void add_cell_info_(box *, uint16_t, uint16_t, valign_type);

		 int32_t maxtablewidth_;
 };

 class hbox : public container_box { 
	 public:
		 hbox(void);
		 virtual ~hbox();

		 virtual point   size(int32_t hsize,  int32_t vsize, bool dont_cache, box *forced_end);

		 virtual void    freeze(int32_t, int32_t, box *forced_end);
		 virtual int     draw(point, const draw_info &, box *forced_end);
		 virtual int     propagate_event(point, event_info& , box *forced_end);
	 protected:
		 void compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
 };

 class viewer_box : public hbox {
	 public:
		 viewer_box(void);
		 virtual ~viewer_box();

		 virtual layout::reflow_manager   *reflow_mgr(void) const;
		 virtual layout::margin_manager   *margin_mgr(void) const;
		 virtual layout::property_manager *property_mgr(void) const;
		 
	 protected:
		 layout::reflow_manager   *reflow_mgr_;
		 layout::margin_manager   *margin_mgr_;
		 layout::property_manager *property_mgr_;

		 point canvas_size_;
 };

 // Usage: While building the tree, marking floating boxes is done by
 // registering them with the margin_manager using
 // margin_mgr()->make_floating().  Subsequently, during 'freeze', the
 // box that encounters the floating box should mark its position by
 // calling margin_mgr()->mark_floater_position().  This vbox will
 // position it immediately following the child (in the vbox's list,
 // not in a sublist) from which the floater was exported.
 class margin_manager {
	 public:
		 ~margin_manager();

		 enum float_type { left_float, right_float };

		 void make_floating(box *, float_type);
		 bool is_floating(box *);
		 void mark_floater_position(box *);
	 protected:
		 void reset_floaters_(void);

		 class float_info {
			 public:
				 const box       *float_after_child;
				 enum float_type  ft;
		 };
		 map<box *, float_info *> floaters_;
		 // If float_after_child is null, the box has only been registered.
		 // If it points to the floater itself, this box has appeared in the 
		 // freeze of the most recent vbox child.
		 // Before a vbox freeze, reset these to null by calling reset_floaters_();
 };

 class vbox : public container_box, public margin_manager { 
	 public:
		 vbox(void);
		 virtual ~vbox();

		 virtual point   size(int32_t hsize,  int32_t vsize, bool dont_cache, box *forced_end);

		 virtual void            freeze(int32_t, int32_t, box *forced_end);
		 virtual int             draw(point, const draw_info &, box *forced_end);
		 virtual int             propagate_event(point, event_info& , box *);
		 virtual margin_manager *margin_mgr(void) const;
	 protected:
		 void compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
 };

 class border_box : public vbox {
	 public:
		 border_box(void);
		 enum border_style { dotted, dashed, solid, double_line, groove, ridge, inset, outset };

		 void         set_border_style(border_style);
		 border_style get_border_style(void) const;
	 private:
		 border_style style_;
 };

 class parbox : public container_box {
	 public:
		 parbox(void);
		 virtual ~parbox();

		 class breakpoint {
			 public:
				 breakpoint(box*, box*, long, int32_t, int32_t, long, long);
				 
				 box           *box_after_last_best_breakpoint;
				 box           *box_before_breakpoint;
				 long         line_number;
				 int32_t      natural_height;
				 int32_t      natural_width;
				 int32_t      natural_shift;
				 long         best_previous_breakpoint;
				 long         total_demerits;
		 };
		 
		 class linebox : public hbox {
			 public:
				 linebox(box *, box*, int32_t, int32_t, int32_t);
				 virtual ~linebox();
				 
				 virtual point size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
				 virtual void  freeze(int32_t, int32_t, box *forced_end);
				 virtual int   position(point, const draw_info &, box *forced_end);
				 virtual int   draw(point, const draw_info &, box *forced_end);
				 virtual int   propagate_event(point, event_info &, box *);
		 };

		 virtual box    *add_child(box *);  // adds to internal children, actually.

		 virtual point   size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end);
	
		 virtual bool   ignore_geometry(void) const;
		 virtual bool   flat_hierarchy(void) const;

		 virtual void   freeze(int32_t, int32_t, box *forced_end);
		 virtual int    draw(point, const draw_info &, box *);
		 virtual int    propagate_event(point, event_info&, box *);
	 protected:
		 virtual void   compute_breakpoints_(int32_t c_hsize, int32_t c_vsize);
		 virtual void   create_hvboxes_(void);
		 virtual void   destroy_hvboxes_(void);
		 child_iterator begin_internal(void) const;
		 child_iterator end_internal(void) const;
	 private:
		 vector<breakpoint> breakpoints_;
		 box               *first_internal_child_;
		 box               *last_internal_child_;
		 mutable int32_t    sum_width_, max_width_; // of internal children FIXME: eliminate one!
 };

void print_layout_tree(box *);



// Inline function definitions follow:

inline void box::freeze_location(const point& pnt)
	{
   frozen_location_=pnt;
	}

inline point box::frozen_location(void) const
	{
	return frozen_location_;
	}

};

#endif 
