// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Font related classes.

  Copyright 1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef fonts_hh_
#define fonts_hh_

#include <string>
#include <iostream.h>
#include <vector.h>
#include <map.h>
#include <layout/numbers.hh>

namespace layout {

 // classes defined here.
 class font_manager; 
 class tfm;
 class enc;
 class font;
 class math_symbol_font;
 class math_extension_font;
 
 class xlfd {  // X Logical Font Description
	 public:
		 xlfd(void);

		 enum weight_value { bold, medium };
		 enum slant_value  { oblique, upright };

		 xlfd(const string& f, weight_value w, slant_value s, length l);
		 xlfd(const string& found, const string& f, weight_value w, slant_value s, length p, 
				const string& enc);

		 string       foundry;
		 string       family;
		 weight_value weight;
		 slant_value  slant;
		 length       size;
		 string       encoding;

		 // first set the above variables, then call this one for the
		 // complete name
		 string xlfd_string(void) const;
	 protected:
 };

 class font_manager {
	 public:
		 virtual ~font_manager();

		 virtual font *get_font(const xlfd&);
		 // The string is a Unicode UTF-8 encoded string. The font
		 // object will take care of mapping the Unicode characters to
		 // font glyphs.
		 virtual void  draw_string(font *, const string& txt, int32_t x, int32_t y)=0;
	 protected:
		 virtual font *create_font_impl(const xlfd&)=0;
	 private:
		 map<string, font *> fonts_;
 };

 class font {
	 public:
		 font(const xlfd&);
		 virtual ~font();

		 string name(void) const;  // FIXME: deprecated
		 xlfd   xl(void) const;  
		 
		 // string measuring
		 virtual int32_t string_width(const string&) const;

		 // metrics
		 fixword designsize(void) const;
		 fixword slant(void) const;
		 virtual int32_t standard_ascent(void) const;
		 virtual int32_t standard_descent(void) const;
		 virtual int32_t space_width(void) const;
		 fixword space_stretch(void) const;
		 fixword space_shrink(void) const;
		 fixword x_height(void) const;
		 fixword quad(void) const;
		 fixword extra_space(void) const;
		 
		 // attributes
		 bool     underline(void);
	 protected:
		 bool fill_tfm_info_(void);
		 bool fill_enc_info_(void);

		 xlfd               xl_;
		 tfm               *tfm_;
		 enc               *enc_;
		 mutable int32_t    standard_ascent_;
		 mutable int32_t    standard_descent_;
		 mutable int32_t    space_width_;
 };
 
/* class math_symbol_font : public font {
	 public:
		 // symbol metrics
		 fixword numer_shiftup_display(void);
		 fixword numer_shiftup_nondisplay_nonatop(void);
		 fixword numer_shiftup_nondisplay_atop(void);
		 fixword denom_shiftdown_display(void);
		 fixword denom_shiftdown_nondisplay(void);
		 fixword super_shiftup_display_uncramped(void);
		 fixword super_shiftup_nondisplay_uncramped(void);
		 fixword super_shiftup_cramped(void);
		 fixword sub_shiftdown_nosuper(void);
		 fixword sub_shiftdown_super(void);
		 fixword superline_belowtop(void);
		 fixword subline_belowbottom(void);
		 fixword delimiter_display(void);
		 fixword delimiter_nondisplay(void);
		 fixword axis_height(void);
	 protected:
		 draw_impl(mathbox *);
 };
 
 class math_extension_font : public font {
	 public:
		 // extensible symbol metrics
		 fixword default_rule_thickness(void);
		 fixword min_clearance_above_op(void);
		 fixword min_clearance_below_op(void);
		 fixword minlineskip_above_op(void);
		 fixword minlineskip_below_op(void);
		 fixword limits_padding(void);
		 
		 class mathextsymbol {
			 public:
				 top();
				 middle();
				 bottom();
				 repeater();
		 };
	 public:
		 draw_impl(vector<mathextsymbol>);
 };
*/

 class enc {
	 public:
		 enc(void);
		 void read(istream&);

		 uint8_t unicode2font(wchar_t);
	 private:
		 map<wchar_t, uint8_t>  u2f_;
 };

 class tfm {
	 public:
		 tfm(void);
		 
		 fixword checksum(void) const;
		 int16_t   tfmlength(void) const;
		 int16_t   min_char(void) const;
		 int16_t   max_char(void) const;
		 
		 fixword designsize(void) const;
		 fixword slant(void) const;
		 fixword space(void) const;
		 fixword space_stretch(void) const;
		 fixword space_shrink(void) const;
		 fixword x_height(void) const;
		 fixword quad(void) const;
		 fixword extra_space(void) const;
		 
		 class char_word_info {
			 public:
				 char_word_info(void);
				 char_word_info(int32_t);
				 int width_index(void) const;
				 int height_index(void) const;
				 int depth_index(void) const;
				 int italic_index(void) const;
				 
				 enum tagtype { no_tag=0, lig_tag, list_tag, ext_tag };
				 tagtype tag(void) const;
				 uint8_t remainder(void) const;
			 private:
				 int32_t raw_;
		 };
		 class ligkern_command {
			 public:
				 int8_t skip_int8_t;
				 int8_t next_char;
				 int8_t op_int8_t;
				 int8_t remainder;
		 };
		 class extensible_recipe {
			 public:
				 int8_t top;
				int8_t mid;
				int8_t bot;
				int8_t rep;
		};
		const vector<fixword>& header(void) const;
		const vector<char_word_info>& charword(void) const;
		const vector<fixword>& widths(void) const;
		const vector<fixword>& heights(void) const;
		const vector<fixword>& depths(void) const;
		const vector<fixword>& itcorr(void) const;
		const vector<ligkern_command>& ligkern(void) const;
		const vector<fixword>& kern(void) const;
		const vector<extensible_recipe>& exten(void) const;
		const vector<fixword>& param(void) const;
		
		void read(istream&);
   private:
		int16_t r16(istream&,int16_t *) const;
		int16_t r16(istream& str) const;
		int32_t r32(istream& str) const;
		
		int16_t filelength_;
		int16_t min_char_,max_char_;
		
		vector<fixword> header_;
		vector<char_word_info> charword_;
		vector<fixword> widths_;
		vector<fixword> heights_;
		vector<fixword> depths_;
		vector<fixword> itcorr_;
		vector<ligkern_command> ligkern_;
		vector<fixword> kern_;
		vector<extensible_recipe> exten_;
		vector<fixword> param_;
};

/*

You could maybe also argue that we still need a box type that has
everything virtual and eliminate the 

	 protected:
		 mutable int32_t  frozen_width_,frozen_height_;
		 point            frozen_location_; 
		 mutable int32_t  shift_;

Then we can make the mathematics a cbox, but also make that one 
virtual such that the base class does not know about the specific
way in which children are stored.

Better yet, since we anyhow do a dynamic_cast<cbox *> now on 
every box, we may want to get rid of cbox altogether. The 
only real difference is the freeze_children member, and modifying that
could go hand in hand with a better way to determine requested size. 
Eg.

	int32_t min_width(c_hsize);
   int32_t max_width(c_hsize); etc


   class font { 
		public:
			virtual draw_location draw(const string&, point, const draw_info &, box *);		 
or maybe
			virtual draw_location draw(const text *, point, const draw_info &, box *);
	};

// In plain TeX:
// family 0: normal text
// family 1: math italic letters
// family 2: ordinary math symbols
// family 3: large symbols
// family 4: text italic
// family 5: slanted roman
// family 6: bold roman
// family 7: typewriter
class font_family {
	public:
		textfont();
		scriptfont();
		scriptscriptfont();
};

// do we want this?
class font_group {
	public:
		font_family text_font();
		font_family math_symbol_font();
		font_family math_extension_font();
};


 \oct{000}, \oct{020}, \oct{022}, and \oct{040}, followed
by the extensible left parenthesis, which is $\oct{060}+[\oct{102}]^n+
\oct{100}$.

TeX does need to have the character codes for the small, and big variants, 
plus something that addresses the extensible version. I guess the latter
could point to any of the characters mentioned in the extensible recipe. 
HOW does TeX do it? \delcode does not seem enough:

\delcode`x="123456

means that if the letter |x| is used as a delimiter, its small variant
is found in position \hex{23} of family~1, and its large variant is found
in position \hex{56} of family~4.




*/


};

#endif
