// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Output device independent layout algorithms, implementation.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdexcept>
#include "layout/boxes.hh"

#define DEBUG_CLASS "LAYOUT"
#include <mnemonic/debug.hh>
#include <mnemonic/stopwatch.hh>

#include <typeinfo>
#include <strstream.h>
#include <algo.h>

#include <layout/fonts.hh>
#include <layout/colors.hh>

using namespace layout;


void layout::print_layout_tree(layout::box *start) 
	{
	long number=0;
	unsigned int level=0;
	container_box *bb=0;
	ostrstream output;
	output << endl;

	while(start) {
		++number;
		for(unsigned int i=0; i<level; ++i)
			output << " ";
		output << typeid(*start).name() << " " << start << " " 
				 << start->frozen_size() 
				 << "," << start->shift() << " @" 
				 << start->frozen_location().x << "," 
				 << start->frozen_location().y << " " 
				 << start->info() << " " << (int)start->break_info() << endl;
		bb=dynamic_cast<container_box *>(start);
		if(bb && bb->first_child()) {
			++level;
			start=bb->first_child();
			}
		else {
			while(start->next_sibling()==0) {
				start=start->parent();
				if(start==0)
					break;
					--level;
				}
			if(start) {
				start=start->next_sibling();
				}
			}
		}
	output << ends;
	debug(output.str());
	debug("total number of boxes: " << number);
	delete [] output.str();
 	}

// property_manager

property_manager::~property_manager()
	{
   set<property_box *>::iterator it=properties_.begin();
	while(it!=properties_.end()) {
		delete (*it);
		++it;
		}
	}

void property_manager::register_property(property_box *pb)
	{
	properties_.insert(pb);
	}

// property_box

property_box::property_box(property_box *inherit, font *fnt=0, color *foreground=0, color *background=0)
	: font_(fnt), fg_color_(foreground), bg_color_(background), inherit_(inherit)
	{
	assert(!(inherit==0 && (fnt==0 || foreground==0 || background==0)));
	}

font *property_box::current_font(void) const
	{
	if(font_)         return font_;
	else if(inherit_) return inherit_->current_font();
	else              return 0;
	}

color *property_box::current_foreground_color(void) const
	{
	if(fg_color_)     return fg_color_;
	else if(inherit_) return inherit_->current_foreground_color();
	else              return 0;
	}

color *property_box::current_background_color(void) const
	{
	if(bg_color_)     return bg_color_;
	else if(inherit_) return inherit_->current_background_color();
	else              return 0;
	}

property_box *property_box::inherit_from(void) const
	{
	return inherit_;
	}

void property_box::set_inherit(property_box *bx)
	{
	inherit_=bx;
	}

// box

box::event_info::event_info(void)
	: accepting_box(0), hovering_box(0)
	{
	}

box::child_iterator::child_iterator(box *start)
	: cangoup_(false), current_node_(start)
	{
	if(current_node_!=0) start_node_=current_node_->parent();
	else                 start_node_=0;
	}

box::child_iterator::child_iterator(box *start, bool cangoup)
	: cangoup_(cangoup), current_node_(start)
	{
	if(current_node_!=0) start_node_=current_node_->parent();
	else                 start_node_=0;
	}

void box::child_iterator::operator=(const box::child_iterator& other)
	{
	current_node_=other.current_node_;
	start_node_=other.start_node_;
	}

bool box::child_iterator::operator==(const box::child_iterator& other) const
	{
	if(current_node_==other.current_node_) 
		return true;
	else
		return false;
	}

box::child_iterator& box::child_iterator::operator++(void)
	{
	if(current_node_->flat_hierarchy()) {
		container_box *bb=dynamic_cast<container_box *>(current_node_);
		if(bb && bb->first_child()) {
			current_node_=bb->first_child();
			return *this;
			}
		}
	while(current_node_->next_sibling()==0) {
		current_node_=current_node_->parent();
		if(current_node_==0 || (cangoup_==false && current_node_==start_node_)) {
			current_node_=0;
			return *this; // end of iteration
			}
		}
	current_node_=current_node_->next_sibling();

	return *this;
	}

box* box::child_iterator::operator*(void)
	{
	return current_node_;
	}



// box

box::box(void)
	: next_sibling_(0), parent_(0), properties_(0)
   {
   }

box::~box()
   {
   }

point box::frozen_size(void) const
   {
   return point(0,0);
   }

int32_t box::shift(void) const
	{
	return 0;
	}

box::baseline_dir box::baseline(void) const
	{
	return horizontal_baseline;
	}

void box::mark_to_root(void)
	{
//	reflow_manager *rm=reflow_mgr();
//	assert(rm!=0);
	container_box *goback=this->parent();
	while(goback) {
//		if(goback->check_recompute_children())
//			break;
		goback->mark_recompute_children();
		goback->mark_refreeze_children();
		goback=goback->parent();
		}
	}

void box::set_parent(container_box *p)
	{
	parent_=p;
	}

void box::set_next_sibling(box* s)
	{
	next_sibling_=s;
	}

void box::set_properties(property_box *pb)
	{
	properties_=pb;
	assert(property_mgr()!=0);
	property_mgr()->register_property(pb);
	}

box::break_property box::break_info(void) const
	{
	return box::no_break;
	}

bool box::subtree_is_interactive(void) const
	{
	if(parent()) return parent()->subtree_is_interactive();
	else return 0;
	}

bool box::ignore_geometry(void) const
	{
	return false;
	}

bool box::flat_hierarchy(void) const
	{
	return false;
	}

string box::info(void) const
	{
	return "";
	}

int box::am_outside(const point& loc, const box::draw_info & inf) const
	{
	int yrel=-1;
	int xrel=-1;
	if(loc.x+frozen_location().x+frozen_size().x > inf.viewport_loc.x) {
		++xrel;
		if(loc.x+frozen_location().x > inf.viewport_botright.x) {
			++xrel;
			}
		}

	if(loc.y+frozen_location().y+shift() >= inf.viewport_loc.y) {
		++yrel;
		if(loc.y+frozen_location().y-frozen_size().y+shift() > inf.viewport_botright.y) {
			++yrel;
			}
		}
	
	int result=5+xrel+3*yrel;
	return result;
	}

int box::point_inside(point off, point pt) const
	{
	if(flat_hierarchy() || ignore_geometry()) return 8;

	int yrel=-1;
	int xrel=-1;
	if(pt.y>=frozen_location().y+off.y-frozen_size().y+shift()) {
		++yrel;
		if(pt.y>=frozen_location().y+shift()+off.y) {
			++yrel;
			}
		}

	if(pt.x>=frozen_location().x+off.x) {
		++xrel;
		if(pt.x>=frozen_location().x+frozen_size().x+off.x) {
			++xrel;
			}
		}
	
	int result=5+xrel+3*yrel;
	return result;
	}

int box::draw(point loc, const box::draw_info &inf, box *)
	{
	// We definitely do not want to test every Mickey mouse box for its location.
	return 5;
	}

void box::draw_outline(point loc, const box::draw_info &inf)
	{
	}

int box::position(point loc, const box::draw_info& inf, box *)
	{
	return 5;
	}

int box::propagate_event(point off, event_info& ev, box *)
	{
/*	debug("box:  " << frozen_location_.x << "," << frozen_location_.y);
	debug("offset:    " << off.x << "," << off.y);
	debug("point:     " << ev.hitpoint.x << "," << ev.hitpoint.y);
	debug("size:      " << frozen_width_ << "," << frozen_height_ << "," << shift_); */
	int tmp=point_inside(off,ev.hitpoint);
	if(tmp==5) {
		if(accept_event(ev))
			return 10;
		return 5;
		}
	else return tmp;
	}

bool box::accept_event(event_info& ev)
	{
	if(parent()) return parent()->accept_event(ev);
	else              return false;
	}

void box::reflow(void)
	{
	if(parent()) 
		parent()->reflow();
	}

void box::redraw(void)
	{
	redraw(rectangle(frozen_location(), frozen_size()));
	}

void box::redraw(rectangle rt)
	{
	rt.top_left.x+=frozen_location().x;
	rt.top_left.y+=frozen_location().y;
	if(parent()) 
		parent()->redraw(rt);
	}

container_box *box::parent(void) const
	{
	return parent_;
	}

property_box *box::properties(void) const
	{
	if(properties_)  return properties_;
	else if(parent_) return parent_->properties();
	else return 0;
	}

box *box::next_sibling(void) const
	{
	return next_sibling_;
	}

font_manager& box::font_mgr(void)
	{
	return parent()->font_mgr();
	}

color_manager& box::color_mgr(void)
	{
	return parent()->color_mgr();
	}

margin_manager *box::margin_mgr(void) const
	{
	if(parent()) return parent()->margin_mgr();
	else return 0;
	}

reflow_manager *box::reflow_mgr(void) const
	{
	if(parent()) return parent()->reflow_mgr();
	else return 0;
	}

property_manager *box::property_mgr(void) const
	{
	if(parent()) return parent()->property_mgr();
	else return 0;
	}

area *box::current_area(void) const
	{
	if(parent()) return parent()->current_area();
	else return 0;
	}

void box::box_added_impl(void)
	{
	}

// area

area::~area()
	{
	}

float area::dpi(void) const
	{
	return 95.0; // just a sensible default
	}

void area::lock(void)
	{
	mutex_.lock();
	}

bool area::trylock(void)
	{
	return mutex_.trylock();
	}

void area::unlock(void)
	{
	mutex_.unlock();
	}

// reflow_manager

void reflow_manager::mark_recompute_children(const box *bx)
	{
	boxes_recompute_children_.insert(bx);
	}

void reflow_manager::mark_propagate_changes(const box *bx)
	{
	boxes_propagate_changes_.insert(bx);
	}

void reflow_manager::mark_refreeze_children(const box *bx)
	{
	boxes_refreeze_children_.insert(bx);
	}

void reflow_manager::mark_needs_redraw(const box *bx)
	{
	boxes_needing_redraw_.insert(bx);
	}


void reflow_manager::unmark_recompute_children(const box *bx)
	{
	boxes_recompute_children_.erase(bx);
	}

void reflow_manager::unmark_propagate_changes(const box *bx)
	{
	boxes_propagate_changes_.erase(bx);
	}

void reflow_manager::unmark_refreeze_children(const box *bx)
	{
	boxes_refreeze_children_.erase(bx);
	}

void reflow_manager::unmark_needs_redraw(const box *bx)
	{
	boxes_needing_redraw_.erase(bx);
	}


bool reflow_manager::check_recompute_children(const box *bx) const
	{
	return boxes_recompute_children_.count(bx);
	}

bool reflow_manager::check_propagate_changes(const box *bx) const
	{
	return boxes_propagate_changes_.count(bx);
	}

bool reflow_manager::check_refreeze_children(const box *bx) const
	{
	return boxes_refreeze_children_.count(bx);
	}

bool reflow_manager::check_needs_redraw(const box *bx) const
	{
	return boxes_needing_redraw_.count(bx);
	}


// margin_manager

margin_manager::~margin_manager()
	{
	map<box *, float_info *>::const_iterator it=floaters_.begin();
	while(it!=floaters_.end()) {
		delete (*it).second;
		++it;
		}
	}

void margin_manager::make_floating(box *bx, float_type ft)
	{
	float_info *tmp=new float_info;
	tmp->float_after_child=0;
	tmp->ft=ft;
	floaters_[bx]=tmp;
	}

bool margin_manager::is_floating(box *bx)
	{
	map<box *, float_info *>::const_iterator it=floaters_.find(bx);
	if(it!=floaters_.end()) return true;
	return false;
	}

void margin_manager::mark_floater_position(box *bx)
	{
	map<box *, float_info *>::const_iterator it=floaters_.find(bx);
	if(it!=floaters_.end()) {
		assert(1==0);
		}
	(*it).second->float_after_child=bx;
	}

void margin_manager::reset_floaters_(void)
	{
	map<box *, float_info *>::const_iterator it=floaters_.begin();
	while(it!=floaters_.end()) {
		(*it).second->float_after_child=0;
		++it;
		}
	
	}

// resizable_box

resizable_box::resizable_box(void) 
	: width_(0), height_(0), shift_(0), frozen_width_(0), frozen_height_(0)
	{
	hwflags_.width_is_percentage_=false;
	hwflags_.height_is_percentage_=false;
	}

point resizable_box::frozen_size(void) const
	{
	return point(frozen_width_, frozen_height_);
	}

int32_t resizable_box::shift(void) const
	{
	return shift_;
	}

point resizable_box::size(int32_t hsize, int32_t vsize, bool, box *)
   {
	point tmp;
	if(hwflags_.width_is_percentage_)  tmp.x=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
	else                               tmp.x=width_;
	if(hwflags_.height_is_percentage_) tmp.y=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
	else                               tmp.y=height_;
	return tmp;
   }

void resizable_box::set_absolute_size(int32_t w, int32_t h, int32_t sh)
	{
	hwflags_.height_is_percentage_=false;
	hwflags_.width_is_percentage_=false;
	width_=w;
	height_=h;
	shift_=sh;
	}

void resizable_box::set_size_fraction(float hsize, float vsize)
	{
	assert(hsize<=1.0);
	assert(vsize<=1.0);
	hwflags_.width_is_percentage_=true;
	width_=(int32_t)(hsize*PERCENTAGE_MULTIPLIER);
	hwflags_.height_is_percentage_=true;
	height_=(int32_t)(vsize*PERCENTAGE_MULTIPLIER);
	}

void resizable_box::set_size_mixed(float hsize, int32_t vsize)
	{
	assert(hsize<=1.0);
	hwflags_.height_is_percentage_=false;
	hwflags_.width_is_percentage_=true;
	width_=(int32_t)(hsize*PERCENTAGE_MULTIPLIER);
	height_=vsize;
	shift_=0;
	}

void resizable_box::set_size_mixed(int32_t hsize, float vsize)
	{
	assert(vsize<=1.0);
	hwflags_.height_is_percentage_=true;
	hwflags_.width_is_percentage_=false;
	height_=(int32_t)(vsize*PERCENTAGE_MULTIPLIER);
	width_=hsize;
	shift_=0;
	}

void resizable_box::freeze(int32_t w, int32_t h, box *)
	{
	frozen_width_=w;
	frozen_height_=h;
	}



// container_box

container_box::container_box(void)
	: first_child_(0), last_child_(0), children_size_(0,0)
	{
	cbox_flags_.baseline_is_top_=false;
	cbox_flags_.center_content_=false;
	cbox_flags_.fill_background_=false;
	cbox_flags_.ignore_initial_ignorables_=false;
	cbox_flags_.alternative_children_int_=false;
	cbox_flags_.recompute_children_=false;
	cbox_flags_.propagate_changes_=false;
	cbox_flags_.refreeze_children_=false;
	}

container_box::~container_box()
	{
	box *currentnode=first_child_;
	box *prevnode;
	while(currentnode!=0) {
		prevnode=currentnode;
		currentnode=currentnode->next_sibling();
		delete prevnode;
		}
	first_child_=0;
	last_child_=0;
	}

box::child_iterator container_box::begin(void) const
	{
	return child_iterator(first_child_);
	}

box* container_box::first_child(void) const
	{
	return first_child_;
	}

box* container_box::last_child(void) const
	{
	return last_child_;
	}

box *container_box::add_child(box *bx)
	{
	assert(bx!=0);
   bx->set_parent(this);
	mark_to_root();

	if(last_child_==0) first_child_=bx;
	else               last_child_->set_next_sibling(bx);
	last_child_=bx;

	mark_recompute_children();
	mark_refreeze_children();
	bx->box_added_impl();
	return bx;
	}

void container_box::set_baseline_is_top(bool onoff)
	{
	cbox_flags_.baseline_is_top_=onoff;
	}

void container_box::set_center_content(bool onoff)
	{
	cbox_flags_.center_content_=onoff;
	}

void container_box::set_fill_background(bool onoff)
	{
	cbox_flags_.fill_background_=onoff;
	}

int container_box::draw(point loc, const box::draw_info& inf, box *forced_end)
	{
	// This is always a good place to look for optimisation: it's in the inner loop.

	if(ignore_geometry() || frozen_size().y==0) return 5;

	int tmp=am_outside(loc, inf);
	if(tmp!=5 || flat_hierarchy()) return tmp;

	if(cbox_flags_.fill_background_) fill_background_impl(loc, inf);

	child_iterator it(begin());

	if((*it)==0) return 5;

	if(forced_end!=0) it.cangoup_=true;
	
	bool didanything=false;
	loc+=frozen_location();
	
	while((*it)!=0) {
		(*it)->draw(loc,inf,forced_end);
		// If you want to see the actual box outlines and baseline, uncomment this one.
		// else (*it)->draw_outline(loc,inf);
		didanything=true;
		if((*it)==forced_end) 
			break;
		++it;
		}
	
	if(didanything) return 5;
	else            return 1;
	}

int container_box::position(point loc, const box::draw_info& inf, box *forced_end)
	{
	if(flat_hierarchy()) return 5;

	child_iterator it=begin();

	while((*it)!=0) {
		(*it)->position(frozen_location()+loc,inf,forced_end);
		if((*it)==forced_end) break;
		++it;
		}

	return 5;
	}

int container_box::propagate_event(point off, event_info& et, box *forced_end)
	{
	// This is a generic one, which does not take child ordering into account.
	// Optimised versions for hbox and vbox are below.

	if(ignore_geometry() || frozen_size().y==0) return 9;

	if(et.hitpoint.y>off.y+frozen_location().y+shift_) {
		return 8;
		}
	if(et.hitpoint.y<off.y+frozen_location().y-frozen_height_+shift_) {
		return 2;
		}

	// FIXME: for _any_ container box (also vbox and hbox, need to be modified!)
	// it is necessary to do an accept_event if no hit was detected for any
	// of the children. Not just at the beginning like this hack here (which was
	// introduced to make images respond to events).
	child_iterator it=begin();
	if((*it)==0) {
		if(accept_event(et))
			return 10;
		else 
			return 8;
		}

	if(forced_end!=0) it.cangoup_=true;

	int lastd=8;
	while((*it)!=0) {
		switch((lastd=(*it)->propagate_event(frozen_location()+off,et,forced_end))) {
			case 10:
				return 10;
			case 1:
			case 2:
			case 3:
				return lastd;
			default:
				break;
			}
		if((*it)==forced_end) 
			break;
		++it;
		}

	return 8;
	}

void container_box::fill_background_impl(point, const draw_info&)
	{
	}

void container_box::set_absolute_size(int32_t w, int32_t h, int32_t sh)
	{
	if(height_!=h || width_!=w || hwflags_.height_is_percentage_ || hwflags_.width_is_percentage_) {
		mark_propagate_changes();
		// FIXME: this implies that you cannot set the shift separately. As shift_
		// is used in a sort of weird way (it is not just a preference, but a real
		// computed value) this is a good thing to do.
		resizable_box::set_absolute_size(w,h,sh);
		}
	}

void container_box::set_size_fraction(float hsize, float vsize)
	{
	if(width_!=hsize || height_!=vsize || !hwflags_.height_is_percentage_ || !hwflags_.width_is_percentage_) {
		mark_propagate_changes();
		}
	resizable_box::set_size_fraction(hsize,vsize);
	}

bool container_box::check_recompute_children(void) const
	{
	return cbox_flags_.recompute_children_;
	}

bool container_box::check_propagate_changes(void) const
	{
	return cbox_flags_.propagate_changes_;
	}

bool container_box::check_refreeze_children(void) const
	{
	return cbox_flags_.refreeze_children_;
	}

void container_box::mark_recompute_children(void)
	{
	cbox_flags_.recompute_children_=true;
	}

void container_box::mark_propagate_changes(void)
	{
	cbox_flags_.propagate_changes_=true;
	}

void container_box::mark_refreeze_children(void)
	{
	cbox_flags_.refreeze_children_=true;
	}

void container_box::unmark_recompute_children(void)
	{
	cbox_flags_.recompute_children_=false;
	}

void container_box::unmark_propagate_changes(void)
	{
	cbox_flags_.propagate_changes_=false;
	}

void container_box::unmark_refreeze_children(void)
	{
	cbox_flags_.refreeze_children_=false;
	}


// glue

glue::glue(void) 
	{
	width_=-1;
	height_=-1;
	}

point glue::size(int32_t hsize, int32_t vsize, bool, box *)
	{
	point tmp;
	if(width_==-1) {
		assert(parent()!=0);
		tmp.x=properties()->current_font()->space_width();
		width_=tmp.x;
		}
	else tmp.x=width_;

	if(height_==-1) {
		assert(parent()!=0);
		tmp.y=properties()->current_font()->standard_ascent();
		height_=tmp.y;
		}
	else tmp.y=height_;

	return tmp;
	}

box::break_property glue::break_info(void) const
   {
   return optional_break_discard;
   }



// text_box

text_box::text_box(const string& str)
	: text_(str), measured_width_(-1)
	{
	}

point text_box::frozen_size(void) const
   {
	if(measured_width_==-1)
		measured_width_=properties()->current_font()->string_width(text_);
	return point(measured_width_,
					 properties()->current_font()->standard_ascent()
					 +properties()->current_font()->standard_descent());
   }

int32_t text_box::shift(void) const
   {
	return properties()->current_font()->standard_descent();
   }

box::break_property text_box::break_info(void) const
	{
	if(text_=="\x0a") return mandatory_break_discard;
	else              return no_break;
	}

string text_box::info(void) const
	{
	if(text_=="\x0a") return "[mandatory break]";
	else return text_;
	}



// event_box

bool event_box::ignore_geometry(void) const
	{
	return true;
	}

bool event_box::flat_hierarchy(void) const
	{
	return true;
	}

bool event_box::subtree_is_interactive(void) const
	{
	return true;
	}

bool event_box::accept_event(event_info& et)
	{		
	if(et.etype==event_info::button_down) et.accepting_box=this;
	else {
		int button=0;
		if(et.etype==event_info::button_up) {
			if(et.accepting_box==this) button=1;
			else                       return false;
			}
		else {
			if(et.hovering_box!=this) et.hovering_box=this;
			}
		}
	return true;
	}

// table

table::table(void)
	: container_box(), cur_row_(0), cur_col_(0)
	{
	}

void table::set_current_cell(uint16_t row, uint16_t col)
	{
	cur_row_=row;
	cur_col_=col;
	}

void table::empty_cell(void)
	{
	++cur_col_;
	}

void table::end_row(void)
	{
	if(!(cur_row_==0 && cur_col_==0)) { // do nothing when we haven't even started yet
		cur_col_=0;
		assert(cur_row_<65535);
		++cur_row_;
		}
	}

uint16_t table::current_row(void) const
	{
	return cur_row_;
	}

uint16_t table::current_column(void) const
	{
	return cur_col_;
	}

point table::size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
	if(!dont_cache)
		if(check_propagate_changes())
			dont_cache=true;

	if((children_size_.x==0 && children_size_.y==0) || check_recompute_children() || dont_cache) {
		compute_natural_size_(hsize, vsize, dont_cache, forced_end);
	   mark_refreeze_children();
		}
	unmark_recompute_children();
	unmark_propagate_changes();
	
	point tmp;
	if(width_==0) {
		if(cbox_flags_.alternative_children_int_ && hsize==EPSILON) tmp.x=children_size_.y;
		else                                                        tmp.x=children_size_.x;
		}
	else {
		if(hwflags_.width_is_percentage_)  {
			if(hsize==INFINITY) tmp.x=0; // FIXME: because we do not know the real width, we have no choice.
			else {
				tmp.x=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
				if(cbox_flags_.alternative_children_int_) {
					if(hsize==EPSILON)                tmp.x=children_size_.y;
					else if(tmp.x<children_size_.y)   tmp.x=children_size_.y;
					}
				else {
					if(tmp.x<children_size_.x) tmp.x=children_size_.x;
					}
				}
			debug("table has percentage width " << tmp.x);
			}
		else {
			debug("table has fixed width " << width_);
			tmp.x=width_;
			if(cbox_flags_.alternative_children_int_) {
				if(hsize==EPSILON || tmp.x < children_size_.y) tmp.x=children_size_.y;
				}
			else {
				if(tmp.x<children_size_.x) tmp.x=children_size_.x;
				}
			}
		}
	if(height_==0) {
		if(cbox_flags_.alternative_children_int_) tmp.y=0;
		else                                      tmp.y=children_size_.y;
		}
	else {
		if(hwflags_.height_is_percentage_) tmp.y=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
		else                               tmp.y=height_;
		}
	return tmp;
	}

void table::compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
	debug("table::compute_natural_size_ " << hsize << "," << vsize);
	map<box *, cellinfo>::const_iterator cit=cells_.begin();
	int32_t number_of_columns=0;
	int32_t number_of_rows=0;
   while(cit!=cells_.end()) {
		number_of_columns=max(number_of_columns, (*cit).second.column+(*cit).second.colspan);
		number_of_rows   =max(number_of_rows,    (*cit).second.row+(*cit).second.rowspan);
		++cit;
		}
	mincolwidths_ =vector<int32_t>(number_of_columns,0);
	natcolwidths_ =vector<int32_t>(number_of_columns,0);
	maxcolwidths_ =vector<int32_t>(number_of_columns,0);
	rowascents_   =vector<int32_t>(number_of_rows,0);
	rowdescents_  =vector<int32_t>(number_of_rows,0);
	debug("table size: " << number_of_columns << "x" << number_of_rows);

	bool height_unspecified=false;
	child_iterator it=begin();
   while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			cellinfo tmp=cells_[(*it)];
			point    sizmin, siznat, sizmax;
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				siznat=res->size(hsize,vsize,dont_cache,forced_end);
				if(siznat.y==0) { // obtain minimal and maximal size too
					sizmin=res->size(EPSILON,0,false,forced_end);
					sizmax=res->size(INFINITY,0,false,forced_end);
					height_unspecified=true;
					}
				else {
					sizmin=siznat;
					sizmax=siznat;
					}
				}
			else {
				sizmin=(*it)->frozen_size();
				siznat=sizmin;
				sizmax=sizmin;
				}

			debug("table cell " << sizmin << "," << siznat << "," << sizmax << ", " << (*it)->shift());
			for(int i=0; i<tmp.colspan; ++i) {
				assert(tmp.column+i<number_of_columns);
				mincolwidths_[tmp.column+i]=max(mincolwidths_[tmp.column+i], 
														  (int32_t)((sizmin.x+0.5)/tmp.colspan));
				natcolwidths_[tmp.column+i]=max(natcolwidths_[tmp.column+i], 
														  (int32_t)((siznat.x+0.5)/tmp.colspan));
				maxcolwidths_[tmp.column+i]=max(maxcolwidths_[tmp.column+i], 
														  (int32_t)((sizmax.x+0.5)/tmp.colspan));
				}
			for(int i=0; i<tmp.rowspan; ++i) {
				assert(tmp.row+i<number_of_rows);
				rowascents_[tmp.row+i]=max(rowascents_[tmp.row+i],   
													(int32_t)((sizmin.y-(*it)->shift()+0.5)/tmp.rowspan));
				rowdescents_[tmp.row+i]=max(rowdescents_[tmp.row+i], 
													 (int32_t)((*it)->shift()/tmp.rowspan));
				}
			}
		if((*it)==forced_end) break;
		++it;
		}
	if(height_unspecified) { // ascents and descents mean nothing, so clear them for safety
		for(int i=0; i<number_of_rows; ++i) {
			rowascents_[i]=0;
		   rowdescents_[i]=0;
			}
		}
	int32_t mintablewidth=0, nattablewidth=0, maxtablewidth=0;
	for(int i=0; i<number_of_columns; ++i) {
		mintablewidth+=mincolwidths_[i];
		nattablewidth+=natcolwidths_[i];
		maxtablewidth+=maxcolwidths_[i];
		debug("column " << i << " min/nat/max " << mincolwidths_[i] 
				<< "," << natcolwidths_[i] << "," << maxcolwidths_[i]);
		}
	debug("table min/nat/max size: " << mintablewidth << ", " << nattablewidth << "," << maxtablewidth);
	cbox_flags_.alternative_children_int_=true;
	maxtablewidth_  =maxtablewidth;
	children_size_.x=nattablewidth;
	children_size_.y=mintablewidth;
	}

// How do we make sure that columns with background filled get filled all the
// way from top to bottom, rather than only in the area for which there is content?
// One solution would be to have a `second pass freeze', which can only be called
// on resizable boxes and should only be called in the direction in which the
// box has returned `0' from size(). For vbox-wrapping-parbox, this would just lead
// to a new vbox vertical size, no reflow. Other options (adding additional boxes etc)
// are problematic as they split the content of the background fill pattern into
// two boxes.

void table::freeze(int32_t hsize, int32_t vsize, box *forced_end)
   {
	if(hsize==frozen_width_ && hsize!=0 && (vsize==0 || vsize==frozen_height_)) {
		if(check_refreeze_children()==false) {
			return;
			}
		}
	unmark_refreeze_children();
	debug("table::freeze " << hsize << "," << vsize);
	if(*(begin())==0) return;

	int32_t mintablewidth=children_size_.y;
	int32_t nattablewidth=children_size_.x;
	int32_t maxtablewidth=maxtablewidth_;
	if(nattablewidth<hsize) {
		if(width_==0) {
			debug("table fits with natural cell widths");
			for(unsigned int i=0; i<mincolwidths_.size(); ++i) {
				mincolwidths_[i]=natcolwidths_[i];
				}
			}
		else {
			debug("table with natural cell widths is too small, stretching");
			if(maxtablewidth<hsize) {
				debug("even with maximal size cells, table is still too small");
				}
			int32_t diff=hsize-nattablewidth;
			debug("stretching cells to fill up content");
			debug("array size " << natcolwidths_.size());
			for(unsigned int i=0; i<natcolwidths_.size(); ++i) {
				debug("something " << i << " and " << mincolwidths_.size());
				mincolwidths_[i]=natcolwidths_[i]+(int32_t)((diff*1.0*natcolwidths_[i]+0.5)/nattablewidth);
				debug("column " << i << " now " << mincolwidths_[i]);
				}
			}
		}
	else if(mintablewidth>hsize) {
		debug("even minimal table does not fit");
		}
	else {
		int32_t diffwidth=nattablewidth-mintablewidth;
		if(diffwidth>0) {
			debug("stretching cells to fill up content");
			int32_t stretchwidth=hsize-mintablewidth;
			for(unsigned int i=0; i<mincolwidths_.size(); ++i) {
				debug("something " << i << " and " << natcolwidths_.size());
				int32_t diff=natcolwidths_[i]-mincolwidths_[i];
				mincolwidths_[i]+=(int32_t)((diff*1.0*stretchwidth+0.5)/diffwidth);
				debug("column " << i << " now " << mincolwidths_[i]);
				}
			}
		}
	int32_t new_height_=0;
	int32_t new_width_=0;
	for(unsigned int i=0; i<mincolwidths_.size(); ++i) {
		new_width_+=mincolwidths_[i];
		if(i>0) mincolwidths_[i]+=mincolwidths_[i-1];
		}
	debug("rowascent accumulate " << rowascents_.size());
	for(unsigned int i=0; i<rowascents_.size(); ++i) {
		new_height_+=rowascents_[i]+rowdescents_[i];
		if(i>0) {
			rowascents_[i]+=rowascents_[i-1];
			rowdescents_[i]+=rowdescents_[i-1];
			}
		}
	if(rowdescents_.size()>1) {
		debug("rowasc/descents:");
		for(unsigned int i=0; i<rowdescents_.size(); ++i) {
			debug(rowascents_[i] << "," << rowdescents_[i]);
			}
		// This will only produce something when there were no unspecified_height boxes.
		shift_=rowdescents_[rowdescents_.size()-1]-rowdescents_[rowdescents_.size()-2];
		}
	else {
		if(rowdescents_.size()==0) shift_=0;
		else                       shift_=rowdescents_[rowdescents_.size()-1];
		}


	// Freeze boxes to their proper size.
	debug("freeze boxes to proper size");
	child_iterator it=begin();
	debug("is it the iterator?");
// FIXME: is this correct? A table is never cut between two lines, so there
// should not be a need to go up. It does anyhow not produce the right result.
//	if(forced_end) it.cangoup_=true;
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			cellinfo tmp=cells_[(*it)];
			
			int32_t setwidth=mincolwidths_[tmp.column+tmp.colspan-1];
			int32_t setheight=rowascents_[tmp.row+tmp.rowspan-1]+rowdescents_[tmp.row+tmp.rowspan-1];
			debug("rowheight stuff: " << rowascents_[tmp.row+tmp.rowspan-1] << "," << 
					rowdescents_[tmp.row+tmp.rowspan-1] << "," << tmp.row+tmp.rowspan-1 << ","
					<< (int)tmp.row << "," << (int)tmp.rowspan);
			if(tmp.column>0)
				setwidth-=mincolwidths_[tmp.column-1];
			if(tmp.row>0) {
				setheight-=rowascents_[tmp.row-1];
				setheight-=rowdescents_[tmp.row-1];
				} 
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				debug("freezing table cell at " << setwidth << "," << setheight);
            // FIXME: have to second-pass freeze later for the height.
				res->freeze(setwidth,setheight, forced_end); 
				}
			}
		if((*it)==forced_end) break;
		++it;
		}
	// Fill ascent/descent arrays in case we have not done that yet.
	debug("fill ascent/descent arrays");
	if(vsize==0) {
		child_iterator it=begin();
		while((*it)!=0) {
			if(!((*it)->ignore_geometry())) {
				cellinfo tmp=cells_[(*it)];
				for(int i=0; i<tmp.rowspan; ++i) {
					rowascents_[tmp.row+i]=max(rowascents_[tmp.row+i],  
														(int32_t)(((*it)->frozen_size().y-(*it)->shift()+0.5)/tmp.rowspan));
					rowdescents_[tmp.row+i]=max(rowdescents_[tmp.row+i], 
														 (int32_t)(((*it)->shift()+0.5)/tmp.rowspan));
					}
				}
			if((*it)==forced_end) break;
			++it;
			}
		if(rowdescents_.size()>1)
			shift_=rowdescents_[rowdescents_.size()-1]-rowdescents_[rowdescents_.size()-2];
		else {
			if(rowdescents_.size()==0) shift_=0;
			else                       shift_=rowdescents_[rowdescents_.size()-1];
			}
		}

//	Compute the correct ascent/descents and total size.
	debug("compute correct ascent/descent");
	new_height_=0;
	for(unsigned int i=0; i<rowascents_.size(); ++i) {
		debug("rowasc/desc " << rowascents_[i] << "," << rowdescents_[i]);
		new_height_+=rowascents_[i]+rowdescents_[i];
		if(i>0) {
			rowascents_[i]+=rowascents_[i-1];
			rowdescents_[i]+=rowdescents_[i-1];
			}
		}
	debug("total height " << new_height_);

	// Now freeze locations.
	it=begin();
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			cellinfo tmp=cells_[(*it)];
			point loc(0,-new_height_+rowascents_[0]+shift_);
			if(tmp.column>0) 
				loc.x=mincolwidths_[tmp.column-1];
			if(tmp.row>0)
				loc.y=-new_height_+shift_+rowascents_[tmp.row]+rowdescents_[tmp.row-1];
			
			if(tmp.valign==align_top) {
				loc.y+=(*it)->frozen_size().y-(*it)->shift()-rowascents_[tmp.row];
				if(tmp.row>0) loc.y+=rowascents_[tmp.row-1];
				}
			(*it)->freeze_location(loc);
			}
		if((*it)==forced_end) break;
		++it;
		}

	// Set our own sizes.
	frozen_width_ =hsize;
	if(vsize==0) frozen_height_=new_height_;
	else         frozen_height_=vsize;
	
	// Make sure we did everything right.
	if(hsize!=0) assert(frozen_width_ ==hsize);
   if(vsize!=0) assert(frozen_height_==vsize);

	debug("table frozen at size " << frozen_width_ << "," << frozen_height_ << "," << shift_);
	}

int table::draw(point loc, const box::draw_info& inf, box *forced_end)
	{
	// This is always a good place to look for optimisation: it's in the inner loop.

	int tmp=am_outside(loc, inf);
	if(tmp!=5) return tmp;

	child_iterator it(begin());

	if((*it)==0) return 5;

//	if(forced_end!=0) it.cangoup_=true;
	
	bool didanything=false;

	loc+=frozen_location();
	
	while((*it)!=0) {
		(*it)->draw(loc,inf,forced_end);
		didanything=true;
		if((*it)==forced_end) break;
		++it;
		}

	if(didanything) return 5;
	else            return 8;
	}

int table::propagate_event(point loc, event_info& ev, box *forced_end)
	{
	int lastd=point_inside(loc,ev.hitpoint);
	if(lastd!=5) return lastd;
	
	child_iterator it=begin();
//	if(forced_end!=0) it.cangoup_=true;

	while((*it)!=0) {
		lastd=(*it)->propagate_event(frozen_location()+loc,ev,forced_end);
		if(lastd==10) return 10;
		if(lastd==5) break;
/*		if(hvbox_flags_.layout_children_==horizontal) {
			if(lastd==1 || lastd==2 || lastd==4 || lastd==5) break;
			}
		else {
			if(lastd==1 || lastd==2 || lastd==4 || lastd==5) break;
			}
 */
		if((*it)==forced_end) break;	
		++it;
		}

	return 5;
	}

box* table::add_child(box *bx, uint16_t colspan, uint16_t rowspan)
	{
	return add_child(bx,colspan,rowspan,align_bottom);
	}

box* table::add_child(box *bx, uint16_t colspan, uint16_t rowspan, valign_type vt)
	{
	assert(bx!=0);
	// Any box can be used as a cell (but you should not use a
	// flat_hierarchy box as a cell, since that would effectively mean
	// that you are using more than one box as one cell, which _is_
	// illegal).
	assert(bx->flat_hierarchy()==false);
   bx->set_parent(this);
	mark_to_root();

	if(last_child_==0) first_child_=bx;
	else               last_child_->set_next_sibling(bx);

	last_child_=bx;

	add_cell_info_(bx,colspan, rowspan,vt);
	assert(cur_col_<65535);
	++cur_col_;
	mark_recompute_children();
	mark_refreeze_children();
	bx->box_added_impl();
	return bx;
	}

void table::add_cell_info_(box * bx, uint16_t colspan, uint16_t rowspan)
	{
	add_cell_info_(bx,colspan,rowspan,align_bottom);
	}

void table::add_cell_info_(box * bx, uint16_t colspan, uint16_t rowspan, valign_type vt)
	{
	debug("adding table cell");
	cellinfo tmp;
	tmp.row=cur_row_;
	tmp.column=cur_col_;
	tmp.rowspan=max((uint16_t)1,rowspan);
	tmp.colspan=max((uint16_t)1,colspan);
	tmp.valign=vt;
	cells_[bx]=tmp;
	}



// hbox

hbox::hbox(void)
   {
   }

hbox::~hbox()
   {
   }

point hbox::size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end) 
	{
	if(!dont_cache)
		if(check_propagate_changes())
			dont_cache=true;

	if((children_size_.x==0 && children_size_.y==0) || check_recompute_children() || dont_cache) {
		compute_natural_size_(hsize, vsize, dont_cache, forced_end);
		mark_refreeze_children();
		}
	unmark_recompute_children();
	unmark_propagate_changes();

	point tmp;
	if(width_==0) {
		if(hsize==EPSILON && cbox_flags_.alternative_children_int_) tmp.x=children_size_.y;
		else                                                        tmp.x=children_size_.x;
		}
	else {
		if(hwflags_.width_is_percentage_) {
			tmp.x=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
			if(cbox_flags_.alternative_children_int_) {
				if(tmp.x<children_size_.y) tmp.x=children_size_.y;
				}
			else {
				if(tmp.x<children_size_.x) tmp.x=children_size_.x;
				}
			}			
		else                               tmp.x=width_;
		}
	if(height_==0) {
		if(cbox_flags_.alternative_children_int_) tmp.y=0;
		else                                      tmp.y=children_size_.y;
		}
	else {
		if(hwflags_.height_is_percentage_) tmp.y=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
		else                               tmp.y=height_;
		}
	return tmp;
	}

void hbox::compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
//	debug("hbox::compute natural size " << dont_cache << " " << forced_end);
	child_iterator it=begin();		
	int32_t maxascent=0;
	int32_t maxdescent=0;
	int32_t sumwidth=0;
	int32_t minwidth=0;
	bool    height_unspecified=false;
	if(width_!=0) {
		if(hwflags_.width_is_percentage_)  hsize=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
		else                               hsize=width_;
		}
	if(height_!=0) {
		if(hwflags_.height_is_percentage_) vsize=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
		else                               vsize=height_;
		}
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				point tmp=res->size(hsize,vsize,dont_cache,forced_end);
				sumwidth     +=tmp.x;
				maxascent     =max(maxascent, tmp.y-res->shift());
				if(tmp.y==0) {
					height_unspecified=true;
					minwidth=max(minwidth,res->size(EPSILON,0,false,forced_end).x);
					}
				else {
					minwidth=max(minwidth,tmp.x);
					}
				}
			else {
				sumwidth     +=(*it)->frozen_size().x;
				maxascent     =max(maxascent, (*it)->frozen_size().y-(*it)->shift());
				assert((*it)->frozen_size().y!=0);
				}
			maxdescent =max(maxdescent, (*it)->shift());
			}
		if((*it)==forced_end) break;
		++it;
		}
	children_size_.x=sumwidth;
	if(height_unspecified) {
		cbox_flags_.alternative_children_int_=true;
		children_size_.y=minwidth;
		}
	else {
		cbox_flags_.alternative_children_int_=false;
		children_size_.y=maxascent+maxdescent;
		}
//	debug("hbox:compute_natural_size_ done " << children_size_);
	}

void hbox::freeze(int32_t hsize, int32_t vsize, box *forced_end)
   {
	if(*(begin())==0) return;
	if(hsize==frozen_width_ && hsize!=0 && (vsize==0 || vsize==frozen_height_)) {
		if(check_refreeze_children()==false) {
			return;
			}
		}
	unmark_refreeze_children();
//	debug("hbox::freeze " << hsize << "," << vsize);
	if(vsize==0 && cbox_flags_.alternative_children_int_) {
		if(hsize<children_size_.x && hsize>=children_size_.y) {
//			debug("need to adapt children size");
			}
		}

	// Freeze the sizes and determine the resulting baseline.
	int32_t maxascent=0, maxshift=0, maxwidth=0;
	int32_t sumheight=0, sumwidth=0, sumshift=0;
	child_iterator it=begin();
	if(forced_end) it.cangoup_=true;
	if(cbox_flags_.ignore_initial_ignorables_ && (*it)) {
		if((*it)->break_info()==optional_break_discard) {
			if((*it)==forced_end) return;
			++it;
			}
		}
	// First fix all the boxes that return a non-zero vsize.
	vector<int32_t> max_sizes;
	vector<int32_t> min_sizes;
	int32_t sum_max_sizes=0;
	int32_t sum_min_sizes=0;

	bool for_second_pass=false;
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				point tmp=res->size(hsize,vsize,false,forced_end);
				if(tmp.y!=0) 
					res->freeze(tmp.x, tmp.y, forced_end);
				else {
					for_second_pass=true;
					max_sizes.push_back(tmp.x);
					sum_max_sizes+=tmp.x;
					int32_t tt=res->size(EPSILON,0,false,forced_end).x;
				   min_sizes.push_back(tt);
//					debug("min/max " << tt << "," << tmp.x);
					sum_min_sizes+=tt;
					}
				}
			}
		if(!for_second_pass) {
			point tmp=(*it)->frozen_size();
			maxascent =max(maxascent, tmp.y-(*it)->shift());
			maxshift  =max(maxshift,  (*it)->shift());
			maxwidth  =max(maxwidth, tmp.x);
			sumheight+=tmp.y;
			sumwidth +=tmp.x;
			sumshift +=(*it)->shift();
			}
		else for_second_pass=false;
		if((*it)==forced_end) break;
		++it;
		}
	// Now fix all the others, using the table column algorithm.
	if(max_sizes.size()>0) {
		int32_t stretchwidth=hsize-sumwidth-sum_min_sizes;
		int32_t diffwidth=sum_max_sizes-sum_min_sizes;
//		debug("uh? " << hsize << "," << sumwidth << "," << sum_min_sizes << "," << sum_max_sizes);
//		debug("stretch parameters " << stretchwidth << "," << diffwidth);
		it=begin();
		if(cbox_flags_.ignore_initial_ignorables_ && (*it)) {
			if((*it)->break_info()==optional_break_discard) {
				++it;
				}
			}
		int colnum=0;
//		debug("hbox: freezing stretchable boxes");
		while((*it)!=0) {
			if(!((*it)->ignore_geometry())) {
				resizable_box *res=dynamic_cast<resizable_box *>((*it));
				if(res) {
					if(res->size(hsize,vsize,false,forced_end).y==0) {
//						debug(max_sizes[colnum] << "," << min_sizes[colnum]);
						int32_t setwidth=min_sizes[colnum];
						if(diffwidth>0) {
							setwidth+=
								(int32_t)(((max_sizes[colnum]-min_sizes[colnum])*1.0*stretchwidth+0.5)/diffwidth);
							}
						res->freeze(setwidth,0,forced_end);

						point tmp=(*it)->frozen_size();
						maxascent =max(maxascent, tmp.y-(*it)->shift());
						maxshift  =max(maxshift,  (*it)->shift());
						maxwidth  =max(maxwidth, tmp.x);
						sumheight+=tmp.y;
						sumwidth +=tmp.x;
						sumshift +=(*it)->shift();
						++colnum;
						}
					}
				}
			if((*it)==forced_end) break;
			++it;
			}
		}
	

   // Freeze locations.
	point current(0,0); 
	it=begin();
	if(cbox_flags_.baseline_is_top_) current.y=maxascent;
	while((*it)!=0) {
		if(!( (*it)->ignore_geometry())) {
//			debug("freezing location of box at " << current);
			(*it)->freeze_location(current);  // this is the baseline
			current.x+=(*it)->frozen_size().x;
			}
		if((*it)==forced_end) break;
		++it;
		}

	// Set our own sizes;
//	debug("hbox " << sumwidth << "," << hsize);
	if(hsize==0)  frozen_width_ =sumwidth;
	else          frozen_width_ =hsize;
	if(vsize==0)  frozen_height_=maxascent+maxshift;
	else          frozen_height_=vsize;

	// Make sure we did everything right.
	if(hsize!=0) assert(frozen_width_ ==hsize);
   if(vsize!=0) assert(frozen_height_==vsize);

	if(cbox_flags_.baseline_is_top_==true) shift_=frozen_height_;
	else                                   shift_=maxshift;
//	debug("hbox " << frozen_width_ << "," << frozen_height_ << "," << shift_);
   }

int hbox::draw(point loc, const box::draw_info& inf, box *forced_end)
	{
	if(ignore_geometry() || frozen_size().y==0) return 5;

	int tmp=am_outside(loc, inf);

	if(tmp!=5) return tmp;

	if(cbox_flags_.fill_background_) fill_background_impl(loc, inf);

	child_iterator it(begin());

	if((*it)==0) return 5;

	if(forced_end!=0) it.cangoup_=true;
	
	loc+=frozen_location();
	
	bool didanything=false;
	while((*it)!=0) {
		int ret=(*it)->draw(loc,inf,forced_end);
		if(ret==3 || ret==6 || ret==9) break;
		didanything=true;
		if((*it)==forced_end) break;
		++it;
		}

	if(didanything) return 5;
	else            return 6;
	}

int hbox::propagate_event(point loc, event_info& ev, box *forced_end)
	{
	if(ignore_geometry()) return 6;
/*	debug("hvbox: " << frozen_location().x << "," << frozen_location().y);
 	debug("offset:     " << loc.x << "," << loc.y);
 	debug("point:      " << ev.hitpoint.x << "," << ev.hitpoint.y);
 	debug("size:       " << frozen_width_ << "," << frozen_height_ << "," << shift_); */
	int lastd=point_inside(loc,ev.hitpoint);
	if(lastd!=5) return lastd;
	
//	debug("hvbox::propagate_event");
	child_iterator it=begin();
	if(forced_end!=0) it.cangoup_=true;

	while((*it)!=0) {
		lastd=(*it)->propagate_event(frozen_location()+loc,ev,forced_end);
		if(lastd==10) return 10;
		if(lastd==1 || lastd==2 || lastd==4 || lastd==5) break;
		if((*it)==forced_end) 
			break;
		++it;
		}

	return 5;
	}


// viewer_box

viewer_box::viewer_box(void)
	{
	cbox_flags_.baseline_is_top_=true;
	reflow_mgr_  =new layout::reflow_manager;
	margin_mgr_  =new layout::margin_manager;
	property_mgr_=new layout::property_manager;
	}

viewer_box::~viewer_box()
	{
	if(reflow_mgr_)   delete reflow_mgr_;
	if(margin_mgr_)   delete margin_mgr_;
	if(property_mgr_) delete property_mgr_;
	}

layout::reflow_manager *viewer_box::reflow_mgr(void) const
	{
	return reflow_mgr_;
	}

layout::margin_manager *viewer_box::margin_mgr(void) const
	{
	return margin_mgr_;
	}

layout::property_manager *viewer_box::property_mgr(void) const
	{
	return property_mgr_;
	}


// vbox

vbox::vbox(void)
   {
   }

vbox::~vbox()
   {
   }

point vbox::size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
	if(!dont_cache)
		if(reflow_mgr()->check_propagate_changes(this))
			dont_cache=true;

	if((children_size_.x==0 && children_size_.y==0) || check_recompute_children() || dont_cache) {
		compute_natural_size_(hsize, vsize, dont_cache, forced_end);
		mark_refreeze_children();
		}
	unmark_recompute_children();
	unmark_propagate_changes();

	point tmp;
	if(width_==0) {
		if(hsize==EPSILON && cbox_flags_.alternative_children_int_) tmp.x=children_size_.y;
		else if(hsize==INFINITY)                                    tmp.x=INFINITY;
		else                                                        tmp.x=children_size_.x;
		}
	else {
		if(hwflags_.width_is_percentage_)  {
			tmp.x=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
			if(cbox_flags_.alternative_children_int_) {
				if(hsize==EPSILON)              tmp.x=children_size_.y;
				else if(tmp.x<children_size_.y) tmp.x=children_size_.y;
				}
			else {
				if(tmp.x<children_size_.x) tmp.x=children_size_.x;
				}
			}
		else {
			tmp.x=width_;
			if(cbox_flags_.alternative_children_int_) {
				if(tmp.x<children_size_.y) tmp.x=children_size_.y;
				}
			else {
				if(tmp.x<children_size_.x) tmp.x=children_size_.x;
				}
			}
		}
	if(height_==0) {
		if(cbox_flags_.alternative_children_int_) tmp.y=0;
		else                                      tmp.y=children_size_.y;
		}
	else {
		if(hwflags_.height_is_percentage_) tmp.y=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
		else                               tmp.y=height_;
		}
	return tmp;
	}

void vbox::compute_natural_size_(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
//	debug("vbox::compute natural size " << dont_cache);
	int32_t sumheight=0;
	int32_t maxwidth=0;
	int32_t minwidth=0;
	int32_t lastshift=0;
	bool    height_unspecified=false;
	if(width_!=0) {
		if(hwflags_.width_is_percentage_)  hsize=(int32_t)(hsize*width_/PERCENTAGE_MULTIPLIER+0.5);
		else                               hsize=width_;
		}
	if(height_!=0) {
		if(hwflags_.height_is_percentage_) vsize=(int32_t)(vsize*height_/PERCENTAGE_MULTIPLIER+0.5);
		else                               vsize=height_;
		}
//	debug(typeid(*this).name() << "::compute_natural_size_ " << hsize << "," << vsize << " " 
//			<< static_cast<const layout::box *>(this));
	child_iterator it=begin();		
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				point tmp=res->size(hsize,vsize,dont_cache,forced_end);
				maxwidth     =max(maxwidth,tmp.x);
				sumheight   +=tmp.y;
				if(tmp.y==0) {
					height_unspecified=true;
					minwidth=max(minwidth,res->size(EPSILON,0,false,forced_end).x);
					}
				else {
					minwidth=max(minwidth,tmp.x);
					}
				}
			else {
				maxwidth     =max(maxwidth,(*it)->frozen_size().x);
				sumheight   +=(*it)->frozen_size().y;
				assert((*it)->frozen_size().y!=0);
				}
			lastshift=(*it)->shift();
//			debug("size now " << maxwidth << "," << sumheight << " " << static_cast<const layout::box *>(this));
			}
		if((*it)==forced_end) break;
		++it;
		}
	children_size_.x=maxwidth;
	if(height_unspecified) {
		cbox_flags_.alternative_children_int_=true;
		children_size_.y=minwidth;
		}
	else {
		cbox_flags_.alternative_children_int_=false;
		children_size_.y=sumheight;
		}
	if(cbox_flags_.baseline_is_top_==true) shift_=sumheight;
	else                                   shift_=lastshift;
	debug("vbox:compute_natural_size_ done " << children_size_ << " " << cbox_flags_.center_content_);
	}

void vbox::freeze(int32_t hsize, int32_t vsize, box *forced_end)
	{
	if(*(begin())==0) {
		frozen_width_ =hsize;
		frozen_height_=vsize;
		return;
		}
	if(hsize==frozen_width_ && hsize!=0 && (vsize==0 || vsize==frozen_height_)) {
		if(check_refreeze_children()==false) {
			return;
			}
		}
	unmark_refreeze_children();
	debug("vbox::freeze " << hsize << "," << vsize);

	// Freeze the sizes and determine the resulting baseline.
	bool firstrealbox=true;
	int32_t maxascent=0, firstascent=0, maxshift=0, maxwidth=0;
	int32_t sumheight=0, sumwidth=0, sumshift=0, lastshift=0;
	child_iterator it=begin();
// FIXME: another dubious one.
//	if(forced_end) it.cangoup_=true;
	while((*it)!=0) {
		if(!((*it)->ignore_geometry())) {
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) {
				point tmp0=res->size(EPSILON,vsize,false,forced_end);
				point tmp1=res->size(hsize,vsize,false,forced_end);
				point tmp2=res->size(INFINITY,vsize,false,forced_end);
				// FIXME: I'm not sure whether this covers all cases now.
				if(vsize==0) {
					if(!(cbox_flags_.center_content_ && tmp1.x<hsize)) {
						if(tmp0.x<tmp1.x && tmp1.x>hsize)      tmp1.x=hsize;
						else if(tmp1.x<hsize && tmp2.x>tmp1.x) tmp1.x=hsize; // stretch box if possible
//					tmp1.x=min(hsize,tmp1.x);  // otherwise size() will be maximum of child
						}
					}
				res->freeze(tmp1.x, tmp1.y, forced_end);
				}
			if(firstrealbox) { 
				if((*it)->frozen_size().y>0) {
					firstascent=(*it)->frozen_size().y-(*it)->shift();
					firstrealbox=false;
					}
				}
			point tmp=(*it)->frozen_size();
			maxascent =max(maxascent, tmp.y-(*it)->shift());
			maxshift  =max(maxshift,  (*it)->shift());
			maxwidth  =max(maxwidth, tmp.x);
			sumheight+=tmp.y;
			sumwidth +=tmp.x;
			sumshift +=(*it)->shift();
			lastshift=(*it)->shift();
			}
		if((*it)==forced_end) break;
		++it;
		}
	debug("max width " << maxwidth << ", maximal ascent " << maxascent << ", first ascent " << firstascent);

   // Freeze locations.
	point current(0,0); 
	it=begin();
	if(cbox_flags_.baseline_is_top_) current.y=0; 
	else                             current.y=-sumheight+lastshift;
	while((*it)!=0) {
		if(!( (*it)->ignore_geometry())) {
			if(cbox_flags_.center_content_) 
				current.x=(hsize-(*it)->frozen_size().x)/2;
			current.y+=(*it)->frozen_size().y-(*it)->shift();
			(*it)->freeze_location(current);  // this is the baseline
			current.y+=(*it)->shift();
			}
		if((*it)==forced_end) break;
		++it;
		}

	// Set our own sizes;
	if(hsize==0)   frozen_width_=maxwidth;
	else           frozen_width_=hsize;
	debug("our size set to " << frozen_width_);

	if(vsize==0)   frozen_height_=sumheight;
	else           frozen_height_=vsize;

	shift_=lastshift;

	// Make sure we did everything right.
	if(hsize!=0) assert(frozen_width_ ==hsize);
   if(vsize!=0) assert(frozen_height_==vsize);
	}

int vbox::draw(point loc, const box::draw_info& inf, box *forced_end)
	{
	if(ignore_geometry() || frozen_size().y==0) return 5;

	int tmp=am_outside(loc, inf);
	if(tmp!=5) return tmp;

	if(cbox_flags_.fill_background_) fill_background_impl(loc, inf);

	child_iterator it(begin());

	if((*it)==0) return 5;

//	if(forced_end!=0) it.cangoup_=true;
	
	loc+=frozen_location();
	
	bool didanything=false;
	while((*it)!=0) {
		int ret=(*it)->draw(loc,inf,forced_end);
		if(ret==7 || ret==8 || ret==9) break;
		didanything=true;
		if((*it)==forced_end) break;
		++it;
		}

	if(didanything) return 5;
	else            return 8;
	}

int vbox::propagate_event(point loc, event_info& ev, box *forced_end)
	{
	if(ignore_geometry()) return 6;
	int lastd=point_inside(loc,ev.hitpoint);
	if(lastd!=5) return lastd;
	
	child_iterator it=begin();
//	if(forced_end!=0) it.cangoup_=true;

	while((*it)!=0) {
		lastd=(*it)->propagate_event(frozen_location()+loc,ev,forced_end);
		if(lastd==10) return 10;
		if(lastd==1 || lastd==2 || lastd==4 || lastd==5) break;
		if((*it)==forced_end) 
			break;
		++it;
		}

	return 5;
	}

margin_manager *vbox::margin_mgr(void) const
	{
	const margin_manager *tmp=this;
	return const_cast<margin_manager *>(tmp);
	}



// border box

border_box::border_box(void)
	: vbox(), style_(inset)
	{
	}

void border_box::set_border_style(border_style bs)
	{
	style_=bs;
	}

border_box::border_style border_box::get_border_style(void) const
	{
	return style_;
	}


// parbox

parbox::parbox(void)
	: first_internal_child_(0), last_internal_child_(0), sum_width_(0), max_width_(0)
   {
	hwflags_.width_is_percentage_=true;
	width_=(int32_t)PERCENTAGE_MULTIPLIER;
   }

parbox::~parbox()
   {
   // If we still have hvboxes open, they will delete
	// our children. Otherwise (in which cast first_internal_child_!=0)
	// we delete them here.
	destroy_hvboxes_();

	box *currentnode=first_internal_child_;
	box *prevnode;
	while(currentnode!=0) {
		prevnode=currentnode;
		currentnode=currentnode->next_sibling();
		delete prevnode;
		} 
   }

box* parbox::add_child(box *bx)
	{
	assert(bx!=0);
	bx->set_parent(this);
	mark_to_root();

	if(last_internal_child_==0) first_internal_child_=bx;
	else                        last_internal_child_->set_next_sibling(bx);

	last_internal_child_=bx;
	mark_recompute_children();
	mark_refreeze_children();
	bx->box_added_impl();
	return bx;
	}

point parbox::size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
	// There is no point calculating height information as we do not know how we
	// will be frozen (boxes providing can_minimise can be frozen anywhere between
	// the size returned here and the minimal size).
	if(max_width_==0 || sum_width_==0 || check_recompute_children() || dont_cache) {
		destroy_hvboxes_();
		child_iterator it=begin_internal();
		max_width_=0;
		sum_width_=0;
		int32_t addwidth=0;
		vector<int32_t> sumwidths;
		while((*it)!=0 && it!=end_internal()) {
			resizable_box *res=dynamic_cast<resizable_box *>((*it));
			if(res) addwidth=res->size(hsize, vsize, true, forced_end).x;
			else    addwidth=(*it)->frozen_size().x;
			
			max_width_=max(max_width_,addwidth);
			sum_width_+=addwidth;
			if((*it)->break_info()==mandatory_break_discard) {
				sumwidths.push_back(sum_width_);
				sum_width_=0;
				}
			++it;
			}
		if(sumwidths.size()!=0) {
			sumwidths.push_back(sum_width_);
		   sum_width_=*(max_element(sumwidths.begin(),sumwidths.end()));
			}
		unmark_recompute_children();
		unmark_propagate_changes();
		}
	if(max_width_>hsize) {
		debug("there's a single overflowing box " << max_width_);
		return point(max_width_,0);
		}
	else {
		debug("parbox is min of " << sum_width_ << " and " << hsize);
		return point(min(sum_width_,hsize),0);
		}
	}


parbox::breakpoint::breakpoint(box* balb, box* bef, long cl, int32_t nath, int32_t natsh, 
										 long prev, long dem)
   : box_after_last_best_breakpoint(balb),
	  box_before_breakpoint(bef), line_number(cl), 
	  natural_height(nath), natural_width(0), natural_shift(0),
     best_previous_breakpoint(prev), total_demerits(dem)
   {
   }

parbox::linebox::linebox(box* fr, box* la, int32_t w, int32_t h, int32_t sh)
	{
	hwflags_.width_is_percentage_=false;
	hwflags_.height_is_percentage_=false;
	width_=0;
	height_=0;
	shift_=sh;
	children_size_.x=w;
	children_size_.y=h;
	first_child_=fr;
	last_child_=la;
	cbox_flags_.ignore_initial_ignorables_=true;
	}

parbox::linebox::~linebox()
	{
	// we do not own our children, so forget them! (no delete)
	first_child_=0;
	last_child_=0;
	}

point parbox::linebox::size(int32_t hsize, int32_t vsize, bool dont_cache, box *forced_end)
	{
	cbox_flags_.recompute_children_=false;
	cbox_flags_.propagate_changes_=false;
	dont_cache=false;
	return hbox::size(hsize,vsize,dont_cache,last_child_);
	}

void parbox::linebox::freeze(int32_t hsize, int32_t vsize, box *forced_end)
	{
	hbox::freeze(hsize, vsize, last_child_);
	}

int parbox::linebox::position(point loc, const box::draw_info& inf, box *forced_end)
	{
	return hbox::position(loc, inf, last_child_);
	}

int parbox::linebox::draw(point loc, const draw_info &inf, box *)
	{
	return hbox::draw(loc, inf, last_child_);
	}

int parbox::linebox::propagate_event(point loc, box::event_info& inf, box *forced_end)
	{
	return container_box::propagate_event(loc, inf, last_child_);
	}

bool parbox::ignore_geometry(void) const
	{
	return false;
	}

bool parbox::flat_hierarchy(void) const
	{
	return true;
	}

void parbox::freeze(int32_t hsize, int32_t vsize, box *forced_end)
   {
	debug("freezing parbox " << static_cast<box *>(this) << " " << hsize << "," << vsize);
	if(hsize==frozen_width_ && hsize!=0 && (vsize==0 || vsize==frozen_height_)) {
		if(check_refreeze_children()==false && breakpoints_.size()>0) {
			return;
			}
		} 
	unmark_refreeze_children();
	// FIXME: do something with forced_end
	destroy_hvboxes_();

	debug("computing breakpoints " << hsize << "," << vsize);
	compute_breakpoints_(hsize, vsize);
	debug("creating hvboxes");
	create_hvboxes_();

 	// we have thrown our children out, so we have zero size!
	// keep some size to force width for horizontal boxes.

	if(!hwflags_.width_is_percentage_) {
		if(width_!=0) frozen_width_=hsize;
		else {
         if(sum_width_>hsize) frozen_width_=hsize; // we promised hsize
			else                 frozen_width_=0;
			}
		}
	else frozen_width_=hsize;
	debug("parbox frozen width " << frozen_width_);
	frozen_height_=0;
   }

int parbox::draw(point loc, const box::draw_info &inf, box *fe)
	{
	return container_box::draw(loc, inf, fe);
	}

int parbox::propagate_event(point off, event_info& et, box *fe)
	{
	return container_box::propagate_event(off, et, fe);
	}

void parbox::compute_breakpoints_(int32_t c_hsize, int32_t c_vsize) 
	{
	point current(0,0);
   int32_t maxascent=0;
	int32_t maxdescent=0;

	if(begin_internal()==end_internal()) return;

 	breakpoint current_breakpoint(*(begin_internal()), *(begin_internal()), 0, 0, 0, 0, 0);
	breakpoints_.clear();

 	child_iterator it=begin_internal();
 	while((*it)!=0 && it!=end_internal()) {
		point thissize;
		resizable_box *res=dynamic_cast<resizable_box *>((*it));
		if(res) {
			thissize=res->size(c_hsize, c_vsize, false, 0); // FIXME: is the '0' forced_end really ok?
			}
		else    thissize=(*it)->frozen_size();

 		// FIXME: we need to get width from margin manager for every line separately.
		if(current_breakpoint.box_before_breakpoint!=(*it)) { // if equal, one box is already too much.
			if((*it)->break_info()==mandatory_break_discard || current.x+thissize.x > c_hsize) { 
				current_breakpoint.natural_height=maxascent+maxdescent;
				current_breakpoint.natural_width=current.x;
				current_breakpoint.natural_shift=maxdescent;
				current.x=0;
				current.y+=current_breakpoint.natural_height;
				maxascent=0;
				maxdescent=0;
				breakpoints_.push_back(current_breakpoint);
				current_breakpoint.box_after_last_best_breakpoint=(*it);
				if((*it)->break_info()==optional_break_discard ||
					(*it)->break_info()==mandatory_break_discard ) { 
					current.x=-thissize.x;
					}
				}
			}
		maxascent=max(thissize.y-(*it)->shift(), maxascent);
		maxdescent=max((*it)->shift(), maxdescent);
 		current.x+=thissize.x;
		current_breakpoint.box_before_breakpoint=(*it);
 		++it;
		}
	current_breakpoint.natural_height=maxascent+maxdescent;
	current_breakpoint.natural_shift=maxdescent;
	current_breakpoint.natural_width=current.x;
	breakpoints_.push_back(current_breakpoint);
	}

void parbox::create_hvboxes_(void)
	{
	if(breakpoints_.size()==0) return;

	for(unsigned long i=0; i<breakpoints_.size(); ++i) {
		linebox *tmph=new linebox(breakpoints_[i].box_after_last_best_breakpoint,
										  breakpoints_[i].box_before_breakpoint,
										  breakpoints_[i].natural_width,
										  breakpoints_[i].natural_height,
										  breakpoints_[i].natural_shift);
		// We could call add_child of the container_box, but that results in
		// this boxed being marked for re-calc of its natural size, which it
		// already knows.
		tmph->set_parent(this);
		mark_to_root();
		if(last_child_==0) first_child_=tmph;
		else               last_child_->set_next_sibling(tmph);
		last_child_=tmph;
//		mark_recompute_children();
//		mark_refreeze_children();
		}
	}

void parbox::destroy_hvboxes_(void)
	{
	if(breakpoints_.size()!=0) {
		box *currentnode=first_child_;
		box *prevnode;
		while(currentnode!=0) {
			prevnode=currentnode;
			currentnode=currentnode->next_sibling();
			delete prevnode;
			}
		}
	first_child_=0;
	last_child_=0;
	breakpoints_.clear();
	}

box::child_iterator parbox::begin_internal(void) const
	{
	return child_iterator(first_internal_child_);
	}

box::child_iterator parbox::end_internal(void) const
	{
	return child_iterator(0);
	}

