// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Font related classes.

  Copyright 1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <string>
#include <layout/fonts.hh>
#include <layout/numbers.hh>
#include <stdexcept>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <iomanip.h>
#include <strstream.h>

#define DEBUG_CLASS "FONTS"
#include <mnemonic/debug.hh>

using namespace layout;

xlfd::xlfd(void)
	: foundry("*"), family("helvetica"), weight(medium), slant(upright), size(10,length::pt), 
	  encoding("iso8859")
	{
	}

xlfd::xlfd(const string& f, weight_value w, slant_value s, length p)
	: foundry("*"), family(f), weight(w), slant(s), size(p), encoding("iso8859")
	{
	}

xlfd::xlfd(const string& found, const string& f, weight_value w, slant_value s, length p, const string& enc)
	: foundry(found), family(f), weight(w), slant(s), size(p), encoding(enc)
	{
	}

string xlfd::xlfd_string(void) const
	{
	strstream fontname;
	fontname << "-" << foundry << "-" << family;
	if(weight==medium) fontname << "-medium-";
	else               fontname << "-bold-";
	if(slant==oblique) fontname << "o-normal-*-";
	else               fontname << "r-normal-*-";
	fontname << (int)(size.pixels(95.0)) << "-";
	fontname << "*-*-*-*-*-" << encoding << "-*" << ends;
	string res=fontname.str();
	delete [] fontname.str();
	return res;
	}

font_manager::~font_manager()
	{
	map<string, font *>::iterator it=fonts_.begin();
	while(it!=fonts_.end()) {
		delete (*it).second;
		++it;
		}
	}

font *font_manager::get_font(const xlfd& xl)
	{
	map<string, font *>::iterator it=fonts_.find(xl.xlfd_string());
	if(it!=fonts_.end()) {
		debug("font " << xl.xlfd_string() << " reused");
		return (*it).second;
		}
	else {
		font *tmp=create_font_impl(xl);
		if(tmp==0) {
			debug("font " << xl.xlfd_string() << " does not exist");
			throw logic_error(string("font ")+xl.xlfd_string()+string(" does not exist"));
			}
		debug("font " << xl.xlfd_string() << " newly created");
		fonts_[xl.xlfd_string()]=tmp;
		debug("end");
		return tmp;
		}
	}

font *font_manager::create_font_impl(const xlfd&)
	{
	return 0;
	}

bool font::fill_tfm_info_(void)
	{
	const char *tfmpath=getenv("MN_TFM_PATH");
	if(tfmpath) {
		string tfmname=string(tfmpath)+"/"+xl_.family+".tfm";
		debug("attempting to open |" << tfmname << "|");
		ifstream tf(tfmname.c_str(), ios::binary | ios::in);
		if(tf.is_open()) {
			debug("succeeded!");
			tfm_=new tfm();
			tfm_->read(tf);
			debug("tfm file read");
			debug("this font has " << tfm_->param().size() << " parameters");
			for(unsigned int i=0; i<tfm_->charword().size(); ++i) {
				debug("character " << i << " tag " << (int)(tfm_->charword()[i].tag()));
				if(tfm_->charword()[i].tag()==tfm::char_word_info::list_tag) {
					debug("next in list is " << (int)tfm_->charword()[i].remainder());
					}
				}
			}
		else {
			debug("failed");
			}
		}
	else {
		debug("MN_TFM_PATH not set, so no tfm info available");
		}
	return true;
	}

bool font::fill_enc_info_(void)
	{
	const char *encpath=getenv("MN_ENC_PATH");
	if(encpath) {
		string encname=string(encpath)+"/"+xl_.family+".enc";
		debug("attempting to open |" << encname << "|");
		ifstream ef(encname.c_str());
		if(ef.is_open()) {
			debug("succeeded!");
			enc_=new enc();
			enc_->read(ef);
			}
		else {
			debug("failed");
			}
		}
	else {
		debug("MN_ENC_PATH not set, so no tfm info available");
		}
	return true;
	}

font::font(const xlfd& xl)
	: xl_(xl), tfm_(0), enc_(0), standard_ascent_(-1)
	{
	fill_tfm_info_();
	fill_enc_info_();
	}

font::~font()
	{
	if(tfm_) delete tfm_;
	if(enc_) delete enc_;
	}

string font::name(void) const
	{
	return xl_.xlfd_string();
	}

xlfd font::xl(void) const
	{
	return xl_;
	}

int32_t font::string_width(const string&) const
	{
	return 0;
	}

int32_t font::space_width(void) const
	{
	// FIXME
	return 0;
	}

int32_t font::standard_ascent(void) const
	{
	return 0;
	}

int32_t font::standard_descent(void) const
	{
	return 0;
	}

enc::enc(void)
	{
	}

void enc::read(istream& str)
	{
	wchar_t       u;
	unsigned long ul;
	unsigned int  i;
	char buffer[100];
	while(str.getline(buffer,99)) {
		if(buffer[0]=='#') continue;
		sscanf(buffer,"%i %li", &i, &ul);
		u=ul;
		debug("unicode " << u << " = font " << i);
		u2f_[u]=i;
		}
	}

void tfm::read(istream& str)
   {
   r16(str,&filelength_);
   header_=vector<fixword>(r16(str));
   r16(str,&min_char_);
   r16(str,&max_char_);
	assert(min_char_>=0);
	debug(min_char_ << ", " << max_char_);
	assert(min_char_<max_char_);
   charword_=vector<char_word_info>(max_char_+1);
   widths_=vector<fixword>(r16(str));
   heights_=vector<fixword>(r16(str));
   depths_=vector<fixword>(r16(str));
   itcorr_=vector<fixword>(r16(str));
   ligkern_=vector<ligkern_command>(r16(str));
   kern_=vector<fixword>(r16(str));
   exten_=vector<extensible_recipe>(r16(str));
   param_=vector<fixword>(r16(str)); 

   for(unsigned int i=0; i<header_.size(); ++i)
		header_[i]=r32(str);

   for(int i=min_char_; i<=max_char_; ++i) charword_[i]=r32(str);
   for(unsigned int i=0; i<widths_.size(); ++i) widths_[i]=r32(str);
   for(unsigned int i=0; i<heights_.size(); ++i) heights_[i]=r32(str);
   for(unsigned int i=0; i<depths_.size(); ++i) depths_[i]=r32(str);
   for(unsigned int i=0; i<itcorr_.size(); ++i) itcorr_[i]=r32(str);
   for(unsigned int i=0; i<ligkern_.size(); ++i) r32(str);
   for(unsigned int i=0; i<kern_.size(); ++i) r32(str);
   for(unsigned int i=0; i<exten_.size(); ++i) r32(str);
   for(unsigned int i=0; i<param_.size(); ++i) param_[i]=r32(str);
   }

tfm::tfm(void)
   {
   }

int16_t tfm::tfmlength(void) const
   {
   return filelength_;
   }

int16_t tfm::min_char(void) const
   {
   return min_char_;
   }

int16_t tfm::max_char(void) const
   {
   return max_char_;
   }

fixword tfm::checksum(void) const
   {
   return header_[0];
   }

fixword tfm::designsize(void) const
   {
   return header_[1];
   }

fixword tfm::slant(void) const
   {
   return param_[1];
   }

fixword tfm::space(void) const
   {
   return param_[2];
   }

fixword tfm::space_stretch(void) const
   {
   return param_[3];
   }

fixword tfm::space_shrink(void) const
   {
   return param_[4];
   }

fixword tfm::x_height(void) const
   {
   return param_[5];
   }

fixword tfm::quad(void) const
   {
   return param_[6];
   }

fixword tfm::extra_space(void) const
   {
   return param_[7];
   }

const vector<fixword>& tfm::header(void) const
   {
   return header_;
   }

const vector<tfm::char_word_info>& tfm::charword(void) const
   {
   return charword_;
   }

const vector<fixword>& tfm::widths(void) const
   {
   return widths_;
   }

const vector<fixword>& tfm::heights(void) const
   {
   return heights_;
   }

const vector<fixword>& tfm::depths(void) const
   {
   return depths_;
   }

const vector<fixword>& tfm::itcorr(void) const
   {
   return itcorr_;
   }

const vector<tfm::ligkern_command>& tfm::ligkern(void) const
   {
   return ligkern_;
   }

const vector<fixword>& tfm::kern(void) const
   {
   return kern_;
   }

const vector<tfm::extensible_recipe>& tfm::exten(void) const
   {
   return exten_;
   }

const vector<fixword>& tfm::param(void) const
   {
   return param_;
   }

int16_t tfm::r16(istream& str,int16_t *target) const
   {
   unsigned char tmp[2];
   str.read(tmp,2);
   *target=tmp[0]*256L+tmp[1];
   return *target;
   }

int16_t tfm::r16(istream& str) const
   {
   unsigned char tmp[2];
   str.read(tmp,2);
   return tmp[0]*256L+tmp[1];
   }

int32_t tfm::r32(istream& str) const
   {
   unsigned char tmp[4];
   if(!str) throw 1;
   str.read(tmp,4);
   return (1L<<24)*tmp[0]+(1L<<16)*tmp[1]+(1L<<8)*tmp[2]+tmp[3];
   }

tfm::char_word_info::char_word_info(void)
   {
   raw_=0;
   }

tfm::char_word_info::char_word_info(int32_t val)
   {
   raw_=val;
   }

int tfm::char_word_info::width_index(void) const
   {
   return ((raw_ & (255L<<24)) >>24);
   }

int tfm::char_word_info::height_index(void) const
   {
   return ((raw_ & (15L<<20)) >>20);
   }

int tfm::char_word_info::depth_index(void) const
   {
   return ((raw_ & (15L<<16)) >>16);
   }

int tfm::char_word_info::italic_index(void) const
   {
   return ((raw_ & (63L<<10)) >>10);
   }

tfm::char_word_info::tagtype tfm::char_word_info::tag(void) const
   {
   return static_cast<tagtype>( ((raw_ & (3L<<8)) >>8) );
   }

uint8_t tfm::char_word_info::remainder(void) const
   {
   return static_cast<uint8_t>( ((raw_ & 255)));
   }

