// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Various number classes.

  Copyright 1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <typeinfo>
#include <strstream.h>
#include <iostream.h>
#include <layout/numbers.hh>

#include <stdexcept>

using namespace layout;

ostream& operator<<(ostream& str, const fixword& fw)
   {
   //   str << fw.approx();
   str << fw.raw << "=";
   str.setf(ios::hex, ios::basefield);
   str << "(" << ((fw.raw&(255L<<24))>>24) << ",";
   str << ((fw.raw&(255L<<16))>>16) << ",";
   str << ((fw.raw&(255L<<8))>>8) << ",";
   str << (fw.raw&(255L)) << ")=";
   str.setf(ios::dec, ios::basefield);
   str << fw.approx();
   return str;
   }

fixword::fixword(void)
   {
   raw=0;
   }

fixword::fixword(const int32_t& val)
   {
   raw=val;
   }

double fixword::approx(void) const
   {
   double tmp;
   if( raw&(255L<<24) ) { 
		tmp= -16 + ((raw&(255L<<16))>>16) / (double)(1L<<4)
			+((raw&(255L<<8))>>8)   / (double)(1L<<12)
			+(raw&255L)            / (double)(1L<<20);
		}
   else { 
		tmp= ((raw&(255L<<16))>>16) / (double)(1L<<4)
			+((raw&(255L<<8))>>8)   / (double)(1L<<12)
			+(raw&255L)            / (double)(1L<<20);
		}
   return tmp;
   }

fixword& fixword::operator=(const int32_t& val)
   {
   raw=val;
   return *this;
   }


point::point(void)
	: x(0), y(0)
   {
   }

point::point(int32_t nx, int32_t ny)
	: x(nx), y(ny)
   {
   }

point& point::operator+=(const point& other)
	{
	x+=other.x;
	y+=other.y;
	return *this;
	}

ostream& operator<<(ostream& str, const layout::point& pt)
	{
	if(pt.x>=INFINITY) str << "(infinity,";
	else               str << "(" << pt.x << ",";
	if(pt.y>=INFINITY) str << "infinity)";
	else               str << pt.y << ")";

	return str;
	}

ostream& operator<<(ostream& str, const layout::length& l)
	{
	str << l.number();
	switch(l.unit()) {
		case length::px: str << "px"; break;
		case length::pt: str << "pt"; break;
		case length::pc: str << "pc"; break;
		case length::in: str << "in"; break;
		case length::bp: str << "bp"; break;
		case length::cm: str << "cm"; break;
		case length::mm: str << "mm"; break;
		case length::dd: str << "dd"; break;
		case length::cc: str << "cc"; break;
		case length::sp: str << "sp"; break;
		}
	return str;
	}

rectangle::rectangle(void)
	{
	}

rectangle::rectangle(point tl, point sz)
	: top_left(tl), size(sz)
	{
	}

point layout::operator+(const layout::point& p1, const layout::point& p2)
	{
	return point(p1.x+p2.x, p1.y+p2.y);
	}

length::length(void)
	: num_(0), unit_(sp)
	{
	}

length::length(const string& str)
	{
	num_=atof(str.c_str());
	string::size_type pos=str.find_first_not_of("-+0123456789. ");
	if(pos==str.npos) unit_=px;
	else {
		string un=str.substr(pos);
		if(un=="pt")      unit_=pt;
		else if(un=="px") unit_=px;
		else if(un=="pc") unit_=pc;
		else if(un=="in") unit_=in;
		else if(un=="bp") unit_=bp;
		else if(un=="cm") unit_=cm;
		else if(un=="mm") unit_=mm;
		else if(un=="dd") unit_=dd;
		else if(un=="cc") unit_=cc;
		else if(un=="sp") unit_=sp;
		else throw logic_error(string("unit ")+un+string(" not known"));
		}
	}

length::length(float fl, length_unit un)
	: num_(fl), unit_(un)
	{
	}

float length::number(void) const
	{
	return num_;
	}

length::length_unit length::unit(void) const
	{
	return unit_;
	}

int32_t length::pixels(float dpi) const
	{
	if(unit_==px) return (int32_t)(num_+0.5);
	float scp=scaled_points_();
	dpi/=65536*72.27;
	return (int32_t)(scp*dpi+0.5);
	}

float length::convert(length_unit un) const
	{
	if(unit_==px) {
		if(un==px) return num_;
		else throw logic_error("cannot convert from px to real units without dpi");
		}
	float scp=scaled_points_();
	switch(un) {
		case px:
			assert(1==0); // cannot happen
			break;
		case pt:
			scp/=65536;
			break;
		case pc:
			scp/=65536*12;
			break;
		case in:
			scp/=65536*72.27;
			break;
		case bp:
			scp/=65536*72.27/72.0;
			break;
		case cm:
			scp/=65536*72.27/2.54;
			break;
		case mm:
			scp/=65536*72.27/25.4;
			break;
		case dd:
			scp/=65536*1238.0/1157.0;
			break;
		case cc:
			scp/=65536*12*1238.0/1157.0;
			break;
		case sp:
			break;
		}
	return scp;
	}

float length::scaled_points_(void) const
	{
	float scp=num_;
	switch(unit_) {
		case px:
			assert(1==0); // should not have been called, internal inconsistency
			break;
		case pt:
			scp*=65536;
			break;
		case pc:
			scp*=65536*12;
			break;
		case in:
			scp*=65536*72.27;
			break;
		case bp:
			scp*=65536*72.27/72.0;
			break;
		case cm:
			scp*=65536*72.27/2.54;
			break;
		case mm:
			scp*=65536*72.27/25.4;
			break;
		case dd:
			scp*=65536*1238.0/1157.0;
			break;
		case cc:
			scp*=65536*12*1238.0/1157.0;
			break;
		case sp:
			break;
		}
	return scp;
	}
