// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  C++ library for streamable raster based images.

  Copyright 1998-1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef raster_hh__
#define raster_hh__

#include <string>
#include <vector.h>
#include <stdint.h>
#include <mnemonic/ptr.hh>

class decoder;
class raster_image;
class scaled_raster_image;
class raster_io;

class decoder {
	public:
		decoder(void);
		virtual ~decoder();

		virtual string            identifier(void) const=0;
		virtual ptr<scaled_raster_image> image_data(void) const;
		virtual bool              still_active(void) const;  // data not rejected yet

	   virtual void              write(const unsigned char *, long length)=0;

		virtual void              data_end(void)=0;
	protected:
		bool still_active_;
		bool finished_;
		ptr<scaled_raster_image> image_;
};

class raster_image {
public:
		raster_image(long,long);
		~raster_image();
		void addFrame(); /*creates a new frame from now on the image_data will return the new pointer*/
		uint8_t *image_data(void) const {return raster_[frames-1];}
		uint8_t *image_alpha(void) const {return alpha_[frames-1];} //returns 0 if no alpha map is available
		void setAlpha();
		void setDelay(int i){delay.push_back(i);}; //the delay stored with an image is the time you have to wait before
							    // displaying it in 100th of seconds


		long width()const {return width_;}
		long height()const {return height_;}

		void    updated(
		    long ax,
		    long ay,
		    long bx, 
		    long by);


		vector<int>       delay;
		vector<uint8_t *> raster_;			//rgbrgbrgb
		vector<uint8_t *> alpha_;			//255 opaque 0 transparent
		int	            frames;
		long              width_, height_;
		bool              hasAlpha_;
		bool	            updated_;

};

class scaled_raster_image {
	public:
		scaled_raster_image(long, long);
		~scaled_raster_image();

		void setAlpha();
		void addFrame(); /*creates a new frame from now on the image_data will return the new pointer*/
		// info
		long    width(void) const;
		long    height(void) const;
		long 	  scaled_width(void) const;
		long    scaled_height(void) const;
		
		void setDelay(int i){image_data_->setDelay(i);}
		// Data. 
		// (image_data())[row*width*3+col*3+0]=red
		//                                 +1]=green
		//                                 +2]=blue
		//
		// The memory pointed to by these pointers is _always_ owned by raster_image.
		uint8_t *image_data(void) const {return image_data_->image_data();}
		uint8_t *image_alpha(void) const {return image_data_->image_alpha();} //returns 0 if no alpha map is available
		uint8_t *scaled_image_data(void);
		uint8_t *scaled_image_alpha(void);
		// algorithms
		void    blend(const raster_image *);
		void    scale(float);
		void    scale(long, long);

		void    updated(
		    long ax,
		    long ay,
		    long bx, 
		    long by);

		//long last_finished_scanline;
		int  nextFrame(){
		    int retv=image_data_->delay[curframe]; 
		    curframe++;
		    if (curframe==image_data_->frames) curframe=0; 
		    return retv;
		} 
		int frames(){return image_data_->frames;}
		                  /*moves the pointer of the current frame to the next and returns the delay time */
                							    
	private:
		ptr<raster_image>   image_data_;
		long	            scaled_width_,scaled_height_;
		int               curframe;
		vector<uint8_t *> scaled_raster_; 
		vector<uint8_t *>	scaled_alpha_;
};

class raster_io { 
	public:
		raster_io(void);
		~raster_io();

		void use_decoder(const string& mimetype); // if you don't call this, all decoders will try.
	   void write(const unsigned char *, long length);
		long read(unsigned char *, long length);

		ptr<scaled_raster_image> image(void) const;
		string            decoder_type(void) const;
		void              await_completion(void) const;
	private:
		bool               initialised_;
		vector<decoder *>  decoders_;
};
		

#endif
