#include "config.h"
#ifdef HAVE_LIBUNGIF
// We have to include this first, BEFORE anyone gets to set the signal
// handler behaviour to follow sysv (leading to EINTR all over the place).
#include <signal.h>


#include <rasterimage/gifraster.hh>
#include <unistd.h>
#include <algo.h>
#include <stdexcept>
#include <errno.h>

#define DEBUG_CLASS "GIF_DECODER"
#include <mnemonic/debug.hh>

gif_decoder::gif_decoder(void)
	: decoder(), gif_thread_(0)
	{
	signal(SIGPIPE,SIG_IGN);
	}

gif_decoder::~gif_decoder()
	{
	debug("destroying gif_decoder");
	if(gif_thread_) {
		gif_thread_->stop();
		debug("thread stopped, deleting");
		delete gif_thread_;
		}
	}

gif_decoder::gif_thread::gif_thread(gif_decoder *owner)
	: error_encountered_(false), finished_decoding_(false), is_running_(false),
	  writepipe_open_(false), readpipe_open_(false), gif_ft_(0), dec_(owner), linebuffer_(0)
	{
	debug("initialising gif thread");
	debug("initialising gif thread done");
	}

gif_decoder::gif_thread::~gif_thread()
	{
	if(readpipe_open_) {
		debug("closing read  pipe " << datapipe_[0] << "," << datapipe_[1]);
		if(DGifCloseFile(gif_ft_)==GIF_ERROR) {
			debug("error closing gif file");
			}
		}
	if(writepipe_open_) {
		debug("closing write pipe " << datapipe_[0] << "," << datapipe_[1]);
		close(datapipe_[1]);
		}
	
	if(linebuffer_) delete [] linebuffer_;
	
	debug("thread destroyed");
	}

void gif_decoder::gif_thread::begin(void)
	{
	// Creating the pipe and _then_ starting the thread led to all sorts of funny
	// problems. So we create it upon thread startup now.
	if(pipe(datapipe_)!=0)
		throw logic_error("gif pipe init failed");
	writepipe_open_=true;
	readpipe_open_=true;
	int frame=0;
        int trans_color_=-1;
	debug("pipe: " << datapipe_[0] << "," << datapipe_[1]);
	
	// We have to do some error checking as libungif doesn't do it...
	// This doesn't seem to go wrong now anymore, with pipe creation moved to the
	// startup of the thread, but it doesn't hurt either.
	if(lseek(datapipe_[0],0,SEEK_CUR)==-1 && errno==EBADF) {
		debug("error on reading pipe");
		error_encountered_=true;
		start_cond.lock();
		is_running_=true;
		start_cond.broadcast();
		start_cond.unlock();
		}
	else {
		if(errno!=ESPIPE) {
			debug("unexpected error: " << string(strerror(errno)));
			}
		start_cond.lock();
		is_running_=true;
		start_cond.broadcast();
		start_cond.unlock();
		gif_ft_=DGifOpenFileHandle(datapipe_[0]);
		}
	if(gif_ft_) {
		debug("this is a GIF image");

//		BackGround = GifFile->SBackGroundColor;
		ColorMapObject *colormap_ = (gif_ft_->Image.ColorMap
											  ? gif_ft_->Image.ColorMap
											  : gif_ft_->SColorMap);

		dec_->image_=new scaled_raster_image(gif_ft_->SWidth, gif_ft_->SHeight);
		int SW=gif_ft_->SWidth;
		int SH=gif_ft_->SHeight;
		debug("reading the image data");
		GifRecordType rectype_;
		linebuffer_ = new unsigned char[gif_ft_->SWidth];
		for(int j=0; j<gif_ft_->SWidth; ++j) {
			linebuffer_[j] = gif_ft_->SBackGroundColor;
			}
		int InterlacedOffset[] = { 0, 4, 2, 1 };
		int InterlacedJumps[] = { 8, 8, 4, 2 };

		do {
			if(DGifGetRecordType(gif_ft_, &rectype_)==GIF_ERROR) {
				debug("oh oh, illegal record");
				error_encountered_=true;
				readpipe_open_=false;
				break;
				}
			switch(rectype_) {
				case IMAGE_DESC_RECORD_TYPE: {
					if(DGifGetImageDesc(gif_ft_)==GIF_ERROR) {
						debug("error reading image description record");
						error_encountered_=true;
						readpipe_open_=false;
						break;
						}
					int Row=gif_ft_->Image.Top;
					int Col=gif_ft_->Image.Left;
					int Width=gif_ft_->Image.Width;
					int Height=gif_ft_->Image.Height;
					GifColorType *cmap_entry_;
					if (frame!=0) dec_->image_->addFrame();
					if (gif_ft_->Image.Interlace) {
						debug("reading interlaced image");
						for (int i=0; i < 4; i++)
							for (int j=Row + InterlacedOffset[i]; j < Row + Height; j += InterlacedJumps[i]) {
								if (DGifGetLine(gif_ft_, &linebuffer_[Col], Width) == GIF_ERROR) {
									debug("error reading line");
									error_encountered_=true;
									readpipe_open_=false;
									break;
									}
								for(int k=Col; k<Width+Col; ++k) {
									cmap_entry_ = &colormap_->Colors[linebuffer_[k]];
									dec_->image_->image_data()[j*SW*3+k*3+0]=cmap_entry_->Red;
									dec_->image_->image_data()[j*SW*3+k*3+1]=cmap_entry_->Green;
									dec_->image_->image_data()[j*SW*3+k*3+2]=cmap_entry_->Blue;
									if (trans_color_!=-1)
									dec_->image_->image_alpha()[j*SW+k]= ((linebuffer_[k]==trans_color_) ? 0 : 255 ); 
									}
//								dec_->image_->last_finished_scanline=Height;
								dec_->image_->updated(0,Height,Width-1,Height);
								}
						debug("reading done");
						}
					else {
						debug("reading normal image " << Row << "," << Col << "-" << Width << "," << Height);
						for (int i = 0; i < Height; i++) {
							debug("line " << i);
							if (DGifGetLine(gif_ft_, &linebuffer_[Col], Width) == GIF_ERROR) {
								debug("error reading line");
								error_encountered_=true;
								readpipe_open_=false;
								break;
								}
							for(int j=Col; j<Width+Col; ++j) {
								cmap_entry_ = &colormap_->Colors[linebuffer_[j]];
								dec_->image_->image_data()[Row*SW*3+j*3+0]=cmap_entry_->Red;
								dec_->image_->image_data()[Row*SW*3+j*3+1]=cmap_entry_->Green;
								dec_->image_->image_data()[Row*SW*3+j*3+2]=cmap_entry_->Blue;
								if (trans_color_!=-1)
							        dec_->image_->image_alpha()[Row*SW+j]= ((linebuffer_[j]==trans_color_) ? 0 : 255 ); 
								}
//							dec_->image_->last_finished_scanline=i+1;
							dec_->image_->updated(0,i+1,Width-1,i+1);
							++Row;
							}
						}
					debug("image data read");
					frame++; 
    					break;
					}
				case EXTENSION_RECORD_TYPE: {
					GifByteType *Extension;
					int ExtCode;
					if (DGifGetExtension(gif_ft_, &ExtCode, &Extension) == GIF_ERROR) {
						debug("error reading extension record");
						error_encountered_=true;
						readpipe_open_=false;
						break;
						}
					while (Extension != NULL) {
						if (ExtCode== GRAPHICS_EXT_FUNC_CODE){
						    if (Extension[1] & 1) {
						        dec_->image_->setAlpha();
							trans_color_=Extension[4];
						    } else {
						        trans_color_=-1;
						    }
						    //Extension[1]>>2&7 disposal not supported
						    dec_->image_->setDelay(Extension[3]*256+Extension[2]); 
						}
						if (DGifGetExtensionNext(gif_ft_, &Extension) == GIF_ERROR) {
							debug("error reading next extension record");
							error_encountered_=true;
							readpipe_open_=false;
							break;
							}
						}
					break;
					}
				case TERMINATE_RECORD_TYPE: 
					debug("terminate record read");
				default:
					break;
				};
			debug("read another record");
			} while(rectype_ != TERMINATE_RECORD_TYPE && error_encountered_==false);
		}
	else {
		error_encountered_=true;
		readpipe_open_=false;
		debug("data is not GIF");
		}
	finished_cond.lock();
	finished_decoding_=true;
	finished_cond.broadcast();
	finished_cond.unlock();
	}

bool gif_decoder::gif_thread::is_running(void) const
	{
	return is_running_;
	}

bool gif_decoder::gif_thread::error_encountered(void) const
	{
	return error_encountered_;
	}

bool gif_decoder::gif_thread::finished_decoding(void) const
	{
	return finished_decoding_;
	}

int gif_decoder::gif_thread::writepipe(void) const
	{
	return datapipe_[1];
	}

string gif_decoder::identifier(void) const
	{
	return "gif";
	}

bool gif_decoder::still_active(void) const
	{
	debug("gif decoder still active?");
	if(still_active_) { 
		debug("decoder is!");
		if(!gif_thread_) { 
			debug("though thread not running");
			return true;
			}
		if(gif_thread_->error_encountered()) return false;
		else                                 return true;
		}
	else return false;
	}

void gif_decoder::write(const unsigned char *dat, long siz)
	{
	debug("gif decoder write");
	if(!gif_thread_) {
		debug("creating gif thread");
		gif_thread_=new gif_thread(this);
		try {
			gif_thread_->start();
			}
		catch(exception& ex) {
			debug("exception: " << ex.what());
			throw;
			}
		gif_thread_->start_cond.lock();
		while(!(gif_thread_->is_running()))
			gif_thread_->start_cond.wait();
		gif_thread_->start_cond.unlock();
		debug("our pipe: " << gif_thread_->writepipe());
		}
	debug("feeding " << siz << " bytes to libgif");
	long tot=::write(gif_thread_->writepipe(),dat,siz);
	debug("written " << tot << " bytes");
	if(tot==-1) {
		still_active_=false;
		debug("cannot feed to " << gif_thread_->writepipe() << ": " << string(strerror(errno)));
		}
	debug("feeding done");
	}

void gif_decoder::data_end(void)
	{
	if(gif_thread_) {
		if(gif_thread_->error_encountered())
			still_active_=false;
		gif_thread_->finished_cond.lock();
		while(!(gif_thread_->finished_decoding()))
			gif_thread_->finished_cond.wait();
		gif_thread_->finished_cond.unlock();
		gif_thread_->stop();
		delete gif_thread_;
		gif_thread_=0;
		}
	}
#endif