#include "config.h"
#ifdef HAVE_LIBJPEG

#include <rasterimage/jpgraster.hh>
#include <unistd.h>
#include <algo.h>
#include <stdexcept>

#define DEBUG_CLASS "JPG_DECODER"
#include <mnemonic/debug.hh>

#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <setjmp.h>

jpg_decoder::jpg_decoder(void)
	: decoder(), error_encountered_(false), buffer_(0), status_(read_header)
	{
	debug("initialising jpg io");

	jpeg_create_decompress(&cinfo_);

 	cinfo_.err=jpeg_std_error(&(jerr_.pub));
	jerr_.pub.error_exit=jpeg_error_exit_;
//	jerr_.pub.output_message=::my_output_message_;

	jerr_.dec=this;
	jsource_.dec=this;
	jsource_.ignore_writes=0;
	jsource_.keep=0;

	cinfo_.src=&jsource_.pub;
	cinfo_.src->init_source = init_source_;
	cinfo_.src->fill_input_buffer = fill_input_buffer_;
	cinfo_.src->skip_input_data = skip_input_data_;
	cinfo_.src->resync_to_restart = jpeg_resync_to_restart; /* use default method */
	cinfo_.src->term_source = term_source_;
	cinfo_.src->bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	cinfo_.src->next_input_byte = NULL; /* until buffer loaded */
	debug("jpg io initialisation done");
	}

jpg_decoder::~jpg_decoder()
	{
	debug("~jpg_decoder " << this);
// DO NOT DELETE buffer_. It has been created by libjpeg and will be cleaned
// up automatically.
//	if(buffer_) delete buffer_;
	debug("~jpg_decoder done");
	}

extern "C" {

	void jpeg_error_exit_(j_common_ptr cinfo)
		{
		struct jpg_error_obj *myerr= (struct jpg_error_obj *) cinfo->err;
		debug("jpg_error_exit_ " << myerr->dec);
		longjmp(myerr->jmpb, 1);
		}
	void my_output_messsage_(j_common_ptr cinfo)
		{
		// FIXME: we probably want to keep these messages somewhere for readout
		}
	void init_source_(j_decompress_ptr cinfo)
	   {
		debug("init_source_");
//		struct jpg_source_obj *mysrc = (struct jpg_source_obj *)cinfo->src;
		}
	boolean fill_input_buffer_(j_decompress_ptr cinfo)
	   {
		debug("fill_input_buffer_");
		struct jpg_source_obj *mysrc = (struct jpg_source_obj *)cinfo->src;
		memmove(mysrc->buffer,mysrc->pub.next_input_byte,mysrc->pub.bytes_in_buffer);
		mysrc->keep=mysrc->pub.bytes_in_buffer;
		return false;
		}
   void skip_input_data_(j_decompress_ptr cinfo, long num_bytes)
	   {
		struct jpg_source_obj *mysrc = (struct jpg_source_obj *)cinfo->src;
		debug("skip_input_data_ " << num_bytes << " while buffer is " << mysrc->pub.bytes_in_buffer);

		if(num_bytes<=0) return;
		if(num_bytes>(long)(mysrc->pub.bytes_in_buffer)) {
			debug("request more data");
			mysrc->ignore_writes=num_bytes-mysrc->pub.bytes_in_buffer;
			mysrc->pub.bytes_in_buffer=0;
			}
		else {
			debug("buffer big enough");
			mysrc->pub.next_input_byte+=(size_t) num_bytes;
			mysrc->pub.bytes_in_buffer-=(size_t) num_bytes;
			}
		}
	void term_source_(j_decompress_ptr cinfo)
	   {
		debug("term_source_");
		}
};

void jpg_decoder::write(const unsigned char *dat, long siz)
	{
	if(!still_active_ || error_encountered_) return;
	if(setjmp(jerr_.jmpb)!=0) {
		debug("returned from longjmp");
		error_encountered_=true;
		return;
		}

	debug("feeding " << siz << " bytes to libjpg");

	if(siz>=jsource_.ignore_writes) {
		dat+=jsource_.ignore_writes;
		siz-=jsource_.ignore_writes;
		jsource_.ignore_writes=0;
		}
	else {
		jsource_.ignore_writes-=siz;
		return;
		}
	const unsigned char *start=dat;
	long                 todo=siz;
	while(todo>0) {
		debug("new bytes " << todo);
		debug("have " << jsource_.keep << " left from last time");
		debug("doing " << min(32768L,todo+jsource_.keep) << " now");
		memcpy(&(jsource_.buffer[jsource_.keep]),start,min(32768L,todo+jsource_.keep));
		jsource_.pub.next_input_byte = jsource_.buffer;
		jsource_.pub.bytes_in_buffer = min(32768L,todo+jsource_.keep);
		todo-=min(32768L,todo+jsource_.keep);
		start+=min(32768L,todo+jsource_.keep);
		jsource_.keep=0;
		
		switch(status_) {
			case read_header:
				if(jpeg_read_header(&cinfo_,TRUE)==JPEG_SUSPENDED || error_encountered_)
					break;
				cinfo_.out_color_space = JCS_RGB;
				status_=start_decompress;
			case start_decompress:
				if(jpeg_start_decompress(&cinfo_)==FALSE || error_encountered_)
					break;
				if(image_==0) {
					image_=new scaled_raster_image(cinfo_.image_width, cinfo_.image_height);
					debug("this is a JPG image " << cinfo_.image_width << "," << cinfo_.image_height);
					}
				status_=read_scanlines;
			case read_scanlines:
				row_stride_ = cinfo_.output_width * cinfo_.output_components;
				if(buffer_==0)
					buffer_ = (*cinfo_.mem->alloc_sarray)((j_common_ptr) &cinfo_, JPOOL_IMAGE, row_stride_, 1);
				while (cinfo_.output_scanline < cinfo_.output_height) {
					debug("reading scanline " << cinfo_.output_scanline);
					if(jpeg_read_scanlines(&cinfo_,buffer_,1)==0 || error_encountered_)
						break;
					for(long i=0; i<row_stride_; ++i) {
						image_->image_data()[(cinfo_.output_scanline-1)*row_stride_+i]=buffer_[0][i];
						}
//					image_->last_finished_scanline=cinfo_.output_scanline;
					image_->updated(0,cinfo_.output_scanline,cinfo_.image_width-1,cinfo_.output_scanline);
					}
				if(cinfo_.output_scanline < cinfo_.output_height) {
					debug(jsource_.keep);
					break;
					}
				status_=finish_decompress;
			case finish_decompress:
				jpeg_finish_decompress(&cinfo_);
				if(error_encountered_) break;
				jpeg_destroy_decompress(&cinfo_);
				status_=all_done;
			case all_done:
				break;
			}
		debug("leaving write with status " << (int)status_);
		}
	}

void jpg_decoder::data_end(void)
	{
	debug("flushing jpg input data");
	}

string jpg_decoder::identifier(void) const
	{
	return "jpg";
	}

#endif
