// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Threaded C++ wrappers for low-level tcp/ip operations, implementation.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string>
#include <strstream.h>
#include <tcpcon/tcp_connection.hh>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <stdexcept>
#include "config.h"

#define DEBUG_CLASS "TCP"
#include <mnemonic/debug.hh>

int         tcp_connection::total_connections_=0;
thread_cond tcp_connection::total_connections_cond_;

tcp_connection::tcp_connection(const string &h, int p) 
	: host_(h), port_(p), proxy_host_(""), proxy_port_(-1)
	{
	max_transfer_ = 0;
	data_transferred_ = 0;
	fd_ = -1;
	status_ = not_connected;
	}

tcp_connection::tcp_connection(const string &h, int p, const string &prox, int prox_p) 
	: host_(h), port_(p), proxy_host_(prox), proxy_port_(prox_p)
	{
	max_transfer_ = 0;
	data_transferred_ = 0;
	fd_ = -1;
	status_ = not_connected;
	}

tcp_connection::~tcp_connection() 
	{
	if(fd_) close_(fd_);
	status_ = closed;
	}

int tcp_connection::lookup(void) 
	{
	string lu_host;
	
	if(proxy_port_==-1) { lu_host=host_; connect_port_=port_; }
	else                { lu_host=proxy_host_; connect_port_=proxy_port_; }

	debug("looking up host... "<< lu_host<<endl);

	int count = 0;
	int the_errno;
	struct hostent *hentp=0;
	struct hostent hent;
	char buffer[8192];

	debug("request for DNS lookup " << lu_host);
	mutex_.lock(); // FIXME: I have no idea why you cannot do this in multiple threads
	// at the same time, but things just start to hang. 
	debug("making system call");
	
#ifdef HAVE_GETHOSTBYNAME_R_6_ARG
	while(gethostbyname_r(lu_host.c_str(), &hent, buffer,sizeof buffer,
								 &hentp, &the_errno) < 0) {
#endif
#ifdef HAVE_GETHOSTBYNAME_R_5_ARG
	while((hentp = gethostbyname_r(lu_host.c_str(), &hent, buffer, 
											sizeof buffer, &the_errno)) == 0) {
#endif
#ifdef HAVE_GETHOSTBYNAME
	while((hentp = gethostbyname(lu_host.c_str()))==0) {
		the_errno = h_errno;
#endif
		switch (the_errno) {
			case HOST_NOT_FOUND:
				debug("host not found");
				mutex_.unlock();
				return -1;
			case NO_ADDRESS:
				debug("no address");
				mutex_.unlock();
				return -1;
			case TRY_AGAIN:
				debug("try again");
				if(++count<5) {
					sleep(1);
					debug("going to try again!");
					continue;
					}
				else {
					mutex_.unlock();
					status_=host_not_found;
					return -1;
					}
			case ERANGE:
				debug("buffer not big enough");
			case NO_RECOVERY:
			default:
				debug("or another error");
				mutex_.unlock();
				return -1;
			}
		}
	mutex_.unlock();
	memcpy(&(connect_ipnum_), hentp->h_addr, sizeof connect_ipnum_);
	debug("dns lookup successful for " << lu_host);

	return 0;
	}

int tcp_connection::connect(void) 
	{
   // FIXME: the maximum number of concurrent connections is fixed here
	// at 5. We probably want to make that configurable.

	total_connections_cond_.lock();
	while(total_connections_ > 4) {
		debug("total connections is " << total_connections_ << ", waiting...");
		total_connections_cond_.wait();
		}
	++total_connections_;
	total_connections_cond_.unlock();

	struct linger ling;
	struct sockaddr_in  name;

	fd_ = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	int size = sizeof(linger);
	ling.l_onoff = 0;
	setsockopt(fd_,SOL_SOCKET,SO_LINGER,(void *)&ling,size);
	
	// open the actual connection
	name.sin_family = AF_INET;
	name.sin_port = htons(connect_port_);
	name.sin_addr.s_addr = connect_ipnum_;
	debug("connecting: "<<connect_ipnum_<<" "<<connect_port_);	

	status_ = ok;
	int retval=0;

	if( ::connect(fd_, (struct sockaddr *)&name, sizeof(name)) ) {
		switch(errno) {
			case ECONNREFUSED:
				status_=connection_refused;
				debug("connection refused");
				retval=-1;
				break;
			case ETIMEDOUT:
				status_=timed_out;
				debug("connection timed out");
				retval=-1;
				break;
			case ENETUNREACH:
				status_=network_unreachable;
				debug("network unreachable");
				retval=-1;
				break;
			case EHOSTUNREACH:
				status_=no_route;
				debug("no route to host");
				retval=-1;
				break;
			default:
				error_ = "unhandled error on connect: "+string(strerror(errno));
				status_ = other_error;
				total_connections_cond_.lock();
				--total_connections_;
				total_connections_cond_.broadcast();
				total_connections_cond_.unlock();
				throw logic_error(error_);
			}
		}
	if(retval!=0) {
		total_connections_cond_.lock();
		--total_connections_;
		total_connections_cond_.broadcast();
		total_connections_cond_.unlock();
		}
	return retval;
	}

void tcp_connection::set_max_transfer(size_t mt) 
	{
	data_transferred_ = 0;
	max_transfer_ = mt;
	if(status_ == max_transferred) 
		status_ = ok;
	}

ssize_t tcp_connection::read(char *buffer, size_t buf_size) 
	{
	if(status_ != ok)
		return -1;
	if(max_transfer_!=0 && data_transferred_ >= max_transfer_) {
		status_ = max_transferred;
		return -1;
		}
	size_t amt_to_read;
	if(max_transfer_) {
		size_t amt_left_on_conn = max_transfer_ - data_transferred_;
		amt_to_read = (amt_left_on_conn < buf_size) ? (amt_left_on_conn) : buf_size;
		} 
	else {
		amt_to_read = buf_size;
		}
	debug("trying to read "<<amt_to_read);
	int amt_read=0;
	
	while((amt_read = ::read(fd_, buffer, amt_to_read)) == -1 || amt_read == 0) {
		if(amt_read == 0) {
			debug("end of file in read()");
			status_ = closed;
			close_(fd_);
			return -1;
			}
		switch(errno) {
			case EINTR:  // ok, let's try again
				debug("received EINTR in read()");
				continue;
			case EAGAIN: // can't happen
				debug("received EAGAIN in read()");
				continue;
			case EIO:    // oof, some bad things happen
				status_ = closed;
				close_(fd_);
				return -1;
			case EBADF:
			case EINVAL:
			case EFAULT:
				error_ = "[tcp_connection]: invalid file descriptor for reading";
				status_ = other_error;
				close_(fd_);
				return -1;
				// FIXME: handle ECONNRESET and retry (this is what makes netscape fail
				// with long downloads). Also see the error handling of wget.
			default:     // should not happen
				throw logic_error("[tcp_connection]: unhandled error during read: "+string(strerror(errno)));
			}
		}
	data_transferred_+=amt_read;
	return amt_read;
	}

int tcp_connection::status() const 
	{
	return status_;
	}

const char *tcp_connection::error_text() const 
	{
	if(status_ != other_error) return "";
	else return error_.c_str();
	}

ssize_t tcp_connection::write(const char *buffer, size_t size) 
	{
	if(status_!=ok) return -1;
	size_t amt_written = 0;
	int res;
	while(amt_written != size) {
		if((res=::write(fd_,buffer+amt_written,size-amt_written)) == -1) {
			switch(errno) {
				case EBADF:
				case EINVAL:
					error_ = "[tcp_connection]: bad file descriptor for writing";
					status_ = other_error;
					close_(fd_);
					return -1;
				case EIO:
					debug("Got EIO");
					status_ = closed;
					close_(fd_);
					return -1;
				case EPIPE:
					debug("Got EPIPE");
					status_ = closed;
					close_(fd_);
					return -1;
				case EINTR: 
					debug("received EINTR in write()");
					continue;
				default:
					status_ = other_error;
					error_ = "[tcp_connection]: unhandled error during write"; 
					throw logic_error(error_.c_str());
				}
			} 
		else amt_written+=res;
		debug("Trying to write data (so far "<<amt_written<<" of"<<size<<")");
		}
	debug("Done writing block");
	return 0;
	}

void tcp_connection::close_(int f)
	{
	close(f);
	total_connections_cond_.lock();
	--total_connections_;
	total_connections_cond_.broadcast();
	total_connections_cond_.unlock();
	}

int tcp_connection::get_port() const
	{
	return port_;
	}

string tcp_connection::get_hostname() const
	{
	return host_;
	}

bool tcp_connection::uses_proxy() const
	{
	if(proxy_port_!=-1) return true;
	else                return false;
	}
