// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  URL string implementation.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string>
#include <strstream.h>
#include "url/url_string.hh"
#define DEBUG_CLASS "URL"
#include <algorithm>
#include "mnemonic/debug.hh"

/* Implementation details:

   We parse the scheme part of the URL at creation time.
   When we are asked about other parts, we parse the rest of the URL.

   We assume the string classes are smart enough to do substrings nicely,
   so we don't eat much space.

   Also, we assuming that + and - and others work nicely on
   string::const_iterators, which is true as they are RandomInterators.

	*/

url_string::url_string(const url_string &u) 
	{
	referrer = u.referrer ? new url_string(*(u.referrer)) : 0;
	original_url = u.original_url;
	decoding_finished_ = u.decoding_finished_;
	if(decoding_finished_) {
		scheme = u.scheme;
		url = u.url;
		net_path = u.net_path;
		abs_path = u.abs_path;
		rel_path = u.rel_path;
		params = u.params;
		port = u.port;
		if(port==0) port=default_port(); //is this correct?
		query = u.query;
		fragment = u.fragment;
		username = u.username;
		passwd = u.passwd;
		}
	}

url_string::url_string(const char *u) 
	: original_url(u), referrer(0), decoding_finished_(0) 
	{
	//port=default_port();
	}

url_string::url_string(const char *u, const url_string &ref) 
	: original_url(u), decoding_finished_(0) 
	{
	//port=default_port(); why set ports on not decoded urls? decoding should set those
	referrer = new url_string(ref);
	}

url_string::url_string() 
	: referrer(0)
	{
	port=default_port();
	decoding_finished_=1; //no use to try anything on nothing
	//is_relative ?
	}

url_string::~url_string() 
	{
	if(referrer) delete referrer;
	}

url_string & url_string::operator=(const url_string &u) 
	{
	//port=default_port();
	if (referrer) delete referrer;
	referrer = u.referrer ? new url_string(*(u.referrer)) : (url_string *)0;
	original_url = u.original_url;
	decoding_finished_ = u.decoding_finished_;
	if(decoding_finished_) {
		scheme = u.scheme;
		url = u.url;
		net_path = u.net_path;
		abs_path = u.abs_path;
		rel_path = u.rel_path;
		params = u.params;
		port = u.port;
		if(port==0) port=default_port();
		query = u.query;
		fragment = u.fragment;
		username = u.username;
		passwd = u.passwd;
		}
	return *this;
	}

void url_string::finish_decoding_(void) const
	{
	if(decoding_finished_) return;
	string::const_iterator loc;
	loc = DecodeScheme(original_url.begin());
	if(is_relative_) { // No scheme found in original_url.
		DoRelDecode(loc);
		} 
	else {// It may still be relative. 
		loc = HandleAbs(loc);
		if(loc != original_url.end()) { // They left us some work to do.
			DoRelDecode(loc);
			}
		}
	fixup_();
	decoding_finished_ = 1;
	}

int url_string::default_port(void) const
	{
	return 0;
	}

void url_string::DoRelDecode(string::const_iterator loc) const
	{
	string::const_iterator loc2;
	loc=DecodeNetPath(loc);
	if((loc2 = DecodeAbsPath(loc)) == loc) {
		loc2 = DecodeRelPath(loc);
		}
	loc = DecodeParams(loc2);
	loc = DecodeQuery(loc);
	DecodeFragment(loc);	

	}

string::const_iterator url_string::DecodeScheme(string::const_iterator loc) const
	{
	//starts at loc searches for a "scheme" and returns
	//the position for the next decoding
	string::const_iterator orig = loc, end=original_url.end();
	//skip spaces
	scheme="";
	while(loc < end) {
    	    if(isalnum(*loc)||*loc=='.') {
		loc++;
		} else break;
	    }
	//step through the letters which are supposed to be the scheme    
	while(loc < end) {
		if(isalpha(*loc))
			loc++;
		else
			break;
		}
	
	if(loc==end || *loc != ':') {
		is_relative_ = 1;
		if(referrer) scheme = referrer->get_scheme(); // better be, or we can't route
    		debug ("scheme: "<<scheme);
		return orig;
		}
	is_relative_ = 0;
	scheme = original_url.substr(orig-original_url.begin(),loc - orig);
	transform(scheme.begin(),scheme.end(),scheme.begin(),tolower);
	loc++; // eat ':'
        debug ("scheme: "<<scheme);
	debug ("got: "<<string(loc,end));
	return loc;
	}

string::const_iterator url_string::DecodeNetPath(string::const_iterator loc)  const
	{
   try {
	if (referrer) {
		//initilize those, if we bail out we use those of the referrer
		username=referrer->get_username();
		passwd=referrer->get_passwd();
		port=referrer->get_port();
		if(port==0) port=default_port();
		net_path=referrer->get_net_path();
		} 
	else
		{
		port=default_port();
		}
	string::const_iterator orig = loc;
	debug ("loc="<<string(loc,original_url.end()));
	if(loc == original_url.end()) return loc;
	/* Eat spaces */
	while(loc != original_url.end() && isspace(*loc)) {
	    loc++;
	    }
	if(loc == original_url.end()) return orig;
	
	if(*loc == '/') loc++; /* this could be more terse */
	else return orig;
	if(loc == original_url.end()) return orig;
	
	if(*loc == '/') loc++;
	else return orig;
	
	if(loc == original_url.end()) return orig;
	//ok there should be some host indicator
	//reset username etc
	username="";
	port=default_port();
	passwd="";
	
	orig = loc;
	/* The next line is not exactly correct -- but unless someone
	   wants to figure out international stuff,  this will work */
	while(loc != original_url.end() && *loc!='/' && *loc!='#') {
		loc++;
		}
	
	net_path = original_url.substr(orig-original_url.begin(),loc - orig);
	debug ("net_path="<<net_path);

		
	string::const_iterator i;
	string::reverse_iterator ri=find(net_path.rbegin(),net_path.rend(),'@');
	if(ri!=net_path.rend()) {
		//username=net_path.substr(0,i-net_path.begin());
		//debug("username found: "<<username);
		//i++;
		//net_path=net_path.substr(i-net_path.begin(),net_path.end()-i);
		//debug("netpath  found: "<<net_path);
		username=net_path.substr(0,&(*ri)-net_path.begin());
		net_path=net_path.substr(&(*ri)-net_path.begin()+1,net_path.size());
		//check username for passwd
		i=find(username.begin(),username.end(),':');
		if(i!=username.end()) {
			i++;
			passwd=username.substr(i-username.begin(),username.end()-i);
			username=username.substr(0,i-username.begin()-1);
			}
		}
	
	i=find(net_path.begin(),net_path.end(),':');
	if(i!=net_path.end()) {
		string::const_iterator nameend=i;
		i++;
		port=0;
		while(i!=net_path.end() && isdigit(*i)) {
			debug ("i="<<*i);
			port = port * 10;
			port += *i - '0';
			i++;
			}
		net_path=net_path.substr(0,nameend-net_path.begin());
		debug ("new net_path: "<<net_path);
		} 
		}
	catch(exception& ex) {
		cerr << "url_string exception: " << ex.what() << endl;
		throw;
		}
	return loc;
	}

string::const_iterator url_string::DecodeAbsPath(string::const_iterator loc)  const
	{
	string::const_iterator orig=loc,end=original_url.end();
	abs_path="";
	if(loc == end) {
		abs_path="/";
		return loc;
		}
	if(*loc != '/') return loc;
	orig = loc;
	while(loc < end && *loc!=';' && *loc!='#' && *loc!='?') {
		loc++;
		}
	abs_path = original_url.substr(orig - original_url.begin(), loc - orig);

	return loc;
	}

string::const_iterator url_string::DecodeRelPath(string::const_iterator loc) const
	{
	rel_path="";
	string::const_iterator orig=loc, end=original_url.end();
	if(loc == end) return loc;
	while(loc < end && *loc!=';' && *loc!='#' && *loc!='?') loc++;
	rel_path = original_url.substr(orig - original_url.begin(), loc - orig);

	if (referrer) abs_path = referrer->get_absolute_path();

	if (abs_path.empty()) {
		abs_path=rel_path;
		return loc;
		}
	if (rel_path.empty()) return loc;
	
	int i=abs_path.size()-1;
	while(i>0 && abs_path[i]!='/') i--;
	if (abs_path[i]!='/') {
		abs_path=rel_path;
		return loc;
		}
	abs_path = abs_path.substr(0,i+1)+rel_path;
	
	return loc;
	}

string::const_iterator url_string::DecodeParams(string::const_iterator loc)  const
	{
	params="";
	string::const_iterator orig=loc, end=original_url.end();
	if(loc == end) return loc;
	while(loc < end && *loc!='?' && *loc !='#') loc++;
	if (loc-orig>0) params = original_url.substr(orig-original_url.begin()+1,loc-orig-1);
	return loc;
	}

string::const_iterator url_string::DecodeQuery(string::const_iterator loc)  const
	{
	query="";
    	string::const_iterator orig=loc, end=original_url.end();
	if(loc == end) return loc;
	while(loc < end && *loc !='#') loc++;
	if (loc-orig>0) query = original_url.substr(orig-original_url.begin()+1,loc-orig-1);
	return loc;
	}

string::const_iterator url_string::DecodeFragment(string::const_iterator loc)  const
	{
	fragment="";
	string::const_iterator orig=loc, end=original_url.end();
	if(loc == end) return loc;
	while(loc < end) loc++;
	if (loc-orig>0) fragment = original_url.substr(orig-original_url.begin()+1,loc-orig-1);
	return loc;
	}

/* Default, do nothing try it as if it were a relative path */
string::const_iterator url_string::HandleAbs(string::const_iterator loc) const
	{
	return loc;
	}

/* We provide a basic sort of relative -> absolute path conversion.
   This is correct for http and probably for most other schemes.
   If this is not the correct one, use a derived class.
   After this, abs_path and url should be correct. */

void url_string::fixup_() const
	{
	//this sets url
	//we've done what we good to find everything belonging to the URL
	//we now recreate it so some one can see what we made of it
	strstream ss1;
	strstream ss2;
	url=scheme;
	if (!username.empty()) {
		ss1<<username;
		if (!passwd.empty()) {
			ss1<<":"<<passwd;
			}
		ss1<<"@";	
		}
	ss1<<net_path;

	if(port!=default_port()) {
		ss1<<":"<<port;
		}
	ss1<<ends;
	url+="://";
	url+=ss1.str();

	// handle ../ pseudo-directories, replace /foo/bar/../foobar with /foo/foobar
	unsigned int uploc, prevslash;
	while((uploc=abs_path.find("/../"))!=abs_path.npos) {
		if(uploc==0) break;
		prevslash=abs_path.rfind("/",uploc-1);
		if(prevslash+1>uploc)
			break; // no more path elements to remove
		abs_path=abs_path.substr(0,prevslash+1)+abs_path.substr(uploc+4);
		}
   // replace /./ elements with /
	while((uploc=abs_path.find("/./"))!=abs_path.npos) {
		abs_path=abs_path.substr(0,uploc)+abs_path.substr(uploc+2);
		}
	
	url+=abs_path;

	if (!params.empty()) {
		ss2<<";"<<params;
		}
	if (!query.empty()) {
		ss2<<"?"<<query;
		}
	if (!fragment.empty()) {
		ss2<<"#"<<fragment;
		}
	ss2<<ends;
	
	url+=ss2.str();	
	}


string url_string::get_url(void) const 
	{
	finish_decoding_();
	return url;
	}

string url_string::get_scheme() const 
	{
	/* Scheme is always decoded */
	finish_decoding_();
	return scheme;
	}

string url_string::get_net_path() const
	{
	finish_decoding_();
	return net_path;
	}

string url_string::get_absolute_path() const
	{
	finish_decoding_();
	return abs_path;
	}

string url_string::get_relative_path() const
	{
	finish_decoding_();
	return rel_path;
	}

string url_string::get_object() const
	{
	unsigned int pos;

	if((pos=abs_path.rfind("/"))<=abs_path.npos) {
		return abs_path.substr(pos+1);
		}
	else return abs_path;
	}

string url_string::get_parameters() const
	{
	finish_decoding_();
	return params;
	}

int url_string::get_port(void) const
	{
	finish_decoding_();
	return port;
	}

string url_string::get_query() const
	{
	finish_decoding_();
	return query;
	}

string url_string::get_fragment() const
	{
	finish_decoding_();
	return fragment;
	}

string url_string::get_username() const
	{
	finish_decoding_();
	return username;
	}

string url_string::get_passwd() const
	{
	finish_decoding_();
	return passwd;
	}

string url_string::get_original() const 
	{
	return original_url;
	}

const  url_string *url_string::get_referrer() const 
	{
	return referrer;
	}

void url_string::set_username(const string& name)
	{
	finish_decoding_();
	username=name;
	}

void url_string::set_passwd(const string& pw)
	{
	finish_decoding_();
	passwd=pw;
	}
