// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Display independent base library for rendering of w3.org defined
  visual layout model. 

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef w3render_hh__
#define w3render_hh__

/* This class takes a DOM Document plus associated StyleSheets, and a
   layout tree (it creates the latter itself). It walks through them,
   updating the layout tree by using boxes which adhere to the
   lib-layout interface, obtained by calling abstract virtual create*
   members. It takes care of style inheritance (we need to think a bit
   about the memory issues involved here).  
*/

#include <dom/dom.hh>
#include <dom/css.hh>
#include <layout/boxes.hh>
#include <layout/fonts.hh>
#include <layout/colors.hh>
#include <layout/factory.hh>
#include <mnemonic/ptr.hh>
#include <mnemonic/stopwatch.hh>

#include <multimap.h>

class w3renderer {
	public:
		w3renderer(ptr<CSS::stylesheet>, DOM::Node *startnode, layout::factory *);
		virtual ~w3renderer();

		void         set_toplevel(layout::container_box *outer, layout::container_box *inner);
		bool         dom_to_layout(const vector<DOM::Node *>& new_nodes);  // true if layout tree was modified
	protected:
		ptr<CSS::stylesheet>             sheet_;
		DOM::Node                       *startnode_;
		layout::factory                 *fac_;
		map<layout::box *, DOM::Node *>  active_boxes_;
		map<string, layout::box *>       anchors_;
		class treeinfo {
			public:
				treeinfo(void);
				
				DOM::Node                       *current_dom_node;
				DOM::Element                    *current_dom_element;     // simple dynamic_cast

				vector<layout::box *>            current_empty_boxes;     // new empty boxes generated
				vector<layout::container_box *>  current_container_boxes; // all new containers generated
				layout::property_box            *current_property_box;

				layout::container_box *          target;                  // is not necessarily a new container
				CSS::declaration::display_value  current_display_mode;

				bool                             is_link;
		};

		enum external_status { failure, success, not_applicable };
		virtual external_status handle_externals_impl(treeinfo&);
	private:
		typedef multimap<DOM::Node *, layout::box *, less<DOM::Node *> >           boxmap_empty;
		typedef pair<DOM::Node *, layout::box *>                                   boxmap_empty_pair;
		typedef multimap<DOM::Node *, layout::container_box *, less<DOM::Node *> > boxmap;
		typedef pair<DOM::Node *, layout::container_box *>                         boxmap_pair;
		typedef multimap<DOM::Node *, layout::property_box *, less<DOM::Node *> >  propmap;
		typedef pair<DOM::Node *, layout::property_box *>                          propmap_pair;

		boxmap       outer_boxes_;  // if empty use inner_boxes_.
		boxmap       inner_boxes_;  // for non-empty elements.
		boxmap_empty empty_boxes_;  // for empty elements.
		propmap      prop_boxes_;   // for pure property elements.

		// The `html' versions handle special HTML tags for which no CSS properties
		// exist (eg. <FONT>).
		void previous_box_info_(DOM::Node::NodeIterator it, treeinfo &);

		void prepare_containers_(treeinfo&);          // setup container boxes and target pointer
		void prepare_containers_html_(treeinfo&);
		void convert_empty_elements_(treeinfo&);      // add any non-container boxes
		void convert_empty_elements_html_(treeinfo&);
		void set_properties_(treeinfo&);              // generate non-geometrical property info
		void set_properties_html_(treeinfo&);

		void append_textboxes_(treeinfo&, const string& txt, bool eatspaces);

		layout::container_box *parent_block_(treeinfo&);
		boxmap_pair            parent_box_(treeinfo&);
		layout::property_box  *parent_properties_(treeinfo&);
		DOM::Node             *previousSibling_mod(DOM::Node *);
		layout::box           *find_inner_box_(DOM::Node *);
		
//		layout::container_box *handle_forms_(DOM::Node::NodeIterator it, layout::container_box *cb, 
//		const treeinfo&);

		bool tree_modified_;

		struct entity_pair {
				char   *entity;
				int8_t  isocode;
		};

		const static entity_pair entities_[];
		int num_entities_;

		bool whitespace_preceding_(DOM::Node *nod);

		stopwatch tim_;
};

#endif
