// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Display independent base library for rendering of w3.org defined
  visual layout model. 

  Copyright 1999  Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// FIXME: also handle BASE tags, eg <base href=http://www.yahoo.co.uk/r/>.

/*
  Tricky bits:

    - Generated parboxes are not associated using boxmap, do we need that?
    - If more than one empty box is generated, only one is stored in boxmap.

  Todo:
    - are pure property boxes properly registered with the property manager 
      so that they get deleted?

 */

#include <w3render/w3render.hh>
#define DEBUG_CLASS "W3RENDER"
#include <mnemonic/debug.hh>
#include <strstream.h>
#include <typeinfo>
#include <algo.h>
 
const w3renderer::entity_pair w3renderer::entities_[] = {
			{ "quot", 34 },
			{ "amp", 38 },
			{ "lt", 60 },
			{ "gt", 62 },
			{ "nbsp", 32 }, 
			{ "Aacute", 0301 },
			{ "aacute", 0341 },
			{ "Acirc",  0302 },
			{ "acirc",  0342 },
			{ "AElig",  0306 },
			{ "aelig",  0346 },
			{ "Agrave", 0300 },
			{ "agrave", 0340 },
			{ "Aring",  0305 },
			{ "aring",  0345 },
			{ "Atilde", 0303 },
			{ "atilde", 0343 },
			{ "Auml",   0304 },
			{ "auml",   0344 },
			{ "Ccedil", 0307 },
			{ "ccedil", 0347 },
			{ "Eacute", 0311 },
			{ "eacute", 0351 },
			{ "Ecirc",  0312 },
			{ "ecirc",  0352 },
			{ "Egrave", 0310 },
			{ "egrave", 0350 },
			{ "ETH",    0320 },
			{ "eth",    0360 },
			{ "Euml",   0313 },
			{ "euml",   0353 },
			{ "Iacute", 0315 },
			{ "iacute", 0355 },
			{ "Icirc",  0316 },
			{ "icirc",  0356 },
			{ "Igrave", 0314 },
			{ "igrave", 0354 },
			{ "Iuml",   0317 },
			{ "iuml",   0357 },
			{ "Ntilde", 0321 },
			{ "ntilde", 0361 },
			{ "Oacute", 0323 },
			{ "oacute", 0363 },
			{ "Ocirc",  0324 },
			{ "ocirc",  0364 },
			{ "Ograve", 0322 },
			{ "ograve", 0362 },
			{ "Oslash", 0330 },
			{ "oslash", 0370 },
			{ "Otilde", 0325 },
			{ "otilde", 0365 },
			{ "Ouml",   0326 },
			{ "ouml",   0366 },
			{ "szlig",  0337 },
			{ "THORN",  0336 },
			{ "thorn",  0376 },
			{ "Uacute", 0332 },
			{ "uacute", 0372 },
			{ "Ucirc",  0333 },
			{ "ucirc",  0373 },
			{ "Ugrave", 0331 },
			{ "ugrave", 0371 },
			{ "Uuml",   0334 },
			{ "uuml",   0374 },
			{ "Yacute", 0335 },
			{ "yacute", 0375 },
			{ "yuml",   0377 },
			{ "reg", 174 },
			{ "copy", 169 } };


w3renderer::w3renderer(ptr<CSS::stylesheet> sheet, DOM::Node *startnode, layout::factory *fac)
	: sheet_(sheet), startnode_(startnode), fac_(fac)
	{
	assert(startnode_!=0);
	num_entities_=sizeof(entities_)/sizeof(entities_[0]);
	debug("startnode is " << startnode->nodeName());
	}

w3renderer::~w3renderer()
	{
	if(fac_) delete fac_;
	}

w3renderer::external_status w3renderer::handle_externals_impl(treeinfo&)
	{
	debug("handle_externals_impl");
	return not_applicable;
	}


w3renderer::treeinfo::treeinfo(void)
	: current_dom_node(0), current_dom_element(0), current_property_box(0), target(0), is_link(false)
	{
	}

void w3renderer::set_toplevel(layout::container_box *outer, layout::container_box *inner)
	{
	assert(inner!=0);
	assert(startnode_!=0);
	inner_boxes_.insert(boxmap_pair(startnode_,inner));
	if(outer)
		outer_boxes_.insert(boxmap_pair(startnode_,outer));
	}

bool w3renderer::dom_to_layout(const vector<DOM::Node *>& new_nodes)
	{
	vector<DOM::Node *>::const_iterator nit=new_nodes.begin();
	tim_.reset();
	tim_.start();
	while(nit!=new_nodes.end()) {
		DOM::Node::NodeIterator it((*nit), true, false);
		while((*it)) {
			if((*it)!=startnode_) {
				treeinfo tinf;
				debug("converting " << (*it)->nodeName());
				// FIXME: this is where the first CSS reference is made, and it fixes the
				// pseudoclasses (link only, for the time being).
				previous_box_info_(it,tinf);
				if(tinf.current_display_mode==CSS::declaration::display_none) {
					debug("display none box, skipping children");
					it.skip_current_children();
					}
				else {
					prepare_containers_(tinf);
					prepare_containers_html_(tinf);
					convert_empty_elements_(tinf);
					convert_empty_elements_html_(tinf);
					handle_externals_impl(tinf);
					set_properties_(tinf);
					set_properties_html_(tinf);
					debug("store info");
					if(tinf.current_container_boxes.size()==0) {
						if(tinf.current_empty_boxes.size()==0) {
							if(tinf.current_property_box==0) {
								debug("=> completely empty element");
								}
							else {
								debug("=> pure property element");
								prop_boxes_.insert(propmap_pair(*it,tinf.current_property_box));
								}
							}
						else {
							debug("=> pure empty element");
							// FIXME: we should really store all empty boxes if there's more than one.
							empty_boxes_.insert(boxmap_empty_pair(*it,tinf.current_empty_boxes.back()));
							}
						}
					else {
						debug("=> container element");
						inner_boxes_.insert(boxmap_pair(*it,tinf.current_container_boxes.back()));
						if(tinf.current_container_boxes.size()>1) {
							outer_boxes_.insert(boxmap_pair(*it,tinf.current_container_boxes[0]));
							}
						if(tinf.current_empty_boxes.size()!=0) {
							debug("with empty inner box");
							empty_boxes_.insert(boxmap_empty_pair(*it,tinf.current_empty_boxes.back()));
							}
						}
					}
				}
			++it;
			}
		++nit;
		}

	debug("dom to layout done");
	tim_.stop();
	debug(tim_);
	return true; //tree_modified_;
	}
			
void w3renderer::previous_box_info_(DOM::Node::NodeIterator it, treeinfo& inf)
	{
	assert((*it)!=0);

	// FIXME: retrieving the style for a node is probably best done in a more memory
	// conservative way: just have the CSS create a new style object and keep a pointer
	// to it in `inf'. Then when we're done, delete it. Also saves on lookup in css.cc.
	inf.current_dom_node=(*it);
	inf.current_dom_element=dynamic_cast<DOM::Element *>(*it);
	if(inf.current_dom_node->nodeName()=="A" &&
		inf.current_dom_element->getAttribute("HREF")!="") {
		inf.is_link=true;
		inf.current_display_mode=sheet_->style((*it),CSS::selector::selector_element::link)->get_display();
		}
	else {
		inf.current_display_mode=sheet_->style((*it))->get_display();
		}
	}

layout::container_box *w3renderer::parent_block_(treeinfo& tinf)
	{
	DOM::Node *nod=tinf.current_dom_node;
	while(nod->parentNode()) {
		nod=nod->parentNode();
		if(nod==startnode_ 
			|| sheet_->style(nod)->get_display()==CSS::declaration::block
			|| sheet_->style(nod)->get_display()==CSS::declaration::table_cell) {
			boxmap::iterator it=inner_boxes_.find(nod);
			return (*it).second;
			}
		}
	return 0;
	}

w3renderer::boxmap_pair w3renderer::parent_box_(treeinfo& tinf)
	{
	DOM::Node *nod=tinf.current_dom_node;
	while(nod->parentNode()) {
		nod=nod->parentNode();
		boxmap::iterator it=inner_boxes_.find(nod);
		if(it!=inner_boxes_.end()) 
			return (*it);
		}
	assert(1==0);
	}

layout::box *w3renderer::find_inner_box_(DOM::Node *nod)
	{
	boxmap_empty::iterator et=empty_boxes_.find(nod);
	if(et==empty_boxes_.end()) {
		boxmap::iterator it=inner_boxes_.find(nod);
		if(it==inner_boxes_.end()) return 0;
		else                       return (*it).second;
		}
	else return (*et).second;
	}

layout::property_box *w3renderer::parent_properties_(treeinfo& tinf)
	{
	DOM::Node *nod=tinf.current_dom_node;
	boxmap::iterator it2;
	do {
		assert(nod->parentNode()!=0);
		nod=nod->parentNode();
		propmap::iterator it1=prop_boxes_.find(nod);
		if(it1!=prop_boxes_.end()) {
			debug("getting properties from " << nod->nodeName());
			return (*it1).second;
			}
		
		it2=inner_boxes_.find(nod);
		} while(it2==inner_boxes_.end());

	debug("getting properties from " << nod->nodeName());
	return (*it2).second->properties();
	}

DOM::Node *w3renderer::previousSibling_mod(DOM::Node *current)
	{
	// foo <em>bar foo</em> bar
	DOM::Node *previous=current->previousSibling();
	while(true) {
		if(previous==0) {
			if(current->parentNode()==0) break;
			if(!(prop_boxes_.find(current->parentNode())==prop_boxes_.end())) {
				current=current->parentNode();
				previous=current->previousSibling();
				}
			else { 
				if(empty_boxes_.find(current->parentNode())==empty_boxes_.end() &&
					inner_boxes_.find(current->parentNode())==inner_boxes_.end() ) { // parent is none
					current=current->parentNode();
					previous=current->previousSibling();
					}
				else
					break;
				}
			}
		else {
			if(!(prop_boxes_.find(previous)==prop_boxes_.end()) ) {	// previous is property box
				if(previous->lastChild()) {
					previous=previous->lastChild();
					current=previous;
					}
				else {
					current=previous;
					previous=previous->previousSibling();
					}
				}
			else {
				if(empty_boxes_.find(previous)==empty_boxes_.end() &&
					inner_boxes_.find(previous)==inner_boxes_.end() ) { // previous is none element
					if(previous->lastChild()) {
						previous=previous->lastChild();
						current=previous;
						}
					else {
						current=previous;
						previous=previous->previousSibling();
						}
					}
				else
					break;
				}
			}
		}
	return previous;
	}

void w3renderer::prepare_containers_(treeinfo& tinf)
	{
	switch(tinf.current_display_mode) {
		case CSS::declaration::block: {
			debug("block");
			tinf.target=parent_block_(tinf);
			assert(tinf.target!=0);

			CSS::declaration::numberproperty *num_left=
				sheet_->style(tinf.current_dom_node)->get_number_property(CSS::declaration::margin_left);
			CSS::declaration::numberproperty *num_right=
				sheet_->style(tinf.current_dom_node)->get_number_property(CSS::declaration::margin_right);
			CSS::declaration::numberproperty *num_top=
				sheet_->style(tinf.current_dom_node)->get_number_property(CSS::declaration::margin_top);
			CSS::declaration::numberproperty *num_bottom=
				sheet_->style(tinf.current_dom_node)->get_number_property(CSS::declaration::margin_bottom);
			layout::hbox *tmph=0;
			if(num_top!=0) {
				layout::glue *tmp=fac_->create_glue();
				tinf.target->add_child(tmp);
				tmp->set_absolute_size(0,num_top->value(),0);
				}

			if(num_left!=0 || num_right!=0) { // margin settings
				tmph=fac_->create_hbox();
				tinf.target->add_child(tmph);
				tinf.current_container_boxes.push_back(tmph);
				tinf.target=tmph;
				}
			if(num_left!=0) {
				layout::glue *tmp1=fac_->create_glue();
				tmph->add_child(tmp1);
				tmp1->set_absolute_size(num_left->value(),0,0);
				}
			layout::vbox *tmpv=0;
			if(sheet_->style(tinf.current_dom_node)->get_border_top_style()!=
				CSS::declaration::border_style_none) {
				CSS::declaration::numberproperty *bwidth=
					sheet_->style(tinf.current_dom_node)->get_number_property(CSS::declaration::border_width);
				debug("making border box");
				tmpv=fac_->create_border_box();

				if(bwidth)
					tmpv->set_size_mixed(1.0,(int)(bwidth->value()));
				else
					tmpv->set_size_mixed(1.0,2);
				}
			else tmpv=fac_->create_vbox();

			tinf.target->add_child(tmpv);
			tinf.current_container_boxes.push_back(tmpv);
			tmpv->set_properties(parent_properties_(tinf));

			if(sheet_->style(tinf.current_dom_node)->get_text_align()==CSS::declaration::align_center) {
				debug("element " << tinf.current_dom_node->nodeName() << " children should be centered");
				tmpv->set_size_fraction(1.0,0.0);
				tmpv->set_center_content(true);
				}

			if(num_right!=0) {
				layout::glue *tmp1=fac_->create_glue();
				tmph->add_child(tmp1);
				tmp1->set_absolute_size(num_right->value(),0,0);
				}

			if(num_bottom!=0) {
				layout::glue *tmp=fac_->create_glue();
				tinf.target->add_child(tmp);
				tmp->set_absolute_size(0,num_bottom->value(),0);
				}
			tinf.target=tmpv;
			break;
			}
		case CSS::declaration::display_inline: {
			debug("inline");
			DOM::Node *prevnode=previousSibling_mod(tinf.current_dom_node);
			if(prevnode!=0) {
				if(sheet_->style(prevnode)->get_display()!=CSS::declaration::display_inline) {
					layout::container_box *target=parent_block_(tinf);
					layout::parbox *pb=fac_->create_parbox();
					target->add_child(pb);
					debug("parent is " << tinf.current_dom_node->parentNode()->nodeName());
				   pb->set_properties(parent_properties_(tinf));
					tinf.current_container_boxes.push_back(pb);
					tinf.target=pb;
					}
				else {
					debug("storing in parent of prevnode " << prevnode->nodeName() << " " << prevnode->nodeValue());
					layout::box *tmp=find_inner_box_(prevnode);
					assert(tmp!=0);
					tinf.target=tmp->parent();
					assert(tinf.target!=0);
					}
				}
			else {
				boxmap_pair target=parent_box_(tinf);
				if(sheet_->style(target.first)->get_display()==CSS::declaration::display_inline) {
					debug("putting in parent inline");
					debug("which is " << target.first->nodeName());
					tinf.target=target.second;
					assert(tinf.target!=0);
					}
				else {
					layout::parbox *pb=fac_->create_parbox();
					target.second->add_child(pb);
					debug("parent is " << tinf.current_dom_node->parentNode()->nodeName());
				   pb->set_properties(parent_properties_(tinf));
					tinf.target=pb;
					}
				}
			break;
			}
		case CSS::declaration::display_empty:
			debug("empty");
			break;
		case CSS::declaration::table: {
			assert(tinf.current_dom_element!=0);
			tinf.target=parent_block_(tinf);
			assert(tinf.target!=0);
			string a_width=tinf.current_dom_element->getAttribute("WIDTH");
 			float  v_width=0;
 			int32_t v_iwidth=0;
 			layout::table *tmp2=fac_->create_table();
 			tinf.target->add_child(tmp2);
 			tinf.target=tmp2;  
			tinf.current_container_boxes.push_back(tmp2);
			tmp2->set_properties(parent_properties_(tinf));

 			if(a_width  !="") {
 				if(a_width[a_width.size()-1]!='%') {
 					debug("table has integer width " << a_width);
 					v_iwidth=atoi(a_width.c_str());
 					tinf.target->set_absolute_size(v_iwidth,0,0);
 					}
 				else {
 					debug("table has percentage width " << a_width);
 					v_width  =atoi(a_width.c_str())/100.0;
 					tinf.target->set_size_fraction(v_width,0);
 					}
				}
			break; 
			}
		case CSS::declaration::table_row: {
			tinf.target=parent_box_(tinf).second;
			layout::table *tb=dynamic_cast<layout::table *>(tinf.target);
			assert(tb!=0);
			tb->end_row();
			tinf.target=tb;
			break;
			}
		case CSS::declaration::table_cell: {
			tinf.target=parent_box_(tinf).second;
			layout::table *tb=dynamic_cast<layout::table *>(tinf.target);
			if(tb==0) throw logic_error("table cell without table row as parent");
			assert(tinf.current_dom_element!=0);
			string a_colspan=tinf.current_dom_element->getAttribute("COLSPAN");
			string a_rowspan=tinf.current_dom_element->getAttribute("ROWSPAN");
			string a_width=tinf.current_dom_element->getAttribute("WIDTH");
			string a_valign=tinf.current_dom_element->getAttribute("VALIGN");
			int8_t v_colspan=1, v_rowspan=1;
			float  v_width=0;
			int32_t v_iwidth=0;
			if(a_colspan!="") v_colspan=atoi(a_colspan.c_str());
			if(a_rowspan!="") v_rowspan=atoi(a_rowspan.c_str());
			if(a_width  !="") {
				if(a_width[a_width.size()-1]!='%') {
					debug("cell has integer width " << a_width);
					v_iwidth=atoi(a_width.c_str());
					}
				else {
					debug("cell has percentage width " << a_width);
					v_width  =atoi(a_width.c_str())/100.0;
					}
				}
			layout::table::valign_type vt=layout::table::align_bottom;
			transform(a_valign.begin(),a_valign.end(),a_valign.begin(),toupper);
			if(a_valign=="TOP") vt=layout::table::align_top;

			// Handle width settings and colors.
			layout::vbox *cb=fac_->create_vbox();
			if(v_width!=0) {
				tb->add_child(cb, v_colspan,v_rowspan,vt);
				cb->set_size_fraction(v_width,0);
				}
			else if(v_iwidth!=0) {
				tb->add_child(cb, v_colspan,v_rowspan,vt);
				cb->set_absolute_size(v_iwidth,0,0);
				}
			else {
				tb->add_child(cb, v_colspan,v_rowspan,vt);
				}
			tinf.target=cb;
			tinf.current_container_boxes.push_back(tinf.target);
			cb->set_properties(parent_properties_(tinf));

			break;
			}
		case CSS::declaration::table_caption: {
/*			layout::table *tb=dynamic_cast<layout::table *>(tinf.parent_layout_node);
			if(tb==0) throw logic_error("table caption without table row as parent");
			cb=fac_->create_vbox();
			tb->add_child(cb, 1,1,layout::table::align_bottom); */
			break; 
			} 
		default:
			debug("unsupported CSS display type " << (int)tinf.current_display_mode);
			throw logic_error("unsupported CSS display type encountered");
		}
	}

void w3renderer::prepare_containers_html_(treeinfo& inf)
	{
	if(inf.current_dom_node->nodeName()=="A") {
		if(inf.current_dom_element->getAttribute("HREF")!="") {
			debug("adding event box");
			assert(inf.target!=0);
			layout::event_box *tmp=fac_->create_event_box();
			inf.target->add_child(tmp);
			inf.current_container_boxes.push_back(tmp);
			inf.target=tmp;
			debug("event box added");
			active_boxes_[tmp]=inf.current_dom_node;
			}
		}
	else if(inf.current_dom_node->nodeName()=="FRAME") {
		throw logic_error("frames not yet supported");
      //	layout::viewer_box *tmp=fac_->create_viewer_box();
		}
	}

void w3renderer::convert_empty_elements_(treeinfo& inf)
	{
	bool eatspace=false;
//	DOM::Node *prevnode=previousSibling_mod(inf.current_dom_node);
/*	if(prevnode!=0) {
		if(sheet_->style(prevnode)->get_display()!=CSS::declaration::display_inline)
			eatspace=true;
			} */
	// FIXME: the above no longer seems to be necessary?!? That's cool!
	// FIXME: this is the place where ':before' and ':after' pseudo classes
	// should be handled. It's now hardcoded for BR. Note that all elements
	// declared as such should also be inline (is that true, doesn't block work too?)
	if(inf.current_dom_node->nodeName()=="BR") {
		debug("appending newline");
		append_textboxes_(inf, "\x0a", eatspace);
		}
	else if(inf.current_dom_node->nodeName()=="#text") {
		debug("appending text boxes");
		append_textboxes_(inf, inf.current_dom_node->nodeValue(), eatspace);
		}
	}

void w3renderer::convert_empty_elements_html_(treeinfo& inf)
	{
	}

/*
<span style="white-space: nowrap">This is a long line of text I don't want
broken.</span>

 */

void w3renderer::append_textboxes_(treeinfo& inf, const string& txt, bool eatspaces)
	{
	const char *bigone=txt.c_str();
	long    totalsize=txt.size();
	string  tmpstr;
	int     curpos=0;

	debug("appending text " << txt);

	while(curpos < totalsize) {
		tmpstr.erase();
		while(curpos < totalsize && bigone[curpos]!=' ' && bigone[curpos]!='\n') {
			if(bigone[curpos]=='&') {
				int tmppos=curpos;
				++curpos;
				if(bigone[curpos]=='#') {
					tmpstr+=atoi(&(bigone[curpos+1]));
					}
				else {
					for(int i=0; i<num_entities_; ++i) {
						int len=strlen(entities_[i].entity);
						if(curpos+len+1 <= totalsize && bigone[curpos+len]==';' && 
							!memcmp(&(bigone[curpos]),entities_[i].entity,len)) {
							tmpstr+=entities_[i].isocode;
							break;
							}
						}
					}
				while(bigone[curpos]!=';') {
					if((++curpos)==totalsize) {
						curpos=tmppos;
						tmpstr+=bigone[curpos];
						break;
						}
					}
				}
			else {
				tmpstr+=bigone[curpos];
				}
			++curpos;
			}
		++curpos;
		if(tmpstr.size()==0 && (bigone[curpos-1]==' ' || bigone[curpos-1]=='\n')) {
			if(!eatspaces) { // this is not the first character in the parbox
				layout::box *tmp;
				debug("white " << curpos << " " << (int)(bigone[curpos-1]));
				if(bigone[curpos-1]==' ') {
					debug("adding space glue");
// Because glue is mandatory_break_discard, the first space will be eaten by the 
// layout engine. Since spaces at the beginning of a node should be kept anyway,
// we use textboxes instead.
					tmp=fac_->create_glue(); 
//					tmp=fac_->create_text_box(" ");
					}
				else {
					debug("adding newline");
					tmp=fac_->create_text_box("\x0a");
					}

				assert(inf.target);
				inf.target->add_child(tmp);
				inf.current_empty_boxes.push_back(tmp);
				} 
			else { 
				debug("eating spaces");
				}
			}
		else {
			if(tmpstr.size()>0) {
				layout::box *tmp=fac_->create_text_box(tmpstr);
				assert(inf.target);
				inf.target->add_child(tmp);
				inf.current_empty_boxes.push_back(tmp);
				if(curpos<totalsize || bigone[curpos-1]==' ' || bigone[curpos-1]=='\n') {
					if(bigone[curpos-1]==' ') inf.target->add_child(fac_->create_glue());
					else                      inf.target->add_child(fac_->create_text_box("\x0a"));
					}
				}
			}
		}
	}


void w3renderer::set_properties_(treeinfo& inf)
	{
	CSS::stylesheet::node_style *ns=sheet_->style(inf.current_dom_node);
	// looks time-consuming but actual time taken is negligible:
	CSS::declaration::colorproperty    *cp=ns->get_color();
	CSS::declaration::colorproperty    *cpb=ns->get_color_property(CSS::declaration::background_color);
	CSS::declaration::stringproperty   *sp=ns->get_font_family();
	CSS::declaration::numberproperty   *np=ns->get_font_size();
	CSS::declaration::font_style_value  styl=ns->get_font_style();
	CSS::declaration::font_weight_value weig=ns->get_font_weight();

	layout::property_box *parentprop=0;
	if(inf.current_container_boxes.size()==0) parentprop=parent_properties_(inf);
	else                                   parentprop=inf.current_container_boxes.back()->properties();

	layout::xlfd fontnam=parentprop->current_font()->xl();

	bool nofont=true;
	// FIXME: handle exceptions for creation of layout::length objects.
	if(sp || np || weig!=CSS::declaration::weight_normal || styl!=CSS::declaration::font_normal ) {
		if(sp) {
			fontnam.family=sp->value();
			nofont=false;
			}
		if(weig==CSS::declaration::weight_bold) {
			fontnam.weight=layout::xlfd::bold;
			nofont=false;
			}
		if(styl==CSS::declaration::font_oblique || styl==CSS::declaration::font_italic) {
			fontnam.slant=layout::xlfd::oblique;
			nofont=false;
			}
		if(np) {
			fontnam.size=layout::length(np->str());
			nofont=false;
			}
		}

	if(!cp && !cpb && nofont) {
		debug("no new font/color info, inheriting");
		if(inf.current_empty_boxes.size()!=0) {
			vector<layout::box *>::iterator ei=inf.current_empty_boxes.begin();
			while(ei!=inf.current_empty_boxes.end()) {
				(*ei)->set_properties(parentprop);
				++ei;
				}
			}
		else if(inf.current_container_boxes.size()!=0) {
			inf.current_container_boxes.back()->set_properties(parentprop);
			}
		}
	else {
		string    colornam, bgcolornam;
		if(cp)  colornam=   cp->colorName();
		if(cpb) bgcolornam= cpb->colorName();
		
		layout::font  *tmp1=0;
		layout::color *tmp2=0, *tmp3=0;
		layout::container_box *cb=parent_block_(inf);
		if(!nofont)        tmp1=cb->font_mgr().get_font(fontnam);
		if(colornam!="")   tmp2=cb->color_mgr().get_color(colornam);
		if(colornam!="") {
			debug("setting color to " << colornam);
			}
		if(bgcolornam!="") tmp3=cb->color_mgr().get_color(bgcolornam);
		
		inf.current_property_box=fac_->create_property_box(cb->properties(),tmp1,tmp2,tmp3);

		if(inf.current_empty_boxes.size()!=0) {
			vector<layout::box *>::iterator ei=inf.current_empty_boxes.begin();
			inf.current_property_box->set_inherit(parentprop);
			while(ei!=inf.current_empty_boxes.end()) {
				(*ei)->set_properties(inf.current_property_box);
				++ei;
				}
			}
		else if(inf.current_container_boxes.size()!=0) {
			// FIXME: this assumes that the properties were duped correctly already!
			inf.current_property_box->set_inherit(inf.current_container_boxes.back()->properties());
			inf.current_container_boxes.back()->set_properties(inf.current_property_box);
			}
		else {
			inf.current_property_box->set_inherit(parentprop);
			}
		}
//	if(inf.current_dom_element->getAttribute("NAME")!="") {
//				anchors_[inf.current_dom_element->getAttribute("NAME")]=cb;
//		}
	}	

void w3renderer::set_properties_html_(treeinfo& inf)
	{
	if(inf.current_dom_node->nodeName()=="BODY") {
		string fgcolor=inf.current_dom_element->getAttribute("TEXT");
		string bgcolor=inf.current_dom_element->getAttribute("BGCOLOR");
		if(bgcolor=="") {
			CSS::declaration::colorproperty *cpb=sheet_->style(inf.current_dom_node)
				->get_color_property(CSS::declaration::background_color);
			if(cpb!=0) bgcolor=cpb->colorName();
			}
		// FIXME: we walk up to the uppermost box, but is that correct?
		assert(inf.current_container_boxes.size()!=0);
		layout::container_box *tmp=inf.current_container_boxes.back();
		while(tmp->parent())
			tmp=tmp->parent();
		
		tmp->set_fill_background(true);
		tmp->set_properties(fac_->create_property_box(tmp->properties(),
																	0,
																	fgcolor!=""?tmp->color_mgr().get_color(fgcolor):0,
																	bgcolor!=""?tmp->color_mgr().get_color(bgcolor):0));
		return;
		}

	// The following are all treated similarly: tag-dependent parsing of attributes and then
	// a generic thing at the end to set the properties on the layout box.
	// FIXME: it's not particularly efficient yet...

	string color, font, siz, bgcolor;
	
	if(inf.current_dom_node->nodeName()=="FONT") {
		color=inf.current_dom_element->getAttribute("COLOR");
		font =inf.current_dom_element->getAttribute("NAME");
		siz  =inf.current_dom_element->getAttribute("SIZE");
		}
	else if(inf.current_dom_node->nodeName()=="TD" ||
			  inf.current_dom_node->nodeName()=="TH" ) {
		bgcolor=inf.current_dom_element->getAttribute("BGCOLOR");
		if(bgcolor=="") {
			// FIXME: this is of course a mess. How does CSS define inheritance of attributes?
			DOM::Node *tab=inf.current_dom_node;
			while(tab->parentNode()!=0 && bgcolor=="") {
				tab=tab->parentNode();
				DOM::Element *table_el=dynamic_cast<DOM::Element *>(tab);
				if(table_el)
					bgcolor=table_el->getAttribute("BGCOLOR");
				}
			}
		}
	else if(inf.current_dom_node->nodeName()=="TABLE") {
		bgcolor=inf.current_dom_element->getAttribute("BGCOLOR");
		}
	else if(inf.current_dom_node->nodeName()=="A") {
		if(inf.current_dom_element->getAttribute("HREF")!="") {
			// FIXME: this is a waste of time, better buffer it.
			DOM::Node *tab=inf.current_dom_node->parentNode();
			while(tab && tab->nodeName()!="BODY")
				tab=tab->parentNode();
			assert(tab!=0);
			DOM::Element *el=dynamic_cast<DOM::Element *>(tab);
			color=el->getAttribute("LINK");
			}
		}

	// The properties have been collected, so now create a property box.

	layout::property_box *parentprop=0;
	if(inf.current_container_boxes.size()==0) parentprop=parent_properties_(inf);
	else                                      parentprop=inf.current_container_boxes.back()->properties();
	layout::xlfd fnt=parentprop->current_font()->xl();

	if(font!="") fnt.family=font;
	if(siz!="")  {
		layout::length num(siz);
		if(num.unit()==layout::length::px) {
			// Size is relative to body, so look up that node and get font size
			// from stylesheet. FIXME: 'SIZE=+1' should be relative...
			int offset=3;
			if(siz.find_first_of("+-")!=siz.npos) offset=0;
			DOM::Node *tmp=inf.current_dom_node;
			while(tmp->nodeName()!="BODY") {
				tmp=tmp->parentNode();
				assert(tmp!=0);
				}
			CSS::declaration::numberproperty *np=sheet_->style(tmp)->get_font_size();
			assert(np!=0);
			fnt.size=layout::length(np->str());
			fnt.size=layout::length(fnt.size.number()+2*(num.number()-offset),fnt.size.unit());
			// Stupid check to avoid negative sizes.
			if(fnt.size.pixels(95.0)<8) fnt.size=layout::length(8,layout::length::px);
			}
		else {
			fnt.size=num;
			}
		}

	if(color=="" && font=="" && siz=="" && bgcolor=="") {
		if(inf.current_empty_boxes.size()!=0) {
			vector<layout::box *>::iterator ei=inf.current_empty_boxes.begin();
			while(ei!=inf.current_empty_boxes.end()) {
				(*ei)->set_properties(parentprop);
				++ei;
				}
			}
		else if(inf.current_container_boxes.size()!=0) {
			inf.current_container_boxes.back()->set_properties(parentprop);
			}
		}
	else {
		layout::container_box *par=parent_block_(inf);
		inf.current_property_box=fac_->create_property_box(
			parentprop,
			(font!="" || siz!="")?par->font_mgr().get_font(fnt):0,
			color!=""?par->color_mgr().get_color(color):0,
			bgcolor!=""?par->color_mgr().get_color(bgcolor):0);
		
		if(inf.current_empty_boxes.size()!=0) {
			vector<layout::box *>::iterator ei=inf.current_empty_boxes.begin();
			inf.current_property_box->set_inherit(parentprop);
			while(ei!=inf.current_empty_boxes.end()) {
				(*ei)->set_properties(inf.current_property_box);
				++ei;
				}
			}
		else if(inf.current_container_boxes.size()!=0) {
			// FIXME: this assumes that the properties were duped correctly already!
			inf.current_property_box->set_inherit(inf.current_container_boxes.back()->properties());
			inf.current_container_boxes.back()->set_properties(inf.current_property_box);
			if(bgcolor!="") 
				inf.current_container_boxes.back()->set_fill_background(true);
			}
		else {
			inf.current_property_box->set_inherit(parentprop);
			}
		}
	}


/*container_box *w3renderer::handle_forms_(DOM::Node::NodeIterator it, container_box *cb, const treeinfo& tinf)
	{
	if((*it)->nodeName()=="INPUT") {
		DOM::Element *el=dynamic_cast<DOM::Element *>((*it));
		string typ=el->getAttribute("BGCOLOR");
		switch(typ) {
			case "text":
			case "password":
			case "checkbox":
			case "radio":
			case "submit":
			case "reset":
			case "file":
			case "hidden":
			case "image":
			case "button":
			}
		} 
		}	*/


bool w3renderer::whitespace_preceding_(DOM::Node *nod)
	{
	assert(nod!=0);

	if((nod=nod->previousSibling())==0) return true;
	// FIXME: should use CSS info to see if its non-white.
	if(nod->nodeName()!="#text") return false;

	for(unsigned int i=0; i<nod->nodeValue().size(); ++i) {
		if(nod->nodeValue()[i]!=' ')
			return false;
		}
	return whitespace_preceding_(nod);
	}

