// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Message announcing creation of new gtk widgets.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <msg/gtkdisplay/object_ready.hh>
#include <unistd.h>
#include <fcntl.h>

#define DEBUG_CLASS "OBJECT_READY"
#include <mnemonic/debug.hh>

msg_gtk_object_ready::msg_gtk_object_ready(void)
	: widget(0), layoutbox(0), new_layoutbox(0)
	{
	}

		/* The oil which is in charge of the Gtk main loop can be interrupted by 
		   writing to the fd below and using the mutexes.

          display_mutex is unlocked when the event loop is running. 
          client_mutex  is unlocked when display thread is NOT accessing gtk

         Schematically: 

          start with:
					gtk_in_use_.lock();
					display_working_.unlock();

           client:
               start loop
               trigger start of display
               wait for signal on pipe
               interrupt loop
					flush pipe
					event_loop_not_restarted=true;
					gtk_in_use_.wait();
					  waiting_.lock();         // for this, the display has to have called wait()
					  waiting_.broadcast();
					  waiting_.unlock();


           display: (NOTE: there can be more than one display!)
               display_working_.lock();
               write to pipe
               gtk_in_use_.lock();
					  waiting_.lock();     // we will always get this
               do gtk stuff
					gtk_in_use_.broadcast();
					gtk_in_use_.unlock();          
					// Here, we should still wait until we are sure the interrupted event loop
					// has picked up our signal and has restarted gtk stuff. Otherwise, another
					// display might go through this entire cycle quickly before the event
					// loop has even had a chance to see the broadcast and restart. So:
					  waiting_.wait();
				     waiting_.unlock();
					display_working_.unlock();

		*/

Threadsafe_Main *Threadsafe_Main::instance_=0;

Threadsafe_Main::Threadsafe_Main(int *argc, char ***argv, bool have_locale=false)
	: main_(argc, argv, have_locale)
	{
	init_();
	}	

Threadsafe_Main::Threadsafe_Main(int &argc, char **&argv, bool have_locale=false)
	: main_(argc, argv, have_locale)
	{
	init_();
	}

void Threadsafe_Main::init_(void)
	{
	fd_[0]=-1;
	if(instance_==0) 
		instance_=this;
	else {
		cerr << "Threadsafe_Main initialised twice!" << endl;
		}
	set_interrupt_handler_();
	}

Gtk::Main& Threadsafe_Main::base(void)
	{
	return main_;
	}

Threadsafe_Main *Threadsafe_Main::instance(void)
	{
	return instance_;
	}

void Threadsafe_Main::set_interrupt_handler_(void)
	{
	if(fd_[0]==-1) {
		if(pipe(fd_)) {
			cerr << "Threadsafe_Main: cannot create signal pipe" << endl;
			return;
			}
		if(fcntl(fd_[0], F_SETFL, O_NONBLOCK)) {
			cerr << "Threadsafe_Main: cannot set O_NONBLOCK on signal pipe" << endl;
			return;
			}
		gtk_in_use_.lock();
		gdk_input_add(fd_[0],GDK_INPUT_READ, interrupt_event_loop_, this);
		}
	else {
		cerr << "Threadsafe_Main: interrupt handler already set" << endl;
		}
	}

void Threadsafe_Main::interrupt_event_loop_(void *arg, int fd, GdkInputCondition cond)
	{
	debug("someone interrupted the event loop");
	assert(arg!=0);
//	gdk_threads_leave();

//	stopwatch sw;
//	sw.start();
	((Threadsafe_Main *)arg)->wait_for_restart_();
//	sw.stop();
//	debug("spent " << sw << " outside event loop");
//	gdk_threads_enter();
	debug("back to gtk event loop");
	}

void Threadsafe_Main::interrupt_event_loop(void)
	{
	display_working_.lock();
	write(fd_[1],"a",1);
	gtk_in_use_.lock();
	waiting_.lock(); 
	}

void Threadsafe_Main::restart_event_loop(void)
	{
	gtk_in_use_.broadcast();
	gtk_in_use_.unlock();
	waiting_.wait();
	waiting_.unlock();
	display_working_.unlock();
	}

void Threadsafe_Main::wait_for_restart_(void)
	{
	static char buf[10];
	// clean the pipe
	while(read(fd_[0],buf,10)>0);
	gtk_in_use_.wait();
	waiting_.lock();
	waiting_.broadcast();
	waiting_.unlock();
	}


untyped_message message<msg_gtk_object_ready>::static_part;
template message<msg_gtk_object_ready>;
template interest_base<msg_gtk_object_ready>;
