// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Raw data to CSS DOM-like tree converter.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define DEBUG_CLASS "CSSPARSER"
#include <mnemonic/debug.hh>
#include <mnemonic/stopwatch.hh>
#include "cssparser.hh"
#include <fstream.h>

//#define VERBOSE

cssparser::cssparser(ptr<CSS::stylesheet> style)
	: keep_whitespace_(-1), offset_in_stream_(0), lines_in_stream_(0),
	  stylesheet_(style), current_selector_(0), current_declaration_(0)
	{
	history_element tmp(INITIAL, 0, 0);
	parse_history.push_back(tmp);
	fill_string2enum_tables_();
	}

cssparser::~cssparser()
	{
	}

// FIXME: for now, we only do HTML 4.0.
void cssparser::setup_default(void)
	{
//	stopwatch sw;
//	sw.start();
	string sheetname=PREFIX"/lib/mnemonic/stylesheets/html40.css";
	if(getenv("MN_DEFAULT_STYLESHEET"))
		sheetname=getenv("MN_DEFAULT_STYLESHEET"); 
	ifstream defsheet(sheetname.c_str());
	if(defsheet.is_open()) {
		while(defsheet) {
			defsheet >> (*this);
			parse_buffer();
			}
		}
	else {
		cerr << "cannot open default stylesheet" << endl;
		}
//	sw.stop();
//	cerr << "loading default stylesheet took " << sw << endl;
	}

void cssparser::append_data(const string& dat)
	{
	raw_data_+=dat;
	}

css_parse_error::css_parse_error(const string& wh, error_type et)
	: logic_error(wh), error(et)
	{
	}

ostream& operator<<(ostream& str, const css_parse_error& er)
	{
	str << "[css_parse_error]: "; 
	str << er.what() << " ";
	switch(er.error) {
		case css_parse_error::UNKNOWN:
			str << "Generic CSS parse error.";
			break;
		default:
			str << "[empty description]";
			break;
		}
	return str;
	}

istream& operator>>(istream& str, cssparser& pars)
	{
	int i=0;
	while(str && i<2000) {
		++i;
		pars.raw_data_+=str.get();
		}
//	debug(pars.raw_data_);
	return str;
	}

cssparser::history_element::history_element(cssparser::parse_state stat, long where, long line)
	: state(stat), offset_in_stream(where), line_in_stream(line)
	{
	}

cssparser::parse_result cssparser::parse_buffer(void)
	{
	history_element hist=parse_history.back();
	// FIXME: can't we just use the hist members?
	unsigned long current_offset=hist.offset_in_stream;
	unsigned long current_line=hist.line_in_stream;

	if(hist.state==INTERRUPTED)
		parse_history.pop_back();

	bool          exit_loop=false;
	parse_result  return_code=OK;
	unsigned char curchar, prevchar;
	curchar=255;

	unsigned long totalsize=raw_data_.size();
	while(!exit_loop && current_offset-offset_in_stream_ < totalsize) {
		prevchar=curchar;
	   curchar=raw_data_[current_offset-offset_in_stream_];
		if(tokenise_(curchar)==255) {
			++current_offset;
			continue;
			}
		if(curchar=='\r')
			++current_line;
		if(curchar=='\n') {
			if(prevchar=='\r') {
				++current_offset;
				continue;
				}
			else ++current_line;
			}

		// Take a deep breath and dive into the state machine logic:
		switch(peek_(0).state) {
			case INITIAL:
#ifdef VERBOSE
				debug("INITIAL mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						++current_offset;
						break;
					case '/':
						parse_history.push_back(history_element(SLASH, current_offset, current_line));
						++current_offset;
						break;
					default:
						selector_elements_.clear();
						selector_="";
						pseudo_class_="";
						parse_history.push_back(history_element(SELECTOR, current_offset, current_line));
						accumulator_+=curchar;
						++current_offset;
						break;
					};
				break;
			case SLASH:
#ifdef VERBOSE
				debug("SLASH mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '*':
						parse_history.push_back(history_element(COMMENT, current_offset, current_line));
						accumulator_="";
						++current_offset;
						break;
					default:
						parse_history.pop_back();
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case COMMENT:
#ifdef VERBOSE
				debug("COMMENT mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '*':
						parse_history.push_back(history_element(MAYBE_COMMENT_END, current_offset, current_line));
						++current_offset;
						break;
					default:
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case MAYBE_COMMENT_END:
#ifdef VERBOSE
				debug("MAYBE_COMMENT_END mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '/':
//						debug("scanned comment |" << accumulator_ << "|");
						accumulator_="";
						parse_history.pop_back();
						parse_history.pop_back();
						parse_history.pop_back();
						++current_offset;
						break;
					default:
						accumulator_+='*';
						accumulator_+=curchar;
						parse_history.pop_back();
						++current_offset;
						break;
					}
				break;
			case EAT_WHITESPACE:
#ifdef VERBOSE
				debug("EAT_WHITESPACE mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						++current_offset;
						break;
					default:
						parse_history.pop_back();
						break;
					}
				break;
			case SELECTOR:
#ifdef VERBOSE
				debug("SELECTOR mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case ':':
						pseudo_class_="";
						parse_history.push_back(history_element(PSEUDO_CLASS, current_offset, current_line));
						++current_offset;
						break;
					case '\n':
					case ' ':
						if(selector_.size()>0) {
							if(current_selector_==0) current_selector_=new CSS::selector;
							else current_selector_->relations_.push_back(CSS::selector::descendant);

							debug("adding element to selector; " << selector_ << ":" << pseudo_class_);
							CSS::selector::selector_element tmp(selector_);
							if(pseudo_class_!="") {
								if(pseudo_class_=="link")
									tmp.pseudoclasses.push_back(CSS::selector::selector_element::link);
								}
							current_selector_->elements_.push_back(tmp);
							selector_="";
						   pseudo_class_="";
							parse_history.push_back(history_element(EAT_WHITESPACE, current_offset, current_line));
							}
						++current_offset;
						break;
					case '{':
					case ',':
						if(current_selector_==0) {
							current_selector_=new CSS::selector;
							CSS::selector::selector_element tmp(selector_);
							if(pseudo_class_!="") {
								if(pseudo_class_=="link")
									tmp.pseudoclasses.push_back(CSS::selector::selector_element::link);
								}
							current_selector_->elements_.push_back(tmp);
							}
						selector_elements_.push_back(current_selector_);
						current_selector_=0;
						selector_="";
						pseudo_class_="";
						++current_offset;
						if(curchar=='{') {
							current_declaration_=new CSS::declaration;
							parse_history.push_back(history_element(BRACE_OPEN, current_offset, current_line));
							}
						break;
					default:
						selector_+=curchar;
						++current_offset;
						break;
					}
				break;
			case PSEUDO_CLASS:
#ifdef VERBOSE
				debug("PSEUDO_CLASS mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
					case '{':
					case ',':
						parse_history.pop_back();
						break;
					default:
						pseudo_class_+=curchar;
						++current_offset;
						break;
					}
				break;
			case BRACE_OPEN:
#ifdef VERBOSE
				debug("BRACE_OPEN mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						++current_offset;
						break;
					case '}': {
						++current_offset;
//						debug("complete declaration finished");
						for(unsigned int j=0; j<selector_elements_.size(); ++j) {
//							debug("selector " << j);
//							for(unsigned int i=0; i<selector_elements_[j]->elements_.size(); ++i) {
//								debug("|" << selector_elements_[j]->elements_[i].name << "|");
//								if(i+1<selector_elements_[j]->elements_.size()) 
//									debug((int)(selector_elements_[j]->relations_[i]));
//								}
							// FIXME: this will crash
							CSS::rule *rule1=new CSS::rule(selector_elements_[j],current_declaration_,j==0);
							stylesheet_->add_rule(rule1);
							}

						current_selector_=0;
						current_declaration_=0;
						selector_elements_.clear();

						parse_history.pop_back();
						break;
						}
					case ':':
					case ';':
//						debug("scanned property: |" << accumulator_ << "|");
						property_name_=accumulator_;
						parse_history.push_back(history_element(PROPVAL, current_offset, current_line));
						++current_offset;
						accumulator_="";
						break;
					default:
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case PROPVAL:
#ifdef VERBOSE
				debug("PROPVAL mode |" << (int)(curchar) << "|" << curchar);
#endif			
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						if(accumulator_.size()==0) {
							++current_offset;
							break;
							}
					case ';':
					case '}':
//						debug("property      : |" << property_name_ << "|");
//						debug("property value: |" << accumulator_ << "|");
						if(!add_declaration_element_(property_name_, accumulator_, current_declaration_)) {
							debug("Uh oh, that is not a valid declaration, ignoring");
							debug(property_name_ << ":" << accumulator_);
							}
						accumulator_="";
						if(curchar!='}')
							++current_offset;
						parse_history.pop_back();
						break;
					default:
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			default:
				++current_offset;
				throw logic_error("unimplemented feature of css parser triggered");
			}
		}
	
	return return_code;
	}
	
bool cssparser::add_declaration_element_(const string& prop, const string& propval, CSS::declaration *dec)
	{
	map<string, int>::const_iterator it=string2propnum_.find(prop);
	if(it==string2propnum_.end()) return false;

	switch((*it).second) {
		case CSS::declaration::display: {
			int val=0;
			if(propval=="block")                 val=CSS::declaration::block;
			else if(propval=="empty")            val=CSS::declaration::display_empty;
			else if(propval=="none")             val=CSS::declaration::display_none;
			else if(propval=="inline")           val=CSS::declaration::display_inline;
		   else if(propval=="table")            val=CSS::declaration::table;
			else if(propval=="table-row")        val=CSS::declaration::table_row;
			else if(propval=="table-cell")       val=CSS::declaration::table_cell;
			else if(propval=="table-caption")    val=CSS::declaration::table_caption;

			if(val==0) throw logic_error("stylesheet contains unsupported display properties");
			dec->set_display((CSS::declaration::display_value)val);
			break;
			}
		case CSS::declaration::font_size: {
			dec->set_font_size(new CSS::declaration::numberproperty(propval));
			break;
			}
		case CSS::declaration::font_weight:
			if(propval=="bold") dec->set_font_weight(CSS::declaration::weight_bold);
			break;
		case CSS::declaration::text_align:
			if(propval=="left")    dec->set_text_align(CSS::declaration::align_left);
			if(propval=="right")   dec->set_text_align(CSS::declaration::align_right);
			if(propval=="center")  dec->set_text_align(CSS::declaration::align_center);
			if(propval=="justify") dec->set_text_align(CSS::declaration::align_justify);
			break;
		case CSS::declaration::font_style:
			if(propval=="italic")       dec->set_font_style(CSS::declaration::font_italic);
			else if(propval=="oblique") dec->set_font_style(CSS::declaration::font_oblique);
			break;
		case CSS::declaration::font_family: {
			dec->set_font_family(new CSS::declaration::stringproperty(propval));
			break;
			}
		case CSS::declaration::margin_left:
		case CSS::declaration::margin_right:
		case CSS::declaration::margin_top:
		case CSS::declaration::margin_bottom: 
		case CSS::declaration::border_width: {
			dec->set_number_property((CSS::declaration::number_properties)((*it).second),
											 new CSS::declaration::numberproperty(propval));
			break;
			}
		case CSS::declaration::border_style: { 
			int val=0;
			if(propval=="none")          val=CSS::declaration::border_style_none;
			else if(propval=="hidden")   val=CSS::declaration::hidden;
			else if(propval=="dotted")   val=CSS::declaration::dotted;
			else if(propval=="dashed")   val=CSS::declaration::dashed;
			else if(propval=="solid")    val=CSS::declaration::solid;
			else if(propval=="double")   val=CSS::declaration::border_double;
			else if(propval=="groove")   val=CSS::declaration::groove;
			else if(propval=="ridge")    val=CSS::declaration::ridge;
			else if(propval=="inset")    val=CSS::declaration::inset;
			else if(propval=="outset")   val=CSS::declaration::outset;
			if(val==0) throw logic_error("unknown CSS border style");
			
			dec->set_border_style((CSS::declaration::border_style_value)val);
			break;
			}
		case CSS::declaration::color:
		case CSS::declaration::background_color: {
			dec->set_color_property(static_cast<CSS::declaration::color_properties>((*it).second),
											new CSS::declaration::colorproperty(propval));
			break;
			}
		default:
			return false;
		}

	return true;
	}

const cssparser::history_element& cssparser::peek_(unsigned int level=0) const
	{
	if(level+1>parse_history.size())
		throw css_parse_error("while examining stack",css_parse_error::INTERNAL_STACK_UNDERFLOW);
	return parse_history[parse_history.size()-1-level];
	}

unsigned char cssparser::tokenise_(unsigned char c) const
	{
	if(c=='\n' || c=='\r')    return '\n';
	if(isspace(c))            return ' ';
	if(c==255 || c==254)
		return 255; // here we can filter out illegal characters
	return c;
	}

void cssparser::fill_string2enum_tables_(void)
	{
	string2propnum_["margin-top"]          =CSS::declaration::margin_top;
	string2propnum_["margin-bottom"]       =CSS::declaration::margin_bottom;
	string2propnum_["margin-left"]         =CSS::declaration::margin_left;
	string2propnum_["margin-right"]        =CSS::declaration::margin_right;
	string2propnum_["max-height"]          =CSS::declaration::max_height;
	string2propnum_["max-width"]           =CSS::declaration::max_width;
	string2propnum_["min-height"]          =CSS::declaration::min_height;
	string2propnum_["min-width"]           =CSS::declaration::min_width;
	string2propnum_["border-spacing"]      =CSS::declaration::border_spacing;
	string2propnum_["font-size"]           =CSS::declaration::font_size;

	string2propnum_["background-color"]    =CSS::declaration::background_color;
	string2propnum_["border-color"]        =CSS::declaration::border_color;
	string2propnum_["border-top-color"]    =CSS::declaration::border_top_color;
	string2propnum_["border-bottom-color"] =CSS::declaration::border_bottom_color;
	string2propnum_["border-left-color"]   =CSS::declaration::border_left_color;
	string2propnum_["color"]               =CSS::declaration::color;
	string2propnum_["outline-color"]       =CSS::declaration::outline_color;
	string2propnum_["text-shadow-color"]   =CSS::declaration::text_shadow_color;

	string2propnum_["font-family"]         =CSS::declaration::font_family;

	string2propnum_["background-image"]    =CSS::declaration::background_image;
	string2propnum_["content-uri"]         =CSS::declaration::content_uri;
	string2propnum_["cursor-uri"]          =CSS::declaration::cursor_uri;
	string2propnum_["list-style-image"]    =CSS::declaration::list_style_image;

	string2propnum_["background-attachment"] =CSS::declaration::background_attachment;
	string2propnum_["display"]             =CSS::declaration::display;
	string2propnum_["font-style"]          =CSS::declaration::font_style;
	string2propnum_["font-weight"]         =CSS::declaration::font_weight;

	string2propnum_["text-align"]          =CSS::declaration::text_align;
	string2propnum_["border-style"]        =CSS::declaration::border_style;
	string2propnum_["border-width"]        =CSS::declaration::border_width;
	}
