// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Classes for browsing logic (history handling, connections to the
  network layer etc). GUI independent (will move to a separate library
  soon).

  Copyright 1998-1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef browser_hh_
#define browser_hh_

#include <string>
#include <map.h>
#include <vector.h>

#include <mnemonic/messagebase.hh>
#include <msg/client/run_client.hh>
#include <msg/client/get_user_authentication.hh>
#include <msg/display/link_clicked.hh>
#include <msg/network/network_status.hh>
#include <msg/network/network_data.hh>
#include <msg/network/stop.hh>
#include <msg/network/get_url.hh>
#include <msg/dom/parse_error.hh>

namespace browser {

 class pageview_manager {
	 public:
		 pageview_manager(void);
		 virtual ~pageview_manager();

		 untyped_interest::callresult accept_run_client(msg_run_client &);
		 virtual void first_window_(void)=0;
		 virtual void open_window_(void)=0;

	 protected:
		 int    *argc_;
		 char ***argv_;

	 private:
		 void setup_interests_(void);
		 void destroy_interests_(void);

		 ptr<interest<pageview_manager, msg_run_client> >              interest_run_client_;
 }; 

 class pageview {
	 public:
		 pageview(void);
		 virtual ~pageview();

		 untyped_interest::callresult accept_network_status(msg_network_status &);
		 untyped_interest::callresult accept_network_data(msg_network_data &);
		 untyped_interest::callresult accept_link_clicked(msg_link_clicked &);
		 untyped_interest::callresult accept_parse_error(msg_parse_error &);
		 untyped_interest::callresult accept_get_url(msg_get_url &);
		 untyped_interest::callresult accept_get_user_authentication(msg_get_user_authentication &);

		 void goto_url(string url, bool nocache);
		 void back(void);
		 void forward(void);
		 void stop(void);
		 void close(void);

		 virtual pageview_manager *owner(void) const=0;
	 protected:
		 enum connection_status { attempted_connections, active_connections, no_connections };

		 // implement these in your derived class to do something fancy
		 virtual void set_status_display(const string&);
		 virtual void set_pageurl_display(const string&);
		 virtual void set_url_display(const string&);
		 virtual void set_broken_html(bool);
		 virtual void set_forward_sensitive(bool);
		 virtual void set_backward_sensitive(bool);
		 virtual void set_connection_status(connection_status);
		 virtual pair<string,string> handle_user_authentication(void);
		 virtual void register_connection_interests(const group_identifiers&);

		 thread_mutex progress_mutex_;

		 class history_buffer {
			 public:
				 history_buffer(void);

				 vector<string>            history_;
				 vector<string>::iterator  history_pos_; // always points to current page
		 };
		 history_buffer     hbuf_;
		 group_identifiers  geturl_ids_;

		 map<group_identifiers, url_string>                attempted_connections_;
		 map<group_identifiers, url_string>                open_connections_;

		 url_string find_connection_(const group_identifiers&);
		 void       add_connection_(const group_identifiers&, const url_string& ur);
		 void       remove_connection_(const group_identifiers&, const string& alldone="Document done.");
		 void       clear_connections_(void);

		 ptr<interest<pageview, msg_link_clicked> >            interest_link_clicked_;
	 private:
		 void setup_interests_(void);
		 void destroy_interests_(void);

		 ptr<interest<pageview, msg_network_status> >          interest_network_status_;
		 ptr<interest<pageview, msg_network_data> >            interest_network_data_;
		 ptr<interest<pageview, msg_parse_error> >             interest_parse_error_;
		 ptr<interest<pageview, msg_get_url> >                 interest_get_url_;
		 ptr<interest<pageview, msg_get_user_authentication> > interest_get_user_authentication_;
 };


};


#endif
