// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Classes for browsing logic (history handling, connections to the
  network layer etc). GUI independent (will move to a separate library
  soon).

  Copyright 1998-1999 Kasper Peeters.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "browser.hh"

#define DEBUG_CLASS "BROWSER_MIO"
#include <mnemonic/debug.hh>

using namespace browser;

pageview_manager::pageview_manager(void)
	{
	setup_interests_();
	}

pageview_manager::~pageview_manager()
	{
	destroy_interests_();
	}

untyped_interest::callresult pageview_manager::accept_run_client(msg_run_client& msg)
	{
	argc_=msg.argc;
	argv_=msg.argv;
	first_window_();
   return untyped_interest::exit_loop;
	}

untyped_interest::callresult pageview::accept_get_user_authentication(msg_get_user_authentication& msg)
	{
	pair<string,string> tmp=handle_user_authentication();

	if(tmp.first!="") {
		msg.user=tmp.first;
		msg.password=tmp.second;
		debug("got a user plus password");
		return untyped_interest::exit_loop;
		}
	else {
		debug("dialog cancelled");
		msg.user="";
		msg.password="";
		return untyped_interest::pass_on;
		}
	}

void pageview_manager::setup_interests_(void)
	{
	interest_run_client_=interest<pageview_manager, msg_run_client>::
		create(this, &pageview_manager::accept_run_client,untyped_interest::normal);
	}

void pageview_manager::destroy_interests_(void)
	{
	interest_run_client_->unregister();
	interest_run_client_.wait();
	}






void pageview::setup_interests_(void)
	{
	interest_network_status_=interest<pageview, msg_network_status>::
		create(this, &pageview::accept_network_status,untyped_interest::normal);
	interest_network_data_=interest<pageview, msg_network_data>::
		create(this, &pageview::accept_network_data,untyped_interest::normal);
	interest_link_clicked_=interest<pageview, msg_link_clicked>::
		create(this, &pageview::accept_link_clicked,untyped_interest::normal);
	interest_parse_error_=interest<pageview, msg_parse_error>::
		create(this, &pageview::accept_parse_error,untyped_interest::normal);
	interest_get_url_=interest<pageview, msg_get_url>::
		create(this, &pageview::accept_get_url,untyped_interest::normal);
	interest_get_user_authentication_=interest<pageview, msg_get_user_authentication>::
		create(this, &pageview::accept_get_user_authentication,untyped_interest::normal);

	interest_network_status_->acceptance=untyped_interest::mid_contains_iid;
	interest_network_data_->acceptance=untyped_interest::mid_contains_iid;
	interest_link_clicked_->acceptance=untyped_interest::mid_contains_iid;
	interest_parse_error_->acceptance=untyped_interest::mid_contains_iid;
	interest_get_url_->acceptance=untyped_interest::mid_contains_iid;
	interest_network_status_->acceptance=untyped_interest::mid_contains_iid;
	}

void pageview::destroy_interests_(void)
	{
	interest_network_status_->unregister();
	interest_network_data_->unregister();
	interest_link_clicked_->unregister();
	interest_parse_error_->unregister();
	interest_get_url_->unregister();
	interest_get_user_authentication_->unregister();

	interest_network_status_.wait();
	interest_network_data_.wait();
	interest_link_clicked_.wait();
	interest_parse_error_.wait();
	interest_get_url_.wait();
	interest_get_user_authentication_.wait();
	}

void pageview::register_connection_interests(const group_identifiers& gid)
	{
	geturl_ids_=gid;
	interest_network_data_->set_group_ids(gid);
	interest_network_status_->set_group_ids(gid);
	interest_parse_error_->set_group_ids(gid);
	interest_get_url_->set_group_ids(gid);
	interest_get_user_authentication_->set_group_ids(gid);
	}

untyped_interest::callresult pageview::accept_network_status(msg_network_status& msg)
	{
	switch(msg.identifier) {
		case msg_network_status::looking_up:
			set_status_display("Looking up host name "+msg.url.get_net_path()+" ...");
			break;
		case msg_network_status::bad_host:
			remove_connection_(msg,"Host "+msg.url.get_net_path()+" unknown, cannot access "
										  +msg.url.get_absolute_path()+".");
			break;
		case msg_network_status::connecting:
			if(msg.text!="") {
				set_status_display("Making connection to "+msg.url.get_net_path()+
											  " ("+string(msg.text)+") ...");
				}
			else {
				set_status_display("Making connection to "+msg.url.get_net_path()+" ...");
				}
			break;
		case msg_network_status::connection_refused:
			remove_connection_(msg,"SigC::Connection to "+msg.url.get_net_path()+" refused.");
			break;
		case msg_network_status::no_route:
			remove_connection_(msg,"No route to host.");
			break;
		case msg_network_status::open_error:
			remove_connection_(msg,"Open error: "+msg.text);
			break;
		case msg_network_status::sending_request:
			add_connection_(msg, msg.url);	
			set_status_display("Connected. Requesting object "+msg.url.get_object()+" ...");
			break;
		case msg_network_status::request_sent:
			set_status_display("Object requested, waiting for reply from "+msg.url.get_net_path()+" ...");
			break;
		case msg_network_status::content:
			add_connection_(msg, msg.url);
			set_status_display("Receiving "+string(msg.text)+" object "+msg.url.get_object()+" ...");
			break;
		case msg_network_status::closed:
		case msg_network_status::finished:
		case msg_network_status::object_truncated:
			set_status_display("Object "+msg.url.get_object()+" received.");
			remove_connection_(msg);
			break;
		default:
			// FIXME: accept all error codes and do this properly.
			set_status_display(msg.status_as_text()+".");
			remove_connection_(msg);
			break;
		}

	return untyped_interest::pass_on;
	}

untyped_interest::callresult pageview::accept_network_data(msg_network_data& msg)
	{
	set_status_display("Receiving data ...");
	return untyped_interest::pass_on;
	}

untyped_interest::callresult pageview::accept_link_clicked(msg_link_clicked& msg)
	{
	if(msg.button==1)
		goto_url(msg.href,false);
	else if(msg.button==0) {
		set_url_display(msg.href);
		}

	return untyped_interest::call_next;
	}

untyped_interest::callresult pageview::accept_parse_error(msg_parse_error& msg)
	{
	set_broken_html(true);	
//	<< "error on line " << msg.line_number << ": " << msg.error_text << endl;
	return untyped_interest::call_next;
	}

untyped_interest::callresult pageview::accept_get_url(msg_get_url& msg)
	{
	add_connection_(msg, msg.url);	
	return untyped_interest::pass_on;
	}

void pageview::stop(void)
	{
	debug("sending stop message " << geturl_ids_);
	msg_stop msg;
	msg.set_group_ids(geturl_ids_);
	try {
		msg.send();
		}
	catch(messagebase_error& ex) {
		debug("ehm, what the bloody hell is happening here?");
		}
	}

template interest<pageview, msg_network_status>;
template interest<pageview, msg_network_data>;
template interest<pageview, msg_link_clicked>;
template interest<pageview, msg_get_url>;
template interest<pageview, msg_parse_error>;
template interest<pageview, msg_get_user_authentication>;

template interest<pageview_manager, msg_run_client>;
