// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  GTK based image rendering unit (one for each image).

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "gtkimagerenderer.hh"
#include <msg/gtkdisplay/object_ready.hh>

#include <unistd.h>
#include <dom/dom.hh>
#include <dom/css.hh>
#include <msg/dom/css_event.hh>
#include <msg/network/get_url.hh>

#define DEBUG_CLASS "GTKIMAGERENDERER_MIO"
#include <mnemonic/debug.hh>

gtkimagerenderer::gtkimagerenderer(const group_identifiers& ids, ptr<CSS::stylesheet> sheet, DOM::Node *sh)
	: 	rio_(new raster_io), sheet_(sheet), start_here_(sh)
	{
	interest_network_data_=interest<gtkimagerenderer, msg_network_data>::create(this, &gtkimagerenderer::accept_network_data,untyped_interest::normal);
	interest_network_data_->acceptance=untyped_interest::mid_equals_iid;

	interest_network_status_=interest<gtkimagerenderer, msg_network_status>::create(this, &gtkimagerenderer::accept_network_status,untyped_interest::normal);
	interest_network_status_->acceptance=untyped_interest::mid_equals_iid;

	interest_stop_=interest<gtkimagerenderer, msg_stop>::create(this, &gtkimagerenderer::accept_stop,untyped_interest::normal);
	interest_stop_->acceptance=untyped_interest::iid_contains_mid;
	}

gtkimagerenderer::~gtkimagerenderer()
	{
	if(toplevel_widget_!=0) {
		debug("someone signalled me to go away");
		debug("locking client once more");
		Threadsafe_Main::instance()->interrupt_event_loop();

		// FIXME: eh, do we have to do something here?

		debug("releasing client forever");
		Threadsafe_Main::instance()->restart_event_loop();
		debug("bye bye");
		}
	else {
		debug("deleted, but we didn't even have an imagebox open yet");
		}
	if(rio_)
		delete rio_;
	}

untyped_interest::callresult gtkimagerenderer::accept_network_status(msg_network_status& msg)
	{
	debug("status message received: " << msg.status_as_text());
	if(msg.identifier==msg_network_status::closed ||
		msg.identifier==msg_network_status::object_truncated ||
		msg.identifier==msg_network_status::timed_out ||
		msg.identifier==msg_network_status::finished) {
		interest_network_data_->remove_group_ids(msg);
		interest_network_status_->remove_group_ids(msg);
		debug("connection closed for " << imgurl_.get_url());
		rio_->await_completion();
		update_box_();
		// We can safely delete the decoder as we have given a ptr<raster_image> to the
		// gtk_imagebox.
      delete rio_;
		rio_=0;
		debug("removed interest");
		}
	else if(msg.identifier==msg_network_status::bad_host) {
		interest_network_data_->remove_group_ids(msg);
		interest_network_status_->remove_group_ids(msg);
		debug("image " << imgurl_.get_url() << " cannot be fetched: bad hostname");
		rio_->await_completion();
		update_box_();
		// We can safely delete the decoder as we have given a ptr<raster_image> to the
		// gtk_imagebox.
      delete rio_;
		rio_=0;
		debug("removed interest");
		}
	else if(msg.identifier==msg_network_status::content) {
		debug("helping the decoders a bit");
		rio_->use_decoder(msg.text);
		}
	else {
		debug("status " << msg.status_as_text() << " for " << imgurl_.get_url());
		}
	return untyped_interest::call_next;
	}


untyped_interest::callresult gtkimagerenderer::accept_network_data(msg_network_data& msg)
	{
	debug("image data (" << msg.length << ") for " << imgurl_.get_url());
	rio_->write((const unsigned char *)msg.data,(long)msg.length);
	update_box_();
	return untyped_interest::exit_loop;
	}

untyped_interest::callresult gtkimagerenderer::accept_stop(msg_stop& msg)
	{
	// FIXME: IMPORTANT, remove from memory and so on.
	interest_network_data_->remove_group_ids(msg);
	interest_network_status_->remove_group_ids(msg);
	interest_stop_->remove_group_ids(msg);
	return untyped_interest::call_next;
	}

void gtkimagerenderer::update_box_(void)
	{
	ptr<scaled_raster_image> im=rio_->image();
	if(im!=0) {
		toplevel_area_->lock();
		debug("image is there");
		if(forced_width_==0) {
			forced_width_=im->scaled_width();
			}
		if(forced_height_==0) {
			forced_height_=im->scaled_height();
			}
		if ((forced_height_!=im->scaled_height())||(forced_width_!=im->scaled_width())) {
		    im->scale(forced_width_,forced_height_);
		} 	
		toplevel_widget_->modify(im);
		// Because we now have access to the tree anyway, we may as well
		// reflow (if necessary) and redraw. If our image widget (toplevel_widget_)
		// changed, force a reflow.
		if(toplevel_widget_->check_recompute_children()) {
			toplevel_area_->unlock();
			gdk_threads_enter();
			toplevel_widget_->reflow();
			gdk_threads_leave();
			}
		else toplevel_area_->unlock();

		toplevel_widget_->redraw(); 
		}
	}

bool gtkimagerenderer::announce(const group_identifiers& gid)
	{
	msg2.set_group_ids(gid);
	debug("sending message");
	try {
		msg2.send();
		
		int w=0, h=0;
		forced_width_=0;
		forced_height_=0;
		string srcname;
		DOM::Element *el=dynamic_cast<DOM::Element *>(start_here_);
		if(el) {
			debug(el->nodeName());
			if(el->firstAttribute()==0) {
				debug("no attributes, so how the hell do we find out what is the source?");
				}
			else {
				debug(el->firstAttribute()->nodeName() << "=" << el->firstAttribute()->nodeValue());
				debug("height=" << el->getAttribute("HEIGHT"));
				debug("width=" << el->getAttribute("WIDTH"));
				debug("src=" << el->getAttribute("SRC"));
				srcname=el->getAttribute("SRC");
				}
			string wid=el->getAttribute("WIDTH");
			string hei=el->getAttribute("HEIGHT");
			if(wid!="") {
				if(wid.find_first_not_of("0123456789")==wid.npos)
					if((forced_width_=atoi(wid.c_str()))!=0)
						w=forced_width_;
				}
			if(hei!="") {
				if(hei.find_first_not_of("0123456789")==hei.npos)
					if((forced_height_=atoi(hei.c_str()))!=0)
						h=forced_height_;
				}
			}

		debug("image size is " << w << "," << h);

		toplevel_widget_=new gtkimagebox(w,h);
		debug("new imagebox made");
		assert(msg2.layoutbox!=0);
		msg2.layoutbox->add_child(toplevel_widget_);
		toplevel_area_=msg2.layoutbox->current_area();
		msg2.new_layoutbox=toplevel_widget_;
		debug("imagebox added");
// This is just a quick hack to see whether buttons still work.
/*		gtkviewer    *tmp=new gtkviewer("white");
		gtkbuttonbox *testbox=new gtkbuttonbox("press me");
		msg2.layoutbox->add_child(tmp);
		debug("new gtkviewer made " << tmp);
		debug("new buttonbox made " << testbox);
tmp->add_child(testbox); */

		
		debug("referrer is " << sheet_->document()->url());
		imgurl_=url_string(srcname.c_str(), sheet_->document()->url().c_str());
		debug("fetching " << imgurl_.get_url());
		msg_get_url imgmsg(imgurl_, false);

		// copy the old ids and add an additional one, too.
		midentifier imgid;
		imgmsg.add_group_id(imgid);
		imgmsg.add_group_ids(gid);

		interest_network_data_->add_group_ids(imgmsg);
		interest_network_status_->add_group_ids(imgmsg);

		debug("ids for new image thread: " << imgmsg);
		try {
			imgmsg.send();
			}
		catch(messagebase_error& er) {
			debug("nobody wants to fetch the image for us...");
			}

		debug("message was accepted");
		}
	catch(messagebase_error &ex) {
		debug(ex);
		return false;
		}
	return true;
	}

template interest<gtkimagerenderer, msg_network_data>;
template interest<gtkimagerenderer, msg_network_status>;
template interest<gtkimagerenderer, msg_stop>;
