// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Mnemonic http thread (does all the work)

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define DEBUG_CLASS "http_thread"
#include <mnemonic/debug.hh>
#include "http_thread.hh"
#include <string>
#include <strstream.h>
#include <stdio.h>
#include <msg/client/get_user_authentication.hh>


// copy all the information we need...
http_thread::http_thread(const http_url &url, bool noc, http_handler *hh,
                         const group_identifiers& ids, tcp_connection *t) : thread() 
	{
	ids_ = ids;
	url_ = url;
	no_cache_ = noc;
	tc_ = t;
	hh_ = hh;
	should_exit_ = 0;
	have_length_ = 1;
	setup_interests_();
	// FIXME: auto-starting is something we usually do not do.
	start();
	}

http_thread::~http_thread() 
	{
	remove_interests_();
	}

void http_thread::cancel(void) 
	{
	should_exit_ = 1;
	}

// on return from this, our thread of execution is finished.
void http_thread::begin() 
	{
	debug("In begin()");
	try {
		if((tc_->status() == tcp_connection::ok || connect_()) && !should_exit_)
			if(send_request_() && !should_exit_)
				if(read_response_() && !should_exit_)
					if(data_flow_() && !should_exit_) {
						debug("Donating this connection to the pool!");
						hh_->donate_to_pool(url_.get_net_path(), tc_);
						debug("Calling thread_exit");
						hh_->thread_exit(this);
						return;
						}
		if(should_exit_) {
			debug("We got told to exit!");
			}
		delete tc_;
		debug("done");
		}
	catch(exception& er) {
		debug("[exception]: " << er.what());
		throw;
		}
	return;
	}

void http_thread::send_status_(msg_network_status::status a, const char *b) 
	{
	msg_network_status mns(a,b,url_);
	mns.add_group_ids(ids_);
	try{mns.send();} 
	catch(messagebase_error&) {
		debug("msg_network_status had no recipients, ignoring");
		};
	}

void http_thread::send_status_(msg_network_status::status a) 
	{
	send_status_(a,"");
	}

int http_thread::read_(char *buf, int length) 
	{
	int res = tc_->read(buf,length);
	if(res == -1) {
		switch(tc_->status()) {
			case tcp_connection::max_transferred:
				send_status_(msg_network_status::finished);
				tc_->set_max_transfer(0);
				return 0; // It's ok, this is what we want.
			default:
			case tcp_connection::closed:
				if(have_length_) {
					send_status_(msg_network_status::object_truncated);
					return -1; // is ok
					} 
				else {
					send_status_(msg_network_status::finished);
					return 0;
					}
			case tcp_connection::other_error:
				send_status_(msg_network_status::closed,tc_->error_text());
				return -1;
			}
		}
	return res;
	}

int http_thread::connect_() 
	{
	debug("port: "<<url_.get_port());
	send_status_(msg_network_status::looking_up);
	if(tc_->lookup()==-1) {
		switch(tc_->status()) {
			case tcp_connection::host_not_found:
				send_status_(msg_network_status::bad_host);
				break;
			case tcp_connection::ns_bad:
				send_status_(msg_network_status::bad_host,
								 "nameserver not responding");
				break;
			default:
				// FIXME: send other error messages.
				break;
			}
		return 0;
		}
	if(should_exit_) return 0; // don't bother connecting
		send_status_(msg_network_status::connecting);
		
		if(tc_->connect()==-1) {
			switch(tc_->status()) {
				case tcp_connection::timed_out:
					send_status_(msg_network_status::timed_out); 
					break;
				case tcp_connection::connection_refused:
					send_status_(msg_network_status::connection_refused); 
					break;
				case tcp_connection::network_unreachable:
					send_status_(msg_network_status::network_unreachable); 
					break;
				case tcp_connection::no_route:
					send_status_(msg_network_status::no_route);
					break;
				default:
					send_status_(msg_network_status::closed, "Unknown error"); 
					break;
					break;
				}
			return 0;
			}
		return 1;
	}

int http_thread::send_request_() 
	{
	string esc;
	strstream buf;

	if(tc_->uses_proxy()) {
		escape_(url_.get_url(), esc);
		buf << "GET " << esc << " HTTP/1.0\r\n";
		send_status_(msg_network_status::sending_request, "using proxy");
		}
	else {
		send_status_(msg_network_status::sending_request);
		escape_(url_.get_absolute_path(), esc);
		if(url_.get_query().size()>0) {
			esc+="?";
			escape_(url_.get_query(),esc);
			}
		if(url_.get_parameters().size()>0) {
			esc+="#";
			escape_(url_.get_parameters(),esc);
			}
		buf << "GET " << esc << " HTTP/1.0\r\n";
		}
	buf << "Connection: Keep-Alive\r\n";
	buf << "User-Agent: Mnemonic/0.1 (X11; Linux)\r\n";
	if(no_cache_)
		buf << "Pragma: no-cache\r\n";
	buf << "Host: " << url_.get_net_path() << "\r\n\r\n" << ends;
	
	string str=buf.str();
	delete [] buf.str();
	
	debug("Writing " << str);
	if (tc_->write(str.c_str(),str.length()) == -1) {
		switch(tc_->status()) {
			default:
			case tcp_connection::closed:
				send_status_(msg_network_status::closed); break;
			case tcp_connection::other_error:
				send_status_(msg_network_status::closed,tc_->error_text());
				break;
			}
		return 0;
		}
	send_status_(msg_network_status::request_sent);
	return 1;
	}

// we scan the response from the 
int http_thread::read_response_() 
	{
	debug("reading response from server");
	enum { START, EOL, CODE, MSG, KEY, COLON, VAL, DONE } state;
	int bytes=-1;
	string header_text,value_text,message_text;
	state=START;
	int i=0;
	int length=0;
	
	while(state != DONE && !should_exit_) {
		length = read_(buf,BUFFER_LENGTH);
		if(length == -1 || length == 0) // length==0 should never happen 
			return 0;
		for (i=0; i<length && state!=DONE; i++) {
			debug(buf[i]<<"\t"<<(int)state);
			switch(state) {
				case START: // FIXME: Should check for a 0.9 response
					if (buf[i] ==  ' ') state=CODE;
					break;
				case CODE:
					if (buf[i] ==  ' ') {
						response_code_ = atoi(header_text.c_str());
						debug("ResponseCODE: " << response_code_);
						header_text="";
						state=MSG;
						}
					else
						header_text += buf[i];
					break;
				case MSG:
					if(buf[i] ==  '\n') {
						state=EOL;
						break;
						}
					if(buf[i] != '\r')
						message_text += buf[i];
					break;
				case EOL:
					if(buf[i]=='\n') {
						state=DONE;
						debug("Got content-length header with value " << headers["content-length"] );
						if(headers["content-length"].size() > 0) {
							debug("Setting bytes from the content-length header\n");
							bytes=atoi(headers["content-length"].c_str());
							have_length_ = bytes;
							} 
						else {
							bytes=-1; /* Signal that we are not an HTTP 1.1 connection */
							have_length_ = 0;
							}
						debug("Bytes available are "<<bytes<<"\n");
						if (!check_header_()) return 0;
						debug("Sending content message " << headers["content-type"] << " to " << ids_);
						msg_network_status ms(msg_network_status::content, headers["content-type"].c_str(),url_);
						ms.add_group_ids(ids_);
						try { ms.send(); }                                 
						catch(messagebase_error& ex) {                               
							debug(ex);
							debug("Since nobody is listening to this content type, we might as well stop immediately");
							// FIXME: we need a message type for internal failures like this. Could be
							// used to regularly send messages to a debug window.
							should_exit_=1;
							break;
							}
						debug("Done sending content message");
						break;
						}		
					if(buf[i]!='\r') {
						i--;
						state=KEY;
						}
					break;
				case KEY:
					if(buf[i]==':') {
						state=COLON;
						break;
						} 
					else
						header_text += buf[i];
					break;
				case COLON:
					if(buf[i]!=' ') {
						header_text += ':';
						header_text += buf[i];
						state=KEY;
						} 
					else {
						state=VAL;
						}
					break;
				case VAL:
					if (buf[i] == '\n') {
						for(string::iterator it=header_text.begin(); it!=header_text.end(); ++it)
							(*it)=tolower((*it));
						debug("Setting header " << header_text << " to " << value_text);
						// FIXME: add cookie handling (ie. header `set-cookie'). Konqueror has the options
                  //  accept
						//  reject
						//  always accept
                  //  always reject
                  //  always ask
                  // on a domain name basis. Of course, this should not go into the
						// http networking module but rather should be part of the client.
						// It is probably also useful to automatically accept headers that match
						// a regexp, like all that ASPSESSIONID stuff, for instance.
						// FIXME: gather information to be dumped somewhere in a status window.
						// I guess all header information would be useful.
						headers[header_text]=value_text;
						state=EOL;
						header_text="";
						value_text="";
						break;
						}
					if(buf[i]!='\r')
						value_text += buf[i];
					break;
				case DONE:
					break;
				}
			}
		}
	if(should_exit_) return 1;
	debug("Header negotiation is done!");
	// we have left-over data.
	// I'm going to cheat here and send the data
	// directly...
	size_t length_left_over = length - i;
	debug("Length left over is "<<length_left_over);
	if(length_left_over != 0) {
		msg_network_data mnd(buf+i,length_left_over);
		mnd.add_group_ids(ids_);
		try { mnd.send(); }
		catch(messagebase_error& ex) { 
			debug("for id " << ids_ << ":" << endl << ex); 
			} // ignore error
		debug("Sent the message");
		}
	// at this point, "bytes" says how long the
	// object is...  (make sure bytes is size_t<--FIXME)
	if(bytes-length_left_over>0) tc_->set_max_transfer(bytes-length_left_over);
	debug("all good");
	return 1; //all good
	}

int http_thread::data_flow_() 
	{
	debug("data_flow_()");
	char buf[BUFFER_LENGTH];
	ssize_t length=0;
	while((length = read_(buf,BUFFER_LENGTH)) != -1 && length != 0  && !should_exit_) {
		debug("Got a chunk of data to send as a data message");
		msg_network_data mnd(buf,length);
		mnd.add_group_ids(ids_);
		debug("msg constructed");
		try { mnd.send(); } 
		catch(messagebase_error&ex) {
			debug("for id " << ids_ << ":" << endl << ex); 
			debug("nobody wants the data, shut down connection");
			send_status_(msg_network_status::object_truncated);
			should_exit_=true;
			}
		debug("msg sent");
		}
	debug("Done with data_flow, length is "<<length);
	if(should_exit_ || length != 0) return 0; // not ok
	else return 1; // ok
	}

void http_thread::escape_(const string &url, string &res) 
	{
	// FIXME: this should not be part of the http module but rather sit in the url_string lib
	int i=0;
	char hex[4];
	while (url[i]) {
		if ((url[i] > 0 && url[i] <=31) || url[i] == ' ' || url[i] == '"' || url[i] == '#' ||
			 url[i] == '%' || url[i] == '<' || url[i] == '>' || url[i] >= 127) {
			sprintf(hex, "%%%02X", url[i]);
			res+=hex;
			debug ("Escaping "<<url[i]<<" to %%"<< hex);
			} 
		else {
			res+=url[i];
			}
		i++;
		}
	debug("Escaped path: "<<res);
	}

int http_thread::check_header_() 
	{
	switch(response_code_) {
		case 200:                                /* OK */
		case 201:                                /* Created */
		case 202:                                /* Accepted */
		case 203:                                /* Non-Authoritative header */
		case 206:                                /* Partial Content */
		case 300:                                /* Multiple Choices */
			debug ("Response ok!");
			break;
		case 301:                                /* Moved Permanently */
		case 302:                                /* Moved Temporarily */
		case 303: {                              /* See Other */
			string url = headers["location"];
			url_string urls(url.c_str());
			debug("We have been redirected to " << url);
			// use the same ids_
			msg_network_status mns(msg_network_status::redirected,url.c_str(),url_);
			mns.add_group_ids(ids_);
			try { mns.send(); } 
			catch(messagebase_error& ex) {
				debug(ex); 
				}
			msg_get_url mgu(urls, no_cache_);
			mgu.add_group_ids(ids_);
			try { mgu.send(); } catch (messagebase_error& ex) {
				debug(ex);
				msg_network_status ms(msg_network_status::generic_error,"unable to handle redirect", url_);
				ms.add_group_ids(ids_);
				try {ms.send();} catch(messagebase_error &ex2) { debug(ex2); }
				}
			// All we need to do is send a new msg_[get|put|post]_url with
			// our same midentifier for this new place.
			return 0; // cancel this attempt
			}
		case 401: {                             /* authenticate required */
			debug("got a 401 authentication error "<< headers["www-authenticate"].c_str());
//				msg_network_status ms(ids_, msg_network_status::authentication,"authentication required");
//				try { ms.send(); }
//				catch(messagebase_error& ex) {
//					debug("mkeyStatus::AUTHENTICATION failed");
//				}
			
			msg_get_user_authentication mgua;
			try { 
				mgua.send(); 
				}
			catch(messagebase_error& ex) {
				debug("msg_get_user_authentication failed");
				return -1;
				}
			debug("got " << mgua.user << ":" << mgua.password);

			url_string new_url(url_);
			new_url.set_username(mgua.user);
			new_url.set_passwd(mgua.password);
			msg_get_url mgu(new_url, no_cache_);

			mgu.add_group_ids(ids_);
			try { 
				mgu.send(); 
				} 
			catch (messagebase_error& ex) {
				debug(ex);
				msg_network_status ms(msg_network_status::generic_error,"unable to re-fetch", url_);
				ms.add_group_ids(ids_);
				try { 
					ms.send();
					} 
				catch(messagebase_error &ex2) { 
					debug(ex2); 
					}
				}
			// All we need to do is send a new msg_[get|put|post]_url with
			// our same midentifier for this new place.

			return 0; // cancel this attempt
			break;
			} 
		default: {
			debug("Error response received.");
			msg_network_status ms(msg_network_status::generic_error,
										 "Error response CODE received from server.", url_);
			ms.add_group_ids(ids_);
			try { ms.send(); }
			catch(messagebase_error& ex) { debug(ex); }
			return -1; // error
			}
		}
   return 1; //no error
	}
