// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Test program to fetch multiple urls simultaneously and dump them 
  to `file%d.out'.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#include <mnemonic/optionals.h>
#include <mnemonic/debugger.hh>
#include <mnemonic/objectmanager.hh>
#include <mnemonic/thread_cond.hh>
#include <mnemonic/midentifier.hh>
#include <msg/network/get_url.hh>
#include <msg/network/network_status.hh>
#include <msg/network/network_data.hh>
#include <msg/network/stop.hh>

class receiver {
	public:
		receiver(const char *url);
		~receiver();
		untyped_interest::callresult accept_msg_network_data(msg_network_data &m);
		untyped_interest::callresult accept_msg_network_status(msg_network_status &m);
		bool start(void);
		void stop(void);
		static thread_cond cond_mutex;
		
	private:
		url_string url_;
		bool connection_ok_;
		ptr<interest<receiver, msg_network_data> >   interest_data_;
		ptr<interest<receiver, msg_network_status> > interest_status_;
		midentifier id_;
		int fd;
		static int id;
};

int receiver::id=0;
thread_cond receiver::cond_mutex;

receiver::receiver(const char *url)
	{
	char name[20];
	url_=url;
	sprintf(name,"file%i.out",id++);
	fd=creat(name,S_IRUSR|S_IWUSR);
	cout<<"writing "<<url<<" to "<<name<<endl;
	interest_data_= interest<receiver, msg_network_data>::create
		(this,&receiver::accept_msg_network_data, untyped_interest::normal);
	
	interest_status_= interest<receiver, msg_network_status>::create
		(this,&receiver::accept_msg_network_status, untyped_interest::normal);
	}

receiver::~receiver()
	{
	cerr<<"destruct \n";
	close(fd);
	interest_data_->unregister();
	interest_status_->unregister();
	
	interest_data_.wait();
	interest_status_.wait();
	}

void receiver::stop() 
	{
	msg_stop ms;
	ms.add_group_id(id_);
	cerr << "Stopping"<<endl;
	try { ms.send(); }
	catch(messagebase_error &) { }; //they may have finished before we stop them
	cerr << "Stopped"<<endl;
	}

bool receiver::start(void)
	{
	interest_data_->add_group_id(id_);
	interest_status_->add_group_id(id_);
	cerr << "making a get_url message" <<endl;
	msg_get_url mgu(url_, false);
	mgu.add_group_id(id_);
	cerr << "fetching " << url_.get_url() << endl;
	connection_ok_=true;
	try { mgu.send(); }
	catch(messagebase_error &ex) {
		cerr << ex << endl;
		throw;
		}
	cerr << "sent message" << endl;
	return connection_ok_;
	}

untyped_interest::callresult receiver::accept_msg_network_data(msg_network_data &m)
	{
//	for(int i=0; i<m.length; ++i)
//		cout << m.data[i];
//	cout << endl;
	write(fd,m.data,m.length);
//	cout << id_ << "\t" << "received chunk of length "<<m.length<<endl;
	return untyped_interest::exit_loop;
	}

untyped_interest::callresult receiver::accept_msg_network_status(msg_network_status &m)
	{
	cerr << "got a status message with identifier " <<(int) m.identifier << ": "
		  << m.status_as_text() << endl;
	switch(m.identifier) {
		case msg_network_status::content:
			cerr << "content type: " << m.text << endl;
			break;
		case msg_network_status::finished:
			interest_data_->remove_group_id(id_);
			interest_status_->remove_group_id(id_);
			cond_mutex.lock();
			cond_mutex.broadcast();
			cond_mutex.unlock();
			break;
		case msg_network_status::connection_failed:
			connection_ok_=false;
			interest_data_->remove_group_id(id_);
			interest_status_->remove_group_id(id_);
			cerr << "connection failed" << endl;
			cond_mutex.lock();
			cond_mutex.broadcast();
			cond_mutex.unlock();
			break;
		case msg_network_status::closed:
		case msg_network_status::object_truncated:
			connection_ok_=false;
			interest_data_->remove_group_id(id_);
			interest_status_->remove_group_id(id_);
			cerr << "object truncated" << endl;
			cond_mutex.lock();
			cond_mutex.broadcast();
			cond_mutex.unlock();
			break;
		case msg_network_status::generic_error:
			connection_ok_=false;
			interest_data_->remove_group_id(id_);
			interest_status_->remove_group_id(id_);
			cerr << "generic error" << endl;
			cond_mutex.lock();
			cond_mutex.broadcast();
			cond_mutex.unlock();
			break;
		default:
			cerr << "default id:"<< (int)(m.identifier) << endl;
			break;
		}
	return untyped_interest::exit_loop;
	}

int main(unsigned int argc, char **argv)
	{
	unsigned int ac=1;
  long int st=-1;
	while (ac < argc && argv[ac][0] == '-') {
		switch (argv[ac][1]) {
#ifdef DEBUG
			case 'd':
				debugger::instance().turn_on(argv[ac]+2);
				break;
#endif
			case 'h':
				cout << "test_fetchurl [options] url[s]"<<endl;
				cout << "-d[debug]             debug is the name of the part which"<< endl;
				cout <<	"                      should be debugged, or try \"all\""<< endl;
				cout << "-sn                   send a stop message after n seconds"<< endl;
				exit(0);		
			case 's':
				st=strtol(&(argv[ac][2]),0,0);
        if (!((st==LONG_MIN) || (st==LONG_MAX))) break;
			default:
				cerr << "[test_fetchurl]: unknown option '" << argv[ac] << "'" << endl;
		
				exit(2);
			}
		ac++;
		}

	if(ac==argc) {
		cerr << "try: test_fetchurl -h" << endl;
		exit(2);
		}
	try {
		object_manager::instance().load_interests();
		vector<receiver *> rec;

		receiver::cond_mutex.lock();
		for(unsigned int i=ac; i<argc; ++i) {
			rec.push_back(new receiver(argv[i]));
			rec.back()->start();
			}
		if (st>=0) {
			sleep(st);
			for(unsigned int i=0; i<rec.size(); ++i) {
				rec[i]->stop();
				cerr << "Sent a stop request to receiver " << i <<endl;
				}
			}
		for(unsigned int i=0; i<rec.size(); ++i) {
			cerr << "going to wait" << endl;
			receiver::cond_mutex.wait();
			}
		receiver::cond_mutex.unlock();
//		sleep(2);
		for(unsigned int i=0; i<rec.size(); ++i) {
			delete rec[i];
			cerr << "Receiver " << i << " deleted" << endl;
			}
		object_manager::free_instance();
		}
	catch(messagebase_error& ex) {
		cout << ex << endl;
		exit(1);
		}
	catch(exception& ex) {
		cout << ex.what() << endl;
		exit(1);
		}
	cout << "exit test_fetchurl" << endl;
	}
