// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Mnemonic xml parser implementation.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string>
#include <unistd.h>
#include <ctype.h>
#include <algo.h>

#include "xmlparser.hh"

/*
   Known bugs:
     - Documents beginning with <TITLE> (ie. skipping both <HTML> and <HEAD>
       are valid, but because we only insert one optional tag, they fail to
       parse. Ditto for documents starting with a tag (we already have a kludge
       to parse documents starting with text).
	  - Namespaces are not yet supported.
     - All element names and attributes are converted to uppercase, which is
       incorrect for XML.
     - We recognise encodings, but don't actually do conversions yet.
     - Style info embedded like <STYLE><!-- H3 { font-size: 12pt } --></STYLE>
       should be expanded; it is easiest to do that immediately in the parser.
       (we are kludging around META tag stuff already anyhow, so it is simplest to
        kludge around this deprecated STYLE thing too and not klutter up the code
        elsewhere).
     - Does not yet handle <em/real/  type shorthands.

   Why would anyone be interested in seeing processing instructions in the
   DOM? 
*/


// if you want to see every single character in the logs
//#define VERBOSE

#define DEBUG_CLASS "XMLPARSER"
#include "mnemonic/debug.hh"

parse_error::parse_error(const string& wh, parse_error::error_type er)
	: logic_error(wh), error(er)
	{
	}

ostream& operator<<(ostream& str, const parse_error& er)
	{
	str << "[parse_error]: "; 
	str << er.what() << " ";
	switch(er.error) {
		case parse_error::ILLEGAL_FIRST:
			str << "First non-white character is not a '<'.";
			break;
		case parse_error::EMPTY_DOCUMENT_TYPE:
			str << "Document type specifier is empty.";
			break;
		case parse_error::STRAY_TEXT:
			str << "Text encountered between document type declartion and first tag.";
			break;
		case parse_error::ILLEGAL_EMPTY_SHORTHAND:
			str << "A '/' in a tag did not get followed by '>'";
			break;
		case parse_error::INTERNAL_STACK_UNDERFLOW:
			str << "Internal error: stack underflow.";
			break;
		default:
			str << "[empty description]";
			break;
		}
	return str;
	}

xmlparser::xmlparser(ptr<DOM::Document> doc)
	: baseparser(), document(doc), dtd_(0),
	  keep_whitespace_(-1), processing_instruction_(false), top_of_subtree_(0)
   { 
   debug( "xmlparser constructor");
	history_element tmp(INITIAL, 0, 0);
	parse_history.push_back(tmp);
	current_node_=&(*document);
   }

xmlparser::~xmlparser()
   {
   debug ("~xmlparser");
	if(dtd_)
		delete dtd_;
   }

xmlparser::history_element::history_element(xmlparser::parse_state stat, long where, long line)
	: state(stat), offset_in_stream(where), line_in_stream(line)
	{
	}

baseparser::baseparser(void)
	: offset_in_stream_(0), lines_in_stream_(0), convert_toupper_(false), iconvdesc_((iconv_t)-1)
	{
	}

void baseparser::append_data(int length, const char *data)
	{
	raw_data_.append(data,length);
	}

bool baseparser::is_whitespace_(const string& str) const
	{
	string::const_iterator it=str.begin();
	while(it!=str.end()) {
		if(!isspace((*it))) 
			return false;
		++it;
		}
	return true;
	}

istream& operator>>(istream& str, baseparser& pars)
	{
	int i=0;
	while(str && i<2000) {
		++i;
		pars.raw_data_+=str.get();
		}
	return str;
	}

xmlparser::parse_result xmlparser::parse_buffer(vector<DOM::Node *>& added_nodes)
	{
	debug("xmlparser::parse_buffer");
	top_of_subtree_=current_node_;

	assert(parse_history.size()!=0);
	history_element hist=parse_history.back();
	unsigned long current_offset=hist.offset_in_stream;
	unsigned long current_line=hist.line_in_stream;

	if(hist.state==INTERRUPTED)
		parse_history.pop_back();

	bool         exit_loop=false;
	parse_result return_code=DOCUMENT_END;

	unsigned char curchar, prevchar;
	curchar=255;

	unsigned long totalsize=raw_data_.size();
	while(!exit_loop && current_offset-offset_in_stream_ < totalsize) {
		prevchar=curchar;
	   curchar=raw_data_[current_offset-offset_in_stream_];
		if(tokenise_(curchar)==255) {
			++current_offset;
			continue;
			}
		if(curchar=='\r')
			++current_line;
		if(curchar=='\n') {
			if(prevchar=='\r') {
				++current_offset;
				continue;
				}
			else ++current_line;
			}

		// Take a deep breath and dive into the state machine logic:
		switch(peek_(0).state) {
			case INITIAL:
#ifdef VERBOSE
				debug("INITIAL mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						++current_offset;
						break;
					case '<':
						parse_history.push_back(history_element(DOCUMENT, current_offset, current_line));
						parse_history.push_back(history_element(COMMANDORTAG, current_offset, current_line));
						++current_offset;
						break;
					default:
						debug("BROKEN");
						debug("first character is not < or whitespace");
						debug("assuming <html> was forgotten");
						send_error(current_line, "First character is not < or whitespace.");
						document->set_doctype(get_doctype_("HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\""));
						convert_toupper_=true;
						dtd_=new htmldtd;
						DOM::Element *tmpel=document->createElement("HTML");
						current_node_=current_node_->appendChild(tmpel);
						added_nodes.push_back(tmpel);
						if(document->content_type()=="text/plain") {
							debug("adding <body> and <pre> too, since this is text/plain");
							tmpel=document->createElement("BODY");
							current_node_=current_node_->appendChild(tmpel);
							tmpel=document->createElement("PRE");
							current_node_=current_node_->appendChild(tmpel);
							}
						parse_history.push_back(history_element(NODE, current_offset, current_line));
//						throw parse_error("",parse_error::ILLEGAL_FIRST);
						break;
					}
				break;
			case COMMANDORTAG:
#ifdef VERBOSE
				debug("COMMANDORTAG mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '?':
						parse_history.pop_back();
						tagname_="";
						accumulator_="";
						processing_instruction_=true;
						parse_history.push_back(history_element(PROCINST, current_offset, current_line));
						++current_offset;
						break;
					case '!':
						// FIXME: when a document starts with a comment, this will dump us straight
						// into document mode afterwards. We should go back to initial, probably.
						parse_history.pop_back();
						parse_history.pop_back();
						parse_history.push_back(history_element(COMMAND, current_offset, current_line));
						++current_offset;
						break;
					default:
						debug("assuming doctype is html 3.2");
						document->set_doctype(get_doctype_("HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\""));
						dtd_=new htmldtd;
						convert_toupper_=true;
						parse_history.pop_back();
						parse_history.push_back(history_element(HOOK_OPEN, current_offset, current_line));
						break;
					}
				break;
			case PROCINST:
#ifdef VERBOSE
				debug("PROCINST mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '?':
						++current_offset;
						break;
					case ' ':
						tagname_=accumulator_;
						accumulator_="";
						parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, current_line));
						++current_offset;
						break;
					case '>': {
						if(tagname_.size()==0)
							throw parse_error("empty processing instruction",parse_error::EMPTY_PROCESSING_INSTRUCTION);
						debug("found processing instruction " << tagname_);
						if(tagname_=="xml-stylesheet") {
							
							}
						else {
							map<string,string>::const_iterator it=keyvalpairs_.begin();
							while(it!=keyvalpairs_.end()) {
								string tmp=(*it).first;
								transform(tmp.begin(),tmp.end(),tmp.begin(),toupper);
								if(tmp=="ENCODING") {
									encoding_=(*it).second;
									debug("encoding " << encoding_);
									}
								++it;
								}
							}
						// FIXME: AAARGHH, w3.org screwed up again. Why can't we get the key/value pairs
						// using DOM? 
/*						DOM::ProcessingInstruction *pri=document->createProcessingInstruction(tagname_,"");
						current_node_->appendChild(pri); */
						parse_history.pop_back();
						tagname_="";
						keyvalpairs_.clear();
						accumulator_="";
						processing_instruction_=false;
						++current_offset;
						parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, 
																			 current_line));
						break;
						}
					default:
						if(tagname_.size()>0) {
							parse_history.push_back(history_element(ATTRIBUTE_KEY, current_offset, current_line));
							break;
							}
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case COMMAND:
#ifdef VERBOSE
				debug("COMMAND mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>':
						if(command.size()==0)
							throw parse_error("empty document type",parse_error::EMPTY_DOCUMENT_TYPE);
//						debug("ignored command " << command);
						command="";
						parse_history.pop_back();
						++current_offset;
						break;
					default:
						command+=toupper(curchar);
						if(command=="--") {
							parse_history.pop_back();
							parse_history.push_back(history_element(SCAN_COMMENT, current_offset, current_line));
							command="";
							}
						if(command=="DOCTYPE") {
							parse_history.pop_back();
							parse_history.push_back(history_element(DOCTYPE, current_offset, current_line));
							parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, current_line));
							}
						++current_offset;
					}
				break;
			case DOCTYPE:
#ifdef VERBOSE
				debug("DOCTYPE mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>': {
						if(doctype.size()==0)
							throw parse_error("empty document type",parse_error::EMPTY_DOCUMENT_TYPE);
						debug("document type is " << doctype);
						document->set_doctype(get_doctype_(doctype));
						string tmp=doctype;
						transform(tmp.begin(),tmp.end(),tmp.begin(),toupper);
						if(tmp.size()>=4)
							if(tmp.substr(0,4)=="HTML")
								convert_toupper_=true;
						// FIXME: use right dtd.
						dtd_=new htmldtd;
						parse_history.push_back(history_element(DOCUMENT, current_offset, current_line));
						++current_offset;
						break;
						}
					default:
						doctype+=curchar;
						++current_offset;
					}
				break;
			case DOCUMENT:
#ifdef VERBOSE
				debug("DOCUMENT mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '<':
						parse_history.push_back(history_element(HOOK_OPEN, current_offset, current_line));
						++current_offset;
						break;
					case '\n':
					case ' ':
						++current_offset;
						break;
					default:
						debug("ILLEGAL");
						debug("ignoring character " << curchar << " between start and first valid tag");
						send_error(current_line, "Illegal character between start and first valid tag.");
						accumulator_="";
						++current_offset;
						break;
					}
				break;
			case STRING:
#ifdef VERBOSE
				debug("STRING mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case 34:   // the usual C expression makes Emacs colour mode go berserk ;-)
						parse_history.pop_back();
						accumulator_+=curchar;
						++current_offset;
						break;
					default:
						accumulator_+=curchar;
						++current_offset;
					}
				break;
			case QUOTE:
#ifdef VERBOSE
				debug("QUOTE mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '\'':
						parse_history.pop_back();
						accumulator_+=curchar;
						++current_offset;
						break;
					default:
						accumulator_+=curchar;
						++current_offset;
					}
				break;
			case COLLECT_WHITESPACE:
#ifdef VERBOSE
				debug("COLLECT_WHITESPACE mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
					case '\t':
						++current_offset;
						break;
					default:
						parse_history.pop_back();
						break;
					}
				break;
			case NODE:
#ifdef VERBOSE
				debug("NODE mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '\n':
					case ' ':
						if(dtd_) {
							if(keep_whitespace_==-1) {
								// This property is inherited, <B> inside <PRE> still keeps whitespace.
								DOM::Node *tmpn=current_node_;
								while(tmpn) {
									if((keep_whitespace_=dtd_->keep_whitespace(tmpn->nodeName()))) {
										break;
										}
									tmpn=tmpn->parentNode();
									}
								}
							if(!keep_whitespace_) {
								accumulator_+=" ";
								// we peek ahead to optimise a little bit
								if(current_offset-offset_in_stream_<totalsize-1) {
									if(tokenise_(raw_data_[current_offset-offset_in_stream_+1])==' ')
									parse_history.push_back(
										history_element(COLLECT_WHITESPACE, current_offset, current_line));
									}
								}
							else accumulator_+=curchar;
							}
						else accumulator_+=curchar;
						++current_offset;
						break;
					case '<': {
						if(accumulator_.size()>0) {
							if(dtd_ && !dtd_->valid_child(current_node_->nodeName(),"#text")) {
								if(!is_whitespace_(accumulator_)) {
									debug("ILLEGAL");
									debug("text " << accumulator_ << " not allowed as child of " 
											<< current_node_->nodeName());
									send_error(current_line, "No text allowed here.");
									// FIXME: try to open optional beginning tags like we do for 
									// insertion of non-textnodes.
									}
								else {
									debug("ignoring whitespace after node " << current_node_->nodeName());
									}
								}
							else {
/*								if(accumulator_.size()>20) {
									debug("found textnode |" << accumulator_.substr(0,20) << " ...|");
									}
								else {
									debug("found textnode |" << accumulator_ << "|");
									} */
								// FIXME: this is not handled correctly by the display, so we cheat
//								if(current_node_->lastChild() && current_node_->lastChild()->nodeName()=="#text") {
//									((DOM::Text *)(current_node_->lastChild()))->appendData(accumulator_);
//									}
//								else {
									DOM::Text *tmpel=document->createTextNode(accumulator_);
									current_node_->appendChild(tmpel);
									if(current_node_==top_of_subtree_) {
										debug("new subtree at " << tmpel->nodeName());
										added_nodes.push_back(tmpel);
										}
//									}
								}
							accumulator_="";
							}
						keep_whitespace_=-1;
						parse_history.push_back(history_element(HOOK_OPEN, current_offset, current_line));
						++current_offset;
						break;
						}
					default:
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case SHORTHAND_EMPTY:
#ifdef VERBOSE
				debug("SHORTHAND_EMPTY mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>': {
//						debug("scanned empty tag <" << accumulator_ << "/>" << " @line " << current_line+1);
						DOM::Element *tmpel=document->createElement(accumulator_);
						current_node_->appendChild(tmpel);
						if(current_node_==top_of_subtree_) {
//							debug("new subtree at " << tmpel->nodeName());
							added_nodes.push_back(tmpel);
							}
						accumulator_="";
						tagname_="";
						parse_history.pop_back();
						parse_history.pop_back();
						++current_offset;
						parse_history.push_back(history_element(NODE, current_offset, current_line));
						break;
						}
					default:
						debug("ILLEGAL");
						debug("probably messed up spacing between tagname and attribute");
						send_error(current_line, "Error scanning attribute.");
						parse_history.pop_back();
//						throw parse_error("",parse_error::ILLEGAL_EMPTY_SHORTHAND);
						break;
					}
				break;
			case HOOK_OPEN:
#ifdef VERBOSE
				debug("HOOK_OPEN mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case 34:
						accumulator_+=curchar;
						parse_history.push_back(history_element(STRING, current_offset, current_line));
						++current_offset;
						break;
					case '\'':
						accumulator_+=curchar;
						parse_history.push_back(history_element(QUOTE, current_offset, current_line));
						++current_offset;
						break;
					case '>': {
						if(tagname_=="")
							tagname_=accumulator_;
						else
							accumulator_=tagname_;
//						debug("scanned tag <" << tagname_ << ">"  << " @line " << current_line+1);
						if(tagname_=="META") {
							map<string,string>::const_iterator it=keyvalpairs_.find("HTTP-EQUIV");
							if(it!=keyvalpairs_.end() && keyvalpairs_["HTTP-EQUIV"]=="Content-Type") {
								it=keyvalpairs_.find("CONTENT");
								if(it!=keyvalpairs_.end()) {
									string tmp=(*it).second;
									size_t sit=tmp.find("charset=");
									if(sit!=tmp.npos) {
										size_t sit2=tmp.find_first_of("\";",sit+8);
										encoding_=tmp.substr(sit+8,sit2-sit-8);
										debug("encoding " << encoding_);
										}
									}
								}
							}
						DOM::Element *tmpel=document->createElement(tagname_);
						map<string,string>::const_iterator it=keyvalpairs_.begin();
						while(it!=keyvalpairs_.end()) {
//							debug((*it).first << "=" << (*it).second);
							if((*it).second=="")
								tmpel->setAttribute((*it).first, "TRUE");
							else
								tmpel->setAttribute((*it).first, (*it).second);
							++it;
							}

						bool allowed_=false;
						bool ignore_=false;
						bool new_subtree_top_=false;
						DOM::Node *tmpnode=current_node_;
						if(dtd_) {
							do {
								allowed_=dtd_->valid_child(tmpnode->nodeName(), tagname_);
								if(!allowed_) {
									if(!(dtd_->optional_endtag(tmpnode->nodeName()))) {
										debug("BROKEN");
										debug("<" << tagname_ << "> not allowed as child of " << tmpnode->nodeName());
										debug("and " << tmpnode->nodeName() << " does not have optional end tag");
										debug("so we assume that </" << tmpnode->nodeName() << "> was forgotten");
										debug("and insert it by hand");
										if(tmpnode==top_of_subtree_)
											new_subtree_top_=true;
										tmpnode=tmpnode->parentNode();
										}
									else {
										debug("assuming implicit end tag for " <<  tmpnode->nodeName());
										if(tmpnode==top_of_subtree_)
											new_subtree_top_=true;
										tmpnode=tmpnode->parentNode();
										}
									}
								} while(!allowed_ && !ignore_ && tmpnode!=0);
							}

						if(tmpnode!=0) {
							if(!ignore_) {
//								debug("adding <" << tagname_ << "> as child of " << tmpnode->nodeName());
								if(!dtd_ || !(dtd_->empty_element(tagname_))) {
									current_node_=tmpnode->appendChild(tmpel);
									if(new_subtree_top_) {
										top_of_subtree_=tmpnode;
//										debug("new subtree at " << tmpel->nodeName());
										added_nodes.push_back(tmpel);
										}
									else if(tmpnode==top_of_subtree_) {
//										debug("new subtree at " << tmpel->nodeName());
										added_nodes.push_back(tmpel);
										}
									}
								else {
									tmpnode->appendChild(tmpel);
									current_node_=tmpnode;
									if(new_subtree_top_) {
										top_of_subtree_=tmpnode;
//										debug("new subtree at " << tmpel->nodeName());
										added_nodes.push_back(tmpel);
										}
									else if(tmpnode==top_of_subtree_) {
//										debug("new subtree at " << tmpel->nodeName());
										added_nodes.push_back(tmpel);
										}
									}
								}
							accumulator_="";
							tagname_="";
							parse_history.pop_back();
							parse_history.push_back(history_element(NODE, current_offset, current_line));
							// Eat whitespace immediately after opening tag, app B.3.1 of HTML 4.0.
							// This property is inherited, so a <B> tag inside a <PRE> still keeps whitespace.
							bool keepwh=false;
							DOM::Node *tmpn=current_node_;
							if(dtd_) {
								while(tmpn) {
									if(dtd_->keep_whitespace(tmpn->nodeName())) {
										keepwh=true;
										break;
										}
									tmpn=tmpn->parentNode();
									}
								}
							if(!keepwh) {
								parse_history.push_back(history_element(
									COLLECT_WHITESPACE, current_offset, current_line));
								}
							}
						else {
							string ins=dtd_->find_optional_enclosing_tag(tagname_);
							if(ins!="" && dtd_->valid_child(current_node_->nodeName(),ins)) {
								debug("that did not work.");
								debug("inserting optional tag <" << ins << "> will do the trick");
								DOM::Element *opttag=document->createElement(ins);
								if(top_of_subtree_==current_node_) {
//									debug("new subtree at " << opttag->nodeName());
									added_nodes.push_back(opttag);
									}
								current_node_=current_node_->appendChild(opttag);
								if(!(dtd_->empty_element(tagname_))) {
									current_node_=current_node_->appendChild(tmpel);
									}
								else {
									current_node_->appendChild(tmpel);
									}
								}
							else {
								debug("ILLEGAL");
								debug("no optional (and valid) enclosing tag found");
								debug("No matter how many forgotten end tags we inserted,");
								debug("the <" << tagname_ << "> dit not fit in. Ignoring it.");
								send_error(current_line, "Tag not allowed here.");
								}
							accumulator_="";
							tagname_="";
							parse_history.pop_back();
							parse_history.push_back(history_element(NODE, current_offset, current_line));
							// eat whitespace immediately after opening tag, app B.3.1 of HTML 4.0.
							parse_history.push_back(history_element(
								COLLECT_WHITESPACE, current_offset, current_line));
							}
						keyvalpairs_.clear();
						++current_offset;
						break;
						}
					case '/':
						if(accumulator_.size()==0 && tagname_.size()==0) {
							parse_history.pop_back();
							parse_history.push_back(history_element(CLOSINGTAG_OPEN, current_offset, current_line));
							++current_offset;
							break;
							}
						else { // shorthand empty tag
							if(accumulator_.size()==0)
								accumulator_=tagname_;
							parse_history.push_back(history_element(SHORTHAND_EMPTY, current_offset, current_line));
							++current_offset;
							break;
							}	
					case '\n':
					case ' ':
					case '\t': {
						tagname_=accumulator_;
						accumulator_="";
						key_="";
						val_="";
						++current_offset;

						if(dtd_ && dtd_->can_contain_brackets(current_node_->nodeName())) {
							debug("as " << current_node_->nodeName() << " can contain brackets");
							debug("I'm assuming that < was meant as text");
							accumulator_="<"+tagname_+" ";
							tagname_="";
							parse_history.pop_back();
							}
						else {
							parse_history.push_back(history_element(ATTRIBUTE_KEY, current_offset, current_line));
							parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, 
																				 current_line));
							}
						break;
						}
					case '<': 
						if(accumulator_=="![CDATA[") {
							tagname_="";
							accumulator_="";
							parse_history.pop_back();
							parse_history.push_back(history_element(CDATA, current_offset, current_line));
							break;
							}
						else {
							parse_history.pop_back();
							accumulator_="<<";
							++current_offset;
							parse_history.push_back(history_element(NODE, current_offset, current_line));
							break;
							}
					case '?':
						if(accumulator_.size()==0) {
							parse_history.pop_back();
							tagname_="";
							accumulator_="";
							processing_instruction_=true;
							parse_history.push_back(history_element(PROCINST, current_offset, current_line));
							++current_offset;
							break;
							}
					case '!':
						if(accumulator_.size()==0) {
							parse_history.pop_back();
							parse_history.push_back(history_element(COMMAND, current_offset, current_line));
							++current_offset;
							break;
							}
					default:
						if(convert_toupper_) accumulator_+=toupper(curchar);
						else                 accumulator_+=curchar;
						if(accumulator_=="!--") {	
							parse_history.pop_back();
							parse_history.push_back(history_element(SCAN_COMMENT, current_offset, current_line));
							accumulator_="";
							}
						++current_offset;
					}
				break;
			case CDATA:
#ifdef VERBOSE
				debug("CDATA mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>':
						if(accumulator_.size()>=2)
							if(accumulator_.substr(accumulator_.size()-2)=="]]") {
								DOM::CDATASection *cdat=document->createCDATASection(
									accumulator_.substr(0,accumulator_.size()-2));
//								debug("scanned CDATA " << cdat->nodeValue());
								current_node_->appendChild(cdat);
								accumulator_="";
								parse_history.pop_back();
								++current_offset;
								break;
								}
					default:
						accumulator_+=curchar;
						++current_offset;
					}
				break;
			case SCAN_COMMENT:
#ifdef VERBOSE
				debug("SCAN_COMMENT mode |" << (int)(curchar) << "|" << curchar);
#endif
				accumulator_+=curchar;
				if(accumulator_.size()>=3)
					if(accumulator_.substr(accumulator_.size()-3)=="-->") {
						parse_history.pop_back();
//						debug("comment " << accumulator_.substr(0,accumulator_.size()-3) << " ignored");
						accumulator_="";
						}
				++current_offset;
				break;
			case SCAN_SGML_COMMENT:
#ifdef VERBOSE
				debug("SCAN_SGML_COMMENT mode |" << (int)(curchar) << "|" << curchar);
#endif
				accumulator_+=curchar;
				if(accumulator_.size()>=2)
					if(accumulator_.substr(accumulator_.size()-2)=="--") {
						parse_history.pop_back();
//						debug("sgml comment " << accumulator_.substr(0,accumulator_.size()-3) << " ignored");
						accumulator_="";
						}
				++current_offset;
				break;
			case ATTRIBUTE_KEY:
#ifdef VERBOSE
				debug("ATTRIBUTE_KEY mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>':
						if(accumulator_.size()>0) {
							if(accumulator_[accumulator_.size()-1]=='/') {
								accumulator_=accumulator_.substr(0,accumulator_.size()-1);
								--current_offset;
								debug("shorthand empty");
								}
							key_=accumulator_;
							}
						if(key_!="") {
							if(val_!="") {
								if(val_[val_.size()-1]=='/') {
									// Slashdot compatibility junk. They use key/val pairs with
									// the value ending in a slash, immediately followed by the closing
									// hook. Would be recognised as a shorthand empty otherwise.
									if(val_.size()>1 && val_[val_.size()-2]==' ') {
										val_=val_.substr(0,val_.size()-1);
										--current_offset;
										debug("shorthand empty with keyval pair");
										}
									else {
										debug("BROKEN");
										debug("last attribute value ends in a slash");
										send_error(current_line, "Attributes should not end with a slash.");
										}
									}
								keyvalpairs_[key_]=val_;
//								debug("attribute pair " << key_ << "=" << val_);
								// FIXME: if the atttribute value is from a fixed set, convert to
								// upper case (need DTD support for this)
								// Maybe we should also keep the quotes on a value so that we can
								// warn if the quotes have been forgotten illegally.
								}
							else {
								keyvalpairs_[key_]="";  // FIXME: is this according to specs?
//								debug("single attribute " << accumulator_);
								}
							}
						parse_history.pop_back();
						// feed it again to HOOK_OPEN.
						break;
					case '=':
						if(key_=="")
							key_=accumulator_;
						accumulator_="";
						parse_history.push_back(history_element(ATTRIBUTE_VAL, current_offset, current_line));
						parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, current_line));
						++current_offset;
						break;
					case '\n':
					case ' ':
						key_=accumulator_;
						val_="";
						parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, current_line));
						++current_offset;
						break;
					case '-':
						if(accumulator_=="-") {
							accumulator_="";
							parse_history.push_back(history_element(SCAN_SGML_COMMENT, current_offset, 
																				 current_line));
							++current_offset;
							break;
							}
					case '?':
						if(processing_instruction_) {
							++current_offset;
							break;
							}
					default:
						if(val_!="") {  // 
//							debug("attribute pair " << key_ << "=" << val_);
								// FIXME: if the atttribute value is from a fixed set, convert to
								// upper case (need DTD support for this)
							keyvalpairs_[key_]=val_;
							key_="";
							val_="";
							}
						if(key_!="") {
//							debug("single attribute " << key_);
							keyvalpairs_[key_]="";
							accumulator_="";
							key_="";
							break;
							}
						if(convert_toupper_) accumulator_+=toupper(curchar);
						else                 accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case ATTRIBUTE_VAL:
#ifdef VERBOSE
				debug("ATTRIBUTE_VAL mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case '>':
						val_=accumulator_;
						accumulator_="";
						parse_history.pop_back();
						// feed it again to HOOK_OPEN.
						break;
					case 34:
						accumulator_+=curchar;
						parse_history.push_back(history_element(STRING, current_offset, current_line));
						++current_offset;
						break;
					case '\'':
						accumulator_+=curchar;
						parse_history.push_back(history_element(QUOTE, current_offset, current_line));
						++current_offset;
						break;
					case '\n':
					case ' ':
						val_=accumulator_;
						accumulator_="";
						parse_history.pop_back();
						parse_history.push_back(history_element(COLLECT_WHITESPACE, current_offset, current_line));
						++current_offset;
						break;
					case '?':
						if(processing_instruction_) {
							++current_offset;
							break;
							}
					default:
						accumulator_+=curchar;
						++current_offset;
						break;
					}
				break;
			case CLOSINGTAG_OPEN:
#ifdef VERBOSE
				debug("CLOSINGTAG_OPEN mode |" << (int)(curchar) << "|" << curchar);
#endif
				switch(tokenise_(curchar)) {
					case 34:
						accumulator_+=curchar;
						parse_history.push_back(history_element(STRING, current_offset, current_line));
						++current_offset;
						break;
					case '\'':
						accumulator_+=curchar;
						parse_history.push_back(history_element(QUOTE, current_offset, current_line));
						++current_offset;
						break;
					case '>':
//						debug("scanned closing tag </" << accumulator_ << ">" << " @line " << current_line+1);
						// FIXME: this is just a dirty hack, will go away once we properly parse DTDs
						//
						// If the closing tag is associated to an empty
						// element, ignore it.  Else, (1) check whether the
						// current parent is the matching opening tag. If
						// so, set parent to parent->parent. If not, check
						// whether the current parent has an optional close
						// tag. If it has, set parent->parent and try again at (1).
						if(!dtd_ || !(dtd_->empty_element(accumulator_))) {
							DOM::Node *tmpnode=current_node_;
							bool new_subtree_top_=false;
							while(tmpnode!=0) {
//								debug(tmpnode->nodeName());
								if(tmpnode->nodeName()==accumulator_)
									break;
/* FIXME: we don't do anything with this, so we may as well not test for optional endtag...
								if(dtd_)
									if(!(dtd_->optional_endtag(tmpnode->nodeName()))) {
										debug("previous tag has non-optional closing tag (trying to close it anyway)");
										}
*/
								if(tmpnode==top_of_subtree_)
									new_subtree_top_=true;
								tmpnode=tmpnode->parentNode();
								}
							if(tmpnode==0) {
								debug("arrived at top without proper opening tag");
								debug("bad. let's ignore the closing tag instead");
								debug("ILLEGAL");
								debug("superfluous closing tag </" << accumulator_ << ">" );
								send_error(current_line, "Superfluous closing tag.");
								}
							else {
								current_node_=tmpnode->parentNode();
								if(new_subtree_top_ || tmpnode==top_of_subtree_)
									top_of_subtree_=current_node_;
								}
							if(current_node_==0)
								throw parse_error("while scanning closing tag",
														parse_error::INTERNAL_STACK_UNDERFLOW);
							}
//						else {
//							debug("that tag is empty, so we ignore the closing tag");
//							}
						accumulator_="";
						parse_history.pop_back();
						parse_history.push_back(history_element(NODE, current_offset, current_line));
						++current_offset;
						break;
					default:
						if(convert_toupper_) accumulator_+=toupper(curchar);
						else                 accumulator_+=curchar;
						++current_offset;
					}
				break;
			default:
				++current_offset;
				debug("unimplemented feature of parser triggered");
			}
		}
	if(!exit_loop)
		parse_history.push_back(history_element(INTERRUPTED, current_offset, current_line));

	// FIXME: be intelligent
	debug("xmlparser::parser_buffer done");
	return return_code;
	}

const xmlparser::history_element& xmlparser::peek_(unsigned int level=0) const
	{
	if(level+1>parse_history.size())
		throw parse_error("while examining stack",parse_error::INTERNAL_STACK_UNDERFLOW);
	return parse_history[parse_history.size()-1-level];
	}

unsigned char xmlparser::tokenise_(unsigned char c) const
	{
	if(c=='\n' || c=='\r')    return '\n';
	if(isspace(c))            return ' ';
	if(c==255 || c==254)
		return 255; // here we can filter out illegal characters
	return c;
	}

DOM::DocumentType *xmlparser::get_doctype_(const string& doctype)
	{
	string mytype=doctype;
	transform(mytype.begin(),mytype.end(),mytype.begin(),toupper);
	if(doctype=="HTML")
		return new DOM::DocumentType("HTML 4.0 blabla"); // FIXME: should read the dtd of course.
	return new DOM::DocumentType(doctype);
	}




/* dtdparser::dtdparser(void)
	{

		// Does DOM want this in the DOM::Element node?!?
		// I suggest DOM::ElementType  as it is called that way in the XML specs
		<!ELEMENT br EMPTY>
		<!ELEMENT p (#PCDATA|emph)* >
		<!ELEMENT %name.para; %content.para; >
		<!ELEMENT container ANY>

		// plus members `ElementContent', `MixedContent' and `Attributes'

		// There is a class DOM::Entity for this; good.
		<!ENTITY open-hatch SYSTEM "http://bla">

		// No idea what this is good for, but DOM::Notation has it.
		<!NOTATION Name (ExternalID | PublicID) >

		// DOM::ProcessingInstruction
		<? stuff

		// CDATA
		<![CDATA[ ... ]]>

	}	 */

