// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Actual XML parser (one of these is associated to each document).
  Sends out msg_dom_mutation_event messages.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef mnemonic_xmlparser_h_
#define mnemonic_xmlparser_h_

#include <string>
//#include <rope>
#include <vector.h>
#include <map.h>
#include <set.h>
#include <multimap.h>
#include <iconv.h>

#include <dom/dom.hh>
#include <mnemonic/ptr.hh>
#include "dtd.hh"
#include "htmldtd.hh"

class parse_error : public logic_error {
	public:
		enum error_type { ILLEGAL_FIRST,            // first non-white character is not '<'
								EMPTY_DOCUMENT_TYPE,      // document type was '<>' with no text.
								EMPTY_PROCESSING_INSTRUCTION, // processing instruction <?>
								STRAY_TEXT,               // text encountered between doctype and first tag
								ILLEGAL_EMPTY_SHORTHAND,  // '/' in tag did not get followed by '>'
								INTERNAL_STACK_UNDERFLOW  // tried to access non-existing element of stack
		};

		parse_error(const string& wh, error_type);
		
		error_type error;

};

ostream& operator<<(ostream&, const parse_error&);

class baseparser {
	public:
		baseparser(void);

		void append_data(int length, const char *data);
		friend istream& operator>>(istream& str, baseparser&);
	protected:
		// We keep all raw data in raw_data, but in case we will ever
		// want to drop parts that have already been processed (eg. to
		// save memory) we also keep offset counters.

		string  raw_data_;          // FIXME: replace with `rope' when they removed the warnings in stl_rope.h
		unsigned long  offset_in_stream_;    // The place where raw_data[0] is located in the stream.
		unsigned long  lines_in_stream_;     // How many lines were spotted before offset_in_stream.

		string                   accumulator_;
		string                   tagname_;
		string                   key_,val_;
		map<string,string>       keyvalpairs_;

		bool is_whitespace_(const string&) const;

		bool    convert_toupper_;
		string  encoding_;
		iconv_t iconvdesc_;
};

class xmlparser : public baseparser {
   public:
		xmlparser(ptr<DOM::Document>);
		virtual ~xmlparser();


		enum parse_result { DOCTYPE_FOUND,      // DTD declaration found
								  TAG_CREATED,        // new tag created
								  OUT_OF_DATA,        // buffer ended while scanning
								  DOCUMENT_END        // 
		};
								  
		parse_result parse_buffer(vector<DOM::Node *>& added_nodes);
		ptr<DOM::Document>  document;		
   private:
		DOM::Node          *current_node_;
		fakedtd            *dtd_;  // 0 if no dtd associated to document (eg. xml)

		enum parse_state { INITIAL,             // nothing done yet
								 COMMANDORTAG,        // have opened <
								 PROCINST,            // scanning processing instruction (<? tag)
								 COMMAND,             // scanning a <! tag
								 DOCTYPE,             // scanning for document type declaration
								 DOCUMENT,            // scanning document following doctype declaration
								 NODE,                // scanning arbitrary text, not knowing what it will become
								 COLLECT_WHITESPACE,  // collapsing whitespace into accumulator
								 HOOK_OPEN,           // reading text after '<'
								 SHORTHAND_EMPTY,     // have read '/' inside tag, hints at shorthand '/>' 
                         CLOSINGTAG_OPEN,     // reading closing tag after '</'
								 ATTRIBUTE_KEY,       // scanning attribute key
								 ATTRIBUTE_VAL,       // scanning attribute value (if any)
								 STRING,              // reading string ("...")
								 QUOTE,               // reading quote  ('...')
								 CDATA,               // scanning argument of <!CDATA[[ 
								 SCAN_COMMENT,        // scanning '<!-- blabla -->'.
								 SCAN_SGML_COMMENT,   // scanning '-- blabla --'.								 
								 INTERRUPTED,         // temporarily ran out of data

								 CONDITIONAL_OPEN,    // read '<!'
								 COMMAND_OPEN,        // read '<?'
								 COMMENT_OPEN1,       // read '<!-'
								 COMMENT_OPEN2,       // read '<!--'
								 TAGCOMMENT_OPEN,     // read '--' in tag while _not_ scanning comment
								 TAGCOMMENT_CLOSE,    // read '--' in tag while scanning comment
								 CDATA_OPEN,          // read '<![CDATA['
								 TAG_CLOSE,           // read '>' while in tag
								 CONDITIONAL_CLOSE    // read ']]>' after successful CDATA or CONDITIONAL
		};

		class history_element {
			public:
				history_element(parse_state, long, long);

				parse_state state;
				unsigned long        offset_in_stream;
				unsigned long        line_in_stream;
		};

		vector<history_element>  parse_history;
		string                   doctype;
		string                   command;

		DOM::DocumentType      *get_doctype_(const string&);
		void                    send_error(long, const string&);
		const history_element&  peek_(unsigned int level=0) const;
		unsigned char           tokenise_(unsigned char c) const;
	   int                     keep_whitespace_;  // -1: don't know, 0: no, 1: yes
		bool                    processing_instruction_;
		DOM::Node              *top_of_subtree_;
};

class dtdparser : public baseparser {
	public:
		dtdparser(ptr<DOM::Document>);
		virtual ~dtdparser(); 

		enum parse_result { DOCTYPE_FOUND,      // DTD declaration found
								  TAG_CREATED,        // new tag created
								  OUT_OF_DATA,        // buffer ended while scanning
								  DOCUMENT_END        // 
		};
								  
		parse_result parse_buffer(void);
		ptr<DOM::Document>  document;		
   private:
		enum parse_state { INITIAL,             // nothing done yet
								 COLLECT_WHITESPACE,  // collapsing whitespace into accumulator
								 HOOK_OPEN,           // reading text after '<'
                         CLOSINGTAG_OPEN,     // reading closing tag after '</'
								 STRING,              // reading string ("...")
								 QUOTE,               // reading quote  ('...')
								 INTERRUPTED,         // temporarily ran out of data
		};

		class history_element {
			public:
				history_element(parse_state, long, long);

				parse_state state;
				unsigned long        offset_in_stream;
				unsigned long        line_in_stream;
		};

		vector<history_element>  parse_history;

		const history_element&  peek_(unsigned int level=0) const;
		unsigned char           tokenise_(unsigned char c) const;
};


istream& operator>>(istream& str, baseparser&);

#endif
