// -*- mode: C++; c-file-style: "mnemonic"; -*-

/*
  Simple program to fetch an url and dump it to stdout.

  Please send bugs, changes, problems to bugs@mnemonic.browser.org

  This program and its components are free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2, or (at your option) any later version.
 
  This program and its components are distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License in
  the file COPYING accompanying this program; if not, write to the
  Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>

#include <mnemonic/optionals.h>
#include <mnemonic/debugger.hh>
#include <mnemonic/objectmanager.hh>
#include <mnemonic/nthreadcond.hh>
#include <mnemonic/midentifier.hh>
#include <msg/network/get_url.hh>
#include <msg/network/network_status.hh>
#include <msg/network/network_data.hh>

class receiver {
	public:
		receiver(const char *url);
		~receiver();
		untyped_interest::callresult accept_msg_network_data(msg_network_data &m);
		untyped_interest::callresult accept_msg_network_status(msg_network_status &m);
		bool start(void);
		
		thread_cond cond_mutex;
	private:
		url_string url_;
		bool connection_ok_;
		ptr<untyped_interest> interest_data_, interest_status_;
		keyed_interest<receiver, msg_network_data, midentifier>   *keyed_data_;
		keyed_interest<receiver, msg_network_status, midentifier> *keyed_status_;
		midentifier id_;
};

receiver::receiver(const char *url)
	{
	url_=url;

	keyed_data_=new keyed_interest<receiver, msg_network_data, midentifier>
		(this,&receiver::accept_msg_network_data, untyped_interest::normal);
	interest_data_=keyed_data_;

	keyed_status_=new keyed_interest<receiver, msg_network_status, midentifier>
		(this,&receiver::accept_msg_network_status, untyped_interest::normal);
	interest_status_=keyed_status_;
	}

receiver::~receiver()
	{
	interest_data_->unregister();
	interest_status_->unregister();
	
	interest_data_.wait();
	interest_status_.wait();
	}

bool receiver::start(void)
	{
	keyed_data_->add_key(id_);
	keyed_status_->add_key(id_);
	cerr << "making a get_url message" <<endl;
	msg_get_url mgu(id_,url_);
	cerr << "fetching " << url_.get_url() << endl;
	connection_ok_=true;
	try { mgu.send(); }
	catch(messagebase_error &ex) {
		cerr << ex << endl;
		throw;
		}
	cerr << "sent message" << endl;
	return connection_ok_;
	}

untyped_interest::callresult receiver::accept_msg_network_data(msg_network_data &m)
	{
	for(int i=0; i<m.length; ++i)
		cout << m.data[i];
	cout << endl;
	return untyped_interest::call_next;
	}

untyped_interest::callresult receiver::accept_msg_network_status(msg_network_status &m)
	{
	cerr << "got a status message with identifier " << m.identifier << ": "
		  << m.status_as_text() << endl;
	switch(m.identifier) {
		case msg_network_status::finished:
			keyed_data_->remove_key(id_);
			keyed_status_->remove_key(id_);
			cond_mutex.broadcast();
			break;
		case msg_network_status::connection_failed:
			connection_ok_=false;
			keyed_data_->remove_key(id_);
			keyed_status_->remove_key(id_);
			cerr << "connection failed" << endl;
			break;
		case msg_network_status::generic_error:
			connection_ok_=false;
			keyed_data_->remove_key(id_);
			keyed_status_->remove_key(id_);
			cerr << "generic error" << endl;
			break;
		default:
			break;
		}
	return untyped_interest::call_next;
	}

int main(int argc, char **argv)
	{
	int ac=1;
	while (ac < argc && argv[ac][0] == '-') {
		switch (argv[ac][1]) {
#ifdef DEBUG
			case 'd':
			case 'D':
				debugger::instance().turn_on(argv[ac]+2);
				break;
#endif
			default:
				cerr << "[test_fetchurl]: unknown option '" << argv[ac] << "'" << endl;
			}
		ac++;
		}

	if(ac==argc) {
		cout << "usage: test_fetchurl [url]" << endl;
		exit(2);
		}
	try {
		object_manager::instance().load_interests();
		for(unsigned int i=ac; i<argc; ++i) {
			receiver tmp(argv[i]);
			if(tmp.start())  { // the connection is ok, wait for msg_network_data to come in
				tmp.cond_mutex.wait(); 
//				sleep(5);
				}
			}
		object_manager::free_instance();
		}
	catch(messagebase_error& ex) {
		cout << ex << endl;
		exit(1);
		}
	catch(exception& ex) {
		cout << ex.what() << endl;
		exit(1);
		}
	cout << "exit test_fetchurl" << endl;
	}
